/**********************************************************************
 **                                                                  **
 **                            RComAPI.h                             **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  The interface class for RCom.                                   **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef RComAPI_H
#define RComAPI_H

/* $Id: RComAPI.h,v 1.1 2008-07-15 22:33:56 endo Exp $ */

#include <string>

using std::string;

class RComAPI {

public:
    //----------------------------------------------------------------
    // Destructor
    //----------------------------------------------------------------
    virtual ~RComAPI(void) {}

    //----------------------------------------------------------------
    // This function will initialize the setups.
    // 
    // Input:
    //  - myID: an unique ID that distinguish itself from others. This
    //    ID will be used used by others to deliver the data to this
    //    connection. Use one of the IDs defined in RComIDs.h.
    //  - readerDataCallbackFunc: Function to be executed when the data is
    //    arrived. (See below.)
    //  - extraCallbackFuncArg: An extra data pointer that is going to be pass
    //    to the last argument of readerCallbackFunc().
    //
    //    Below is how the callback function should look like:
    //
    //    bool readerCallbackFunc(
    //        int recipientID,
    //        const void *data,
    //        size_t dataSize,
    //        bool reliableEnforced
    //        void *extraArg);
    // 
    //     * recipientID: The ID of the application this data needs to
    //       be shipped to.
    //     * data: pointer to the data
    //     * dataSize: size of the data.
    //     * reliableEnforced: the flag to notifies that whether this
    //       data was sent via absolutely reliable connection or
    //       if the loss of the data could have been tolerated. 
    //     * extraArg: An extra data pointer that was passed by the caller
    //       via the last argument of initialization().
    // 
    // Output:
    //  - Returns true if success. Otherwise, false.
    //----------------------------------------------------------------
    virtual bool initialize(
        int myID,
        void(*readerCallbackFunc)(int, const void *, size_t, bool, void *),
        void *extraCallbackFuncArg) = 0;

    //----------------------------------------------------------------
    // This function will send the data.
    // 
    // Input:
    //  - recipientID: ID of the recipient application.
    //  - data: data to be sent.
    //  - dataSize: the size of the data.
    //  - enforceReliable: the flag to notifies that whether this
    //    data should be sent via absolutely reliable connection or
    //    if the loss of the data can be tolerated. 
    //
    // Output:
    //  - Returns true if success. Otherwise, false.
    //----------------------------------------------------------------
    virtual bool sendData(
        int recipientID,
        const void *data,
        size_t dataSize,
        bool enfoceReliable) = 0;
};

#endif
/**********************************************************************
 * $Log: RComAPI.h,v $
 * Revision 1.1  2008-07-15 22:33:56  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.6  2006/02/25 16:17:49  endo
 * rcomserver added
 *
 * Revision 1.5  2006/02/22 07:36:15  endo
 * *** empty log message ***
 *
 * Revision 1.4  2006/02/22 07:17:22  endo
 * RComServer and RComClient implemented.
 *
 * Revision 1.3  2006/02/20 19:07:28  endo
 * Port added
 *
 * Revision 1.2  2006/02/20 16:15:50  endo
 * Typo
 *
 * Revision 1.1  2006/02/20 14:14:48  endo
 * RComAPI checked.
 *
 **********************************************************************/
