/**********************************************************************
 **                                                                  **
 **                            RComClient.cc                         **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This file contains functions for the RComClient class.          **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: RComClient.cc,v 1.1 2008-07-15 22:33:59 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <netdb.h>
#include <pthread.h>
#include <fstream>
#include <iostream>
#include <string>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/un.h>
#include <pthread.h>

#include "RComClient.h"
#include "RComServerProtocols.h"
#include "RComDefaultSetups.h"

using std::string;

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------
extern bool gRComDebug;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const struct timespec RComClient::READER_SLEEP_TIME_NSEC_ = {0, 1000000};
const string RComClient::EMPTY_STRING_ = "";
const int RComClient::INVALID_SOCKET_FD_ = -1;

//-----------------------------------------------------------------------
// Mutex initialization
//-----------------------------------------------------------------------
pthread_mutex_t RComClient::socketMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constructor for RComClient class.
//-----------------------------------------------------------------------
RComClient::RComClient(void) :
    serverHostMachine_(DEFAULT_RCOM_SERVER_HOST_MACHINE),
    shSocketFd_(INVALID_SOCKET_FD_),
    serverPort_(DEFAULT_RCOM_SERVER_PORT),
    myID_(-1),
    readerThreadIsUp_(false),
    extraCallbackFuncArg_(NULL),
    readerCallbackFunc_(NULL)
{
    // If you want non-default host machine name or port number,
    // specify them in here, but be careful.

    // Initialize the mutex (to be sure).
    pthread_mutex_init(&socketMutex_, NULL);
}

//-----------------------------------------------------------------------
// Distructor for RComClient class.
//-----------------------------------------------------------------------
RComClient::~RComClient(void)
{
}

//-----------------------------------------------------------------------
// This function initializes the RComClient.
//-----------------------------------------------------------------------
bool RComClient::initialize(
    int myID,
    void (*readerCallbackFunc)(int, const void *, size_t, bool, void *),
    void *extraCallbackFuncArg)
{
    const struct timespec RETRY_SLEEP_TIME_NSEC = {1, 0}; // 1 second sleep
    int connStatus;
    bool done = false;

    myID_ = myID;
    readerCallbackFunc_ = readerCallbackFunc;
    extraCallbackFuncArg_ = extraCallbackFuncArg;

    // Connect to server.
    fprintf(
        stdout,
        "RComClient: Connecting to the server on %s [port %d]...",
        serverHostMachine_.c_str(),
        serverPort_);
    fflush(stdout);

    while (!done)
    {
        connStatus = connectToServer_(serverHostMachine_, serverPort_);

        switch (connStatus) {

        case INIT_CONNECTION_STATUS_NOT_READY:
            fprintf(stdout, ".");
            fflush(stdout);
            nanosleep(&RETRY_SLEEP_TIME_NSEC, NULL);
            break;

        case INIT_CONNECTION_STATUS_SUCCESS:
            fprintf(stdout, "\nRComClient: Server connected.\n");
            fflush(stdout);
            done = true;
            break;

        default:
            fprintf(
                stderr,
                "\nError: RComClient::initialize(). Failed to connect to the server.\n");
            return false;
        }
    }

    // Send ID.
    sendCommandToServer_(RCOMSERVER_PROTOCOL_INIT);
    writeInteger_(shSocketFd_, myID);

    // Start the reader thread.
    pthread_create(&readerThread_, NULL, &startReaderThread_, (void *)this);

    return true;
}

//-----------------------------------------------------------------------
// This function writes an integer number.
//-----------------------------------------------------------------------
void RComClient::writeInteger_(const int fd, int value)
{
    write(fd, &value, sizeof(int));
}

//-----------------------------------------------------------------------
// This function sends a command to the server.
//-----------------------------------------------------------------------
void RComClient::sendCommandToServer_(const unsigned char cmd)
{
    unsigned char threeBytesBuf[3];

    threeBytesBuf[0] = RCOMSERVER_PROTOCOL_SYNC_BYTE1;
    threeBytesBuf[1] = RCOMSERVER_PROTOCOL_SYNC_BYTE2;
    threeBytesBuf[2] = cmd;

    write(shSocketFd_, (unsigned char *)&threeBytesBuf, 3);
}

//-----------------------------------------------------------------------
// This function will ship the data to server.
//-----------------------------------------------------------------------
bool RComClient::sendData(
        int recipientID,
        const void *data,
        size_t dataSize,
        bool enfoceReliable)
{
    // Check the validity of the data.
    if ((recipientID < 0) || (data == NULL))
    {
        return false;
    }

    pthread_mutex_lock(&socketMutex_);

    sendCommandToServer_(RCOMSERVER_PROTOCOL_DATA);
    writeInteger_(shSocketFd_, recipientID);
    writeInteger_(shSocketFd_, dataSize);
    write(shSocketFd_, (unsigned char *)data, dataSize);
    writeInteger_(shSocketFd_, enfoceReliable);

    pthread_mutex_unlock(&socketMutex_);

    return true;
}

//-----------------------------------------------------------------------
// This function connects to the server through a socket.
//-----------------------------------------------------------------------
int RComClient::connectToServer_(string hostMacine, int port)
{
    struct sockaddr_in sin;
    struct hostent *hent = NULL;
    int sockfd;
    int status = INIT_CONNECTION_STATUS_FAILURE;

    sin.sin_family = AF_INET;
    sin.sin_port = htons(port);

    if (hostMacine == EMPTY_STRING_)
    {
        fprintf(
            stderr,
            "\nError: RComClient::connectToServer_(). Host name is empty.\n");
        return INIT_CONNECTION_STATUS_FAILURE;
    }

    if (isdigit(hostMacine[0]))
    {
        sin.sin_addr.s_addr = inet_addr(hostMacine.c_str());
    }
    else
    {
        // Get the IP address.
        if((hent = gethostbyname(hostMacine.c_str())) == NULL)
        {
            fprintf(
                stderr,
                "\nError: RComClient::connectToServer_(). Unknown host: %s\n",
                hostMacine.c_str());
            return INIT_CONNECTION_STATUS_FAILURE;
        }

        memcpy((char*) &sin.sin_addr, hent->h_addr, sizeof(sin.sin_addr));
    }

    bzero(&(sin.sin_zero), 8);

    // Open the socket.
    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        fprintf(
            stderr,
            "\nError: RComClient::connectToServer_(). Socket could not be opened.\n");
        return INIT_CONNECTION_STATUS_FAILURE;
    }

    // Connect
    if (connect(sockfd, (sockaddr*)&sin, sizeof(sin)))
    {
        close(sockfd);

        switch(errno) {

        case ECONNREFUSED:
            status = INIT_CONNECTION_STATUS_NOT_READY;
            break;

        case ENETUNREACH:
            fprintf(
                stderr,
                "\nError: RComClient::connectToServer_(). Network unreachable from this host.");
            status = INIT_CONNECTION_STATUS_FAILURE;
            break;

        default:
            fprintf(
                stderr,
                "\nError: RComClient::connectToServer_(). Could not connect to socket.");
            status = INIT_CONNECTION_STATUS_FAILURE;
            break;
        }

        return status;
    }

    shSocketFd_ = sockfd;

    return INIT_CONNECTION_STATUS_SUCCESS;
}

//-----------------------------------------------------------------------
// This function starts up a thread that reads the server command.
//-----------------------------------------------------------------------
void *RComClient::startReaderThread_(void *data)
{
    RComClient *rcomClientInstance;

    rcomClientInstance = (RComClient *)data;
    
    // Start the main loop for the connection
    rcomClientInstance->readerMainLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to read commands from the server.
//-----------------------------------------------------------------------
void RComClient::readerMainLoop_(void)
{
    vector<ServerData_t> serverDataList;
    int i, numQueuedData = 0;

    readerThreadIsUp_ = true;

    while(true)
    {
        pthread_testcancel();

        pthread_mutex_lock(&socketMutex_);

        numQueuedData = pollServer_();

        // Read the arrived data
        for (i = 0; i < numQueuedData; i++)
        {
            serverDataList.push_back(readServerData_());
        }

        pthread_mutex_unlock(&socketMutex_);

        if (numQueuedData == 0)
        {
            pthread_testcancel();
            nanosleep(&READER_SLEEP_TIME_NSEC_, NULL);
        }
        else
        {
            // Process the server data.
            for (i = 0; i < numQueuedData; i++)
            {
                pthread_testcancel();
                processServerData_(&(serverDataList[i]));
            }

            serverDataList.clear();
        }
    }
}

//-----------------------------------------------------------------------
// This function read the server data.
//-----------------------------------------------------------------------
RComClient::ServerData_t RComClient::readServerData_(void)
{
    ServerData_t serverData;

    serverData.recipientID = readInteger_(shSocketFd_);
    serverData.dataSize = readInteger_(shSocketFd_);
 
    if (serverData.dataSize > 0)
    {
        serverData.data = new char[(serverData.dataSize)+1];
        readNBytes_(
            shSocketFd_,
            (unsigned char *)(serverData.data),
            serverData.dataSize);
        serverData.data[serverData.dataSize] = '\0';
    }

    serverData.enfoceReliable = (bool)(readInteger_(shSocketFd_));

    return serverData;
}

//-----------------------------------------------------------------------
// This function process the server data and delete it after done with it.
//-----------------------------------------------------------------------
void RComClient::processServerData_(ServerData_t *serverData)
{
    // Call the callback function.
    readerCallbackFunc_(
        serverData->recipientID,
        serverData->data,
        serverData->dataSize,
        serverData->recipientID,
        extraCallbackFuncArg_);

    // Delete the data.
    if ((serverData->data) != NULL)
    {
        delete [] serverData->data;
        serverData->data = NULL;
    }
    serverData->dataSize = 0;
}

//-----------------------------------------------------------------------
// This function reads an integer number.
//-----------------------------------------------------------------------
int RComClient::readInteger_(const int fd)
{
    int number;

    readNBytes_(fd, (unsigned char *)&number, sizeof(int));

    return number;
}

//-----------------------------------------------------------------------
// This function reads n bytes from the file descriptor.
//-----------------------------------------------------------------------
bool RComClient::readNBytes_(const int fd, unsigned char *buf, const int n)
{
    int msgLength = -1, index = 0, remaining = 0;

    remaining = n;

    while (remaining > 0)
    {
        msgLength = read(fd, &buf[index], remaining);

        if (msgLength == 0)
        {
            fprintf(
                stderr,
                "RComClient: WARNING! Server may have died. Reconnection invoked.\n");
            initialize(myID_, readerCallbackFunc_, extraCallbackFuncArg_);
            return false;
        }
        else if (msgLength <= 0)
        {
            fprintf(
                stderr,
                "RComClient: WARNING! Read error %d occured. Reconnection invoked.\n",
                errno);
            initialize(myID_, readerCallbackFunc_, extraCallbackFuncArg_);
            return false;
        }

        index += msgLength;
        remaining -= msgLength;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function checks to see if any incoming data is ready to be
// delivered.
//-----------------------------------------------------------------------
int RComClient::pollServer_(void)
{
    int numQueuedData;

    sendCommandToServer_(RCOMSERVER_PROTOCOL_POLL);
    numQueuedData = readInteger_(shSocketFd_);

    return numQueuedData;
}

/**********************************************************************
 * $Log: RComClient.cc,v $
 * Revision 1.1  2008-07-15 22:33:59  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.2  2006/08/09 21:59:55  endo
 * gDebug -> gRComDebug
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.6  2006/03/13 19:27:05  endo
 * gDebug added.
 *
 * Revision 1.5  2006/02/25 22:10:23  endo
 * The user can now specify ports to be used.
 *
 * Revision 1.4  2006/02/25 16:17:49  endo
 * rcomserver added
 *
 * Revision 1.3  2006/02/24 11:50:34  endo
 * RComm_Controller checked in.
 *
 * Revision 1.2  2006/02/22 07:17:22  endo
 * RComServer and RComClient implemented.
 *
 * Revision 1.1  2006/02/20 14:23:55  endo
 * RComAPI checked in.
 *
 **********************************************************************/
