/**********************************************************************
 **                                                                  **
 **                            RComServer.cc                         **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This file contains functions for the RComServer class.          **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: RComServer.cc,v 1.1 2008-07-15 22:34:00 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <netdb.h>
#include <fstream>
#include <iostream>
#include <string>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/un.h>
#include <pthread.h>

#include "RComServer.h"
#include "RComServerProtocols.h"
#include "RComDefaultSetups.h"
#include "RComAPIFactory.h"
#include "RComIDs.h"
#include "RComInterthread.h"

using std::string;

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------
extern bool gRComDebug;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const string RComServer::EMPTY_STRING_ = "";
const int RComServer::INVALID_SOCKET_FD_ = -1;
const int RComServer::SOCKET_BACKLOG_SIZE_ = 5;
const int RComServer::MAX_SYNC_ERROR_COUNT_ = 5;
const int RComServer::MAX_QUEUED_DATA = 10;

//-----------------------------------------------------------------------
// Mutex initialization
//-----------------------------------------------------------------------
pthread_mutex_t RComServer::privateConnectionInfoListMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t RComServer::queuedDataMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constructor for RComServer class.
//-----------------------------------------------------------------------
RComServer::RComServer(void) :
    rcomHandlerMain_(NULL),
    localPort_(DEFAULT_RCOM_SERVER_PORT),
    socketFd_(INVALID_SOCKET_FD_),
    myID_(-1),
    daemonThreadIsUp_(false),
    extraCallbackFuncArg_(NULL),
    readerCallbackFunc_(NULL)
{
    // If you want to a non-default port number, specify them in here,
    // but be careful.

    // Initialize the mutexes (to be sure).
    pthread_mutex_init(&privateConnectionInfoListMutex_, NULL);
    pthread_mutex_init(&queuedDataMutex_, NULL);

    // Create the main RCom handler to send data to peer.
    rcomHandlerMain_ = RComAPIFactory::createRComAPI(RCOMAPI_TYPE_RCOM_INTERTHREAD);

    if (rcomHandlerMain_ == NULL)
    {
        fprintf(
            stderr,
            "Error: RComServer::initClient_(). RComInterthread could not be created.\n");
        return;
    }
}

//-----------------------------------------------------------------------
// Distructor for RComServer class.
//-----------------------------------------------------------------------
RComServer::~RComServer(void)
{
    // Close all the private connections.
    disconnectAllPrivateConnections_();

    // Make sure to kill the deamon thread.
    if (daemonThreadIsUp_)
    {
        pthread_cancel(daemonThread_);
        pthread_join(daemonThread_, NULL );
        daemonThreadIsUp_ = false;
    }
}

//-----------------------------------------------------------------------
// This function disconnects all of the private connections.
//-----------------------------------------------------------------------
void RComServer::disconnectAllPrivateConnections_(void)
{
    vector<PrivateConnectionInfo_t> privateConnectionInfoList;
    int i, listSize;
    
    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    privateConnectionInfoList = shPrivateConnectionInfoList_;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    listSize = privateConnectionInfoList.size();

    for (i = 0; i < listSize; i++)
    {
        if (privateConnectionInfoList[i].connectionThreadIsUp)
        {
            pthread_cancel(privateConnectionInfoList[i].connectionThread);
            pthread_join(privateConnectionInfoList[i].connectionThread, NULL);

            privateConnectionInfoList[i].connectionThreadIsUp = false;
        }
    }

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    shPrivateConnectionInfoList_ = privateConnectionInfoList;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);
}

//-----------------------------------------------------------------------
// This function initializes the RComServer.
//-----------------------------------------------------------------------
bool RComServer::initialize(
    int myID,
    void (*readerCallbackFunc)(int, const void *, size_t, bool, void *),
    void *extraCallbackFuncArg)
{
    myID_ = myID;

    // Callback function to be called when some application wants to send
    // data to rcomserver itself.
    readerCallbackFunc_ = readerCallbackFunc;
    extraCallbackFuncArg_ = extraCallbackFuncArg;

    // Connect to server.
    fprintf(
        stdout,
        "RComServer: Starting the server on port %d...\n",
        localPort_);

    if (!(openSocket_(localPort_)))
    {
        fprintf(
            stderr,
            "Error: RComServer::initialize(). Failed to open the socket.\n");
        return false;
    }

    fprintf(stdout, "RComServer: Connected.\n");
    fflush(stdout);

    // Start the deamon thread.
    pthread_create(&daemonThread_, NULL, &startDaemonThread_, (void *)this );

    // Initialize the main RCom handler.
    if (!(rcomHandlerMain_->initialize(
              myID_,
              readerCallbackFunc_,
              extraCallbackFuncArg_)))
    {
        fprintf(
            stderr,
            "Error: RComServer::initialize(). RCom handler could not be initialized.\n");
        return false;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function opens a socket.
//-----------------------------------------------------------------------
bool RComServer::openSocket_(int port)
{
    struct sockaddr_in sin;
    int sockfd;
    unsigned int sockReuseOption;

   bzero((char *) &sin, sizeof(sin));
  
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = htonl(INADDR_ANY);
    sin.sin_port = htons(port);
  
    // Open the socket.
    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        fprintf(
            stderr,
            "Error: RComServer::connectToServer_(). Socket could not be opened.\n");
        return false;
    }

    // Allow multiple sickets to use the same port number.
    sockReuseOption = 1;
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &sockReuseOption, sizeof(unsigned int)) < 0)
    {
         fprintf(stderr, "Error: RComServer::connectToServer_(). Reusing ADDR failed.\n");
         return false;
    }

    // Bind
    if (bind(sockfd, (struct sockaddr *)&sin, sizeof(sin)) < 0)
    {
        fprintf(
            stderr,
            "Error: RComServer::connectToServer_(). Bind to local address failed.\n");
        return false;
    }

    listen(sockfd, 3);

    socketFd_ = sockfd;

    return true;
}

//-----------------------------------------------------------------------
// This function starts up a thread to run the daemon.
//-----------------------------------------------------------------------
void *RComServer::startDaemonThread_(void *rcomServerInstance)
{
    ((RComServer *)rcomServerInstance)->daemonMainLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This starts the daemon to listen to the incoming data.
//-----------------------------------------------------------------------
void RComServer::daemonMainLoop_(void)
{
    int count = 0;
    int fd = INVALID_SOCKET_FD_;

    daemonThreadIsUp_ = true;

    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Wait for a client connection at the socket.
        fd = waitForNewConnection_(count);

        // Make sure termination was not requested.
        pthread_testcancel();

        // Start another thread specific to this connection.
        startPrivateConnection_(fd);

        count++;
    }
}

//-----------------------------------------------------------------------
// This function listens the socket for a client connection. After the
// communication is established and it returns a new file descriptor to
// talks to the cilent privately.
//-----------------------------------------------------------------------
int RComServer::waitForNewConnection_(int count)
{
    int newFd = INVALID_SOCKET_FD_, status = -1;
    sockaddr client;
    socklen_t len;

    // Listen the socket for connection.
    status = listen(socketFd_, SOCKET_BACKLOG_SIZE_);

    if (status < 0)
    {
        fprintf(
            stderr,
            "Error: RComServer::waitForNewConnection_(). Listening socket failed.\n");
        return INVALID_SOCKET_FD_;
    }

    fprintf(
        stdout,
        "RComServer: Waiting for %s client connection...\n",
        (count > 0)? "another" : "the first");
    fflush(stdout);

    // Wait for the connection with client.
    len = sizeof(sockaddr);
    newFd = accept(socketFd_, &client, &len);

    if (newFd < 0)
    {
        fprintf(
            stderr,
            "Error: RComServer::waitForNewConnection_(). Connection to the client failed [%d].\n", newFd);
        return INVALID_SOCKET_FD_;
    }

    return newFd;
}

//-----------------------------------------------------------------------
// This function starts another thread to talk only to this connection.
//-----------------------------------------------------------------------
void RComServer::startPrivateConnection_(const int fd)
{
    PrivateConnectionThreadInput_t *data = NULL;
    PrivateConnectionInfo_t connectionInfo;
    pthread_t newThread;

    // First, add this fd to the connection info list.
    connectionInfo.fd = fd;
    connectionInfo.clientID = RCOM_USER_ID_INVALID;
    connectionInfo.connectionThreadIsUp = false;
    connectionInfo.rcomHandler = NULL;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    shPrivateConnectionInfoList_.push_back(connectionInfo);
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    // Set up the input for the thread.
    data = new PrivateConnectionThreadInput_t;
    data->fd = fd;
    data->rcomServerInstance = this;

    // Start the new thread to talk to this connection.
    pthread_create(&newThread, NULL, &startPrivateConnectionThread_, (void *)data);

    saveConnectionThread_(fd, newThread);
}

//-----------------------------------------------------------------------
// This function updates the status of the thread in the info list.
//-----------------------------------------------------------------------
void RComServer::saveConnectionThread_(const int fd, const pthread_t connectionThread)
{
    vector<PrivateConnectionInfo_t> privateConnectionInfoList;
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    privateConnectionInfoList = shPrivateConnectionInfoList_;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    listSize = privateConnectionInfoList.size();

    for (i = 0; i < listSize; i++)
    {
        pthread_testcancel();

        if (privateConnectionInfoList[i].fd == fd)
        {
            privateConnectionInfoList[i].connectionThread = connectionThread;
            break;
        }
    }

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    shPrivateConnectionInfoList_ = privateConnectionInfoList;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);
}

//-----------------------------------------------------------------------
// This function starts up a thread to communicate with this connection
// privately.
//-----------------------------------------------------------------------
void *RComServer::startPrivateConnectionThread_(void *input)
{
    PrivateConnectionThreadInput_t *data = NULL;
    int fd;
    RComServer *rcomServerInstance;

    // Unpack the data
    data = (PrivateConnectionThreadInput_t *)input;
    fd = data->fd;
    rcomServerInstance = data->rcomServerInstance;
    
    // Start the main loop for the connection
    rcomServerInstance->privateConnectionMainLoop_(fd);

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to read commands from the client.
//-----------------------------------------------------------------------
void RComServer::privateConnectionMainLoop_(const int fd)
{
    setConnectionThreadIsUp_(fd, true);
    
    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Read the arrived data, and process it based on the instruction.
        readAndProcessPrivateConnectionData_(fd);
    }
}

//-----------------------------------------------------------------------
// This function reads the data sent by the client, and process it
// according to the instruction.
//-----------------------------------------------------------------------
bool RComServer::readAndProcessPrivateConnectionData_(const int fd)
{
    unsigned char oneByteBuf[1];
    int clientID;
    bool synced = false, received = false;
    static long int debugReadAndProcessPrivateConnectionDataCounter = 0;
    const bool DEBUG_READ_AND_PROCESS_PRIVATE_CONNECTION_DATA = false;

    // Make sure that it is synced to the client.
    synced = syncToClient_(fd);

    if (!synced)
    {
        // Not synced. Something went wrong. Disconnect it.
        disconnectPrivateConnection_(fd);
        return false;
    }

    // Now it is synced. Get a command byte.
    received = readNBytes_(fd, (unsigned char *)&oneByteBuf, 1);

    if (DEBUG_READ_AND_PROCESS_PRIVATE_CONNECTION_DATA)
    {
        fprintf(
            stderr,
            "[%ld] Byte(%x) receieved.\n",
            debugReadAndProcessPrivateConnectionDataCounter++,
            oneByteBuf[0]);
    }

    switch (oneByteBuf[0]) {

    case RCOMSERVER_PROTOCOL_INIT:
        clientID = readInteger_(fd);
        initClient_(fd, clientID);
        fprintf(
            stdout,
            "RComServer: Connection to client [ID %d] established.\n",
            clientID);
        fflush(stdout);
        break;

    case RCOMSERVER_PROTOCOL_EXIT:
        disconnectPrivateConnection_(fd);
        break;

    case RCOMSERVER_PROTOCOL_POLL:
        sendAllQueuedData_(fd);
        break;

    case RCOMSERVER_PROTOCOL_DATA:
        readAndProcessClientData_(fd);
        break;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function process the client data and delete it after done with it.
//-----------------------------------------------------------------------
void RComServer::readAndProcessClientData_(const int fd)
{
    int recipientID;
    char *data = NULL;
    int dataSize;
    bool enfoceReliable = false;
    bool connected = false;

    recipientID = readInteger_(fd);
    dataSize = readInteger_(fd);
 
    if (dataSize > 0)
    {
        data = new char[(dataSize)+1];
        readNBytes_(fd, (unsigned char *)data, dataSize);
        data[dataSize] = '\0';
    }

    enfoceReliable = (bool)(readInteger_(fd));

    if (recipientID == myID_)
    {
        // Data is send to RCom itself. Call the callback function.
        readerCallbackFunc_(
            recipientID,
            data,
            dataSize,
            enfoceReliable,
            extraCallbackFuncArg_);
    }
    else if (isClient_(recipientID, &connected))
    {
        if (connected)
        {
            // Send the data to client.
            sendData(recipientID, data, dataSize, enfoceReliable);
        }
    }
    else
    {
        // Send the data to RCom.
        rcomHandlerMain_->sendData(recipientID, data, dataSize, enfoceReliable);
    }

    // Delete the data.
    if (data != NULL)
    {
        delete [] data;
        data = NULL;
    }
}

//-----------------------------------------------------------------------
// This function sends the queued data
//-----------------------------------------------------------------------
void RComServer::sendAllQueuedData_(int fd)
{
    int clientID;
    vector<QueuedData_t> thisQueuedDataList, nextQueuedDataList;
    int i, numQueuedData;
    const bool DEBUG_SEND_ALL_QUEUED_DATA = false;

    clientID = getClientID_(fd);
    numQueuedData = 0;

    pthread_mutex_lock(&queuedDataMutex_);

    for (i = 0; i < (int)(shQueuedDataList_.size()); i++)
    {
        if (shQueuedDataList_[i].recipientID == clientID)
        {
            thisQueuedDataList.push_back(shQueuedDataList_[i]);

	    if ((int)(thisQueuedDataList.size()) > MAX_QUEUED_DATA)
	    {
	        thisQueuedDataList.erase(thisQueuedDataList.begin());
	    }
        }
        else
        {
            nextQueuedDataList.push_back(shQueuedDataList_[i]);
        }
    }

    shQueuedDataList_ = nextQueuedDataList;
        
    pthread_mutex_unlock(&queuedDataMutex_);

    pthread_testcancel();

    numQueuedData = thisQueuedDataList.size();

    if (DEBUG_SEND_ALL_QUEUED_DATA)
    {
        if (numQueuedData > 0)
	{
            fprintf(stderr, "Num Queued Data = %d\n", numQueuedData);
	}
    }

    writeInteger_(fd, numQueuedData);

    for (i = 0; i < numQueuedData; i++)
    {
        sendQueuedData_(fd, thisQueuedDataList[i]);
    }
}

//-----------------------------------------------------------------------
// This function writes an integer number.
//-----------------------------------------------------------------------
void RComServer::writeInteger_(const int fd, int value)
{
    write(fd, &value, sizeof(int));
}

//-----------------------------------------------------------------------
// This function sends the queued data
//-----------------------------------------------------------------------
void RComServer::sendQueuedData_(int fd, QueuedData_t queuedData)
{
    writeInteger_(fd, queuedData.recipientID);
    writeInteger_(fd, queuedData.dataSize);

    if (queuedData.dataSize > 0)
    {
        write(fd, queuedData.data, queuedData.dataSize);
        delete [] queuedData.data;
        queuedData.data = NULL;
    }

    writeInteger_(fd, queuedData.enfoceReliable);
}

//-----------------------------------------------------------------------
// This function updates the client list.
//-----------------------------------------------------------------------
bool RComServer::initClient_(const int fd, const int clientID)
{
    RComAPI *rcomHandler = NULL;
    vector<PrivateConnectionInfo_t> privateConnectionInfoList;
    int i, listSize;

    // Create the RCom handler for this client.
    rcomHandler = RComAPIFactory::createRComAPI(RCOMAPI_TYPE_RCOM_INTERTHREAD);

    if (rcomHandler == NULL)
    {
        fprintf(
            stderr,
            "Error: RComServer::initClient_(). RComInterthread could not be created.\n");
        return false;
    }

    // Initialize the RCom handler.
    if (!(rcomHandler->initialize(
              clientID,
              cbProcessRecievedDataFromRComPeer_,
              this)))
    {
        fprintf(
            stderr,
            "Error: RComServer::initClient_(). RComInterthread could not be initialized.\n");
        return false;
    }

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    privateConnectionInfoList = shPrivateConnectionInfoList_;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    listSize = privateConnectionInfoList.size();

    for (i = 0; i < listSize; i++)
    {
        pthread_testcancel();

        if (privateConnectionInfoList[i].fd == fd)
        {
            privateConnectionInfoList[i].clientID = clientID;
            privateConnectionInfoList[i].rcomHandler = rcomHandler;
            break;
        }
    }

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    shPrivateConnectionInfoList_ = privateConnectionInfoList;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    return true;
}

//-----------------------------------------------------------------------
// The callback function that will be called when a new data is arrived
// from the peer RCom.
//-----------------------------------------------------------------------
void RComServer::cbProcessRecievedDataFromRComPeer_(
    int recipientID,
    const void *data,
    size_t dataSize,
    bool reliableEnforced,
    void *extraArg)
{
    RComServer *rcomServerInstance;

    // Pointer to RComServer instance passed via initialize()'s last
    // argument (i.e., extraCallbackFuncArg).
    rcomServerInstance = (RComServer *)extraArg;

    if (recipientID == (rcomServerInstance->myID_))
    {
        // Data is send to RCom itself. Call the callback function.
        rcomServerInstance->readerCallbackFunc_(
            recipientID,
            data,
            dataSize,
            reliableEnforced,
            rcomServerInstance->extraCallbackFuncArg_);
    }
    else
    {
        // Send the data to the client right away.
        rcomServerInstance->sendData(
            recipientID,
            data,
            dataSize,
            reliableEnforced);
    }

    if (gRComDebug)
    {
        fprintf(stderr, "RComServer::cbProcessRecievedDataFromRComPeer_():\n");
        fprintf(stderr, " Data received from peer RCom.\n");
        fprintf(stderr, " Recipient ID: %d\n", recipientID);
        fprintf(stderr, " Data Size: %d\n", dataSize);
        fprintf(stderr, " Enforced Reliability: %d\n", reliableEnforced);
        fprintf(stderr, " Extra Arg from initialize(): %s\n", (extraArg)? "Specified" : "None");
        fprintf(stderr, " Data: %s\n", (char *)data);
    }
}

//-----------------------------------------------------------------------
// This function reads an integer number.
//-----------------------------------------------------------------------
int RComServer::readInteger_(const int fd)
{
    int number;

    readNBytes_(fd, (unsigned char *)&number, sizeof(int));

    return number;
}

//-----------------------------------------------------------------------
// This function reads n bytes from the file descriptor.
//-----------------------------------------------------------------------
bool RComServer::readNBytes_(const int fd, unsigned char *buf, const int n)
{
    int msgLength = -1, index = 0, remaining = 0;

    remaining = n;

    while (remaining > 0)
    {
        msgLength = read(fd, &buf[index], remaining);

        if (msgLength == 0)
        {
            disconnectPrivateConnection_(fd);
        }
        else if (msgLength <= 0)
        {
            fprintf(
                stderr,
                "WARNING: RComServer::readNBytes_(). Read error %d.\n", errno);

            disconnectPrivateConnection_(fd);
        }

        index += msgLength;
        remaining -= msgLength;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function disconnects the private connection.
//-----------------------------------------------------------------------
void RComServer::disconnectPrivateConnection_(const int fd)
{
    pthread_t connectionThread;
    int clientID;
    char buf[1024];

    // First, report the disconnection to the user.
    clientID = getClientID_(fd);
    if (clientID > RCOM_USER_ID_INVALID)
    {
        sprintf(buf, "%d", clientID);
    }
    else
    {
        sprintf(buf, "<Unknown>");
    }

    fprintf(
        stderr,
        "RComServer: Client (ID %s) disconnected.\n",
        buf);

    // Update the info list.
    setConnectionThreadIsUp_(fd, false);
    removeRComHandler_(fd);

    // Finally, quit the thread.
    connectionThread = getConnectionThread_(fd);
    pthread_cancel(connectionThread);
    pthread_join(connectionThread, NULL);
}

//-----------------------------------------------------------------------
// This function gets the connectionThread stored in the info list.
//-----------------------------------------------------------------------
pthread_t RComServer::getConnectionThread_(const int fd)
{
    vector<PrivateConnectionInfo_t> privateConnectionInfoList;
    pthread_t connectionThread;
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    privateConnectionInfoList = shPrivateConnectionInfoList_;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    listSize = privateConnectionInfoList.size();

    for (i = 0; i < listSize; i++)
    {
        pthread_testcancel();

        if (privateConnectionInfoList[i].fd == fd)
        {
            connectionThread = privateConnectionInfoList[i].connectionThread;
            break;
        }
    }

    return connectionThread;
}

//-----------------------------------------------------------------------
// This function gets the connectionThread stored in the info list.
//-----------------------------------------------------------------------
int RComServer::getClientID_(const int fd)
{
    vector<PrivateConnectionInfo_t> privateConnectionInfoList;
    int clientID;
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    privateConnectionInfoList = shPrivateConnectionInfoList_;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    listSize = privateConnectionInfoList.size();

    for (i = 0; i < listSize; i++)
    {
        pthread_testcancel();

        if (privateConnectionInfoList[i].fd == fd)
        {
            clientID = privateConnectionInfoList[i].clientID;
            break;
        }
    }

    return clientID;
}

//-----------------------------------------------------------------------
// This function makes sure that the server and the client are syncronized.
// Two keys are used to check the syncronization.
//-----------------------------------------------------------------------
bool RComServer::syncToClient_(const int fd)
{
    int errorCount = 0;
    unsigned char oneByteBuf[1];
    bool synced = false, received = false;


    while (!synced)
    {
        received = readNBytes_(fd, (unsigned char *)&oneByteBuf, 1);

        if (received)
        {
            if (oneByteBuf[0] == RCOMSERVER_PROTOCOL_SYNC_BYTE1)
            {
                // The first sync byte is OK. Check the second one.
                received = readNBytes_(fd, (unsigned char *)&oneByteBuf, 1);

                if (received)
                {
                    if (oneByteBuf[0] == RCOMSERVER_PROTOCOL_SYNC_BYTE2)
                    {
                        // The second sync byte matched as well. Assuming
                        // this is a right data.
                        synced = true;
                        break;
                    }
                    else
                    {
                        // It seems, it was not the header. Continue to read.
                        continue;
                    }
                }
                else
                {
                    // Read error occured.
                    errorCount++;
                }
            }
            else
            {
                // It seems, it was not the header. Continue to read.
                continue;
            }
        }
        else
        {
            // Read error occured.
            errorCount++;
        }

        // Check if the errors were occured too often.
        if (errorCount > MAX_SYNC_ERROR_COUNT_)
        {
            fprintf(
                stderr,
                "WARNING: RComServer::syncToClient_(). Max Sync count exceeded.\n");
            break;
        }
    }

    return synced;
}

//-----------------------------------------------------------------------
// This function updates the status of the thread in the info list.
//-----------------------------------------------------------------------
void RComServer::setConnectionThreadIsUp_(const int fd, const bool isUp)
{
    vector<PrivateConnectionInfo_t> privateConnectionInfoList;
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    privateConnectionInfoList = shPrivateConnectionInfoList_;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    listSize = privateConnectionInfoList.size();

    for (i = 0; i < listSize; i++)
    {
        pthread_testcancel();

        if (privateConnectionInfoList[i].fd == fd)
        {
            privateConnectionInfoList[i].connectionThreadIsUp = isUp;
            break;
        }
    }

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    shPrivateConnectionInfoList_ = privateConnectionInfoList;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);
}

//-----------------------------------------------------------------------
// This function removes the RCom Handler for this priviate connection.
//-----------------------------------------------------------------------
void RComServer::removeRComHandler_(const int fd)
{
    vector<PrivateConnectionInfo_t> privateConnectionInfoList;
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    privateConnectionInfoList = shPrivateConnectionInfoList_;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    listSize = privateConnectionInfoList.size();

    for (i = 0; i < listSize; i++)
    {
        pthread_testcancel();

        if (privateConnectionInfoList[i].fd == fd)
        {
            if (privateConnectionInfoList[i].rcomHandler != NULL)
            {
                delete privateConnectionInfoList[i].rcomHandler;
                privateConnectionInfoList[i].rcomHandler = NULL;
            }
            break;
        }
    }

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    shPrivateConnectionInfoList_ = privateConnectionInfoList;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);
}

//-----------------------------------------------------------------------
// This function will ship the data to client. It is done by saving the
// data in a queue. When next time the client polls, it will be delivered.
//-----------------------------------------------------------------------
bool RComServer::sendData(
    int recipientID,
    const void *data,
    size_t dataSize,
    bool enfoceReliable)
{
    QueuedData_t queuedData;

    queuedData.recipientID = recipientID;
    queuedData.dataSize = dataSize;
    queuedData.data = new char[queuedData.dataSize];
    memcpy(queuedData.data, data, queuedData.dataSize);

    pthread_mutex_lock(&queuedDataMutex_);
    shQueuedDataList_.push_back(queuedData);
    pthread_mutex_unlock(&queuedDataMutex_);

    return true;
}

//-----------------------------------------------------------------------
// This function checks to see specified ID is connected is the client.
// It also checks to see if the connection is up.
//-----------------------------------------------------------------------
bool RComServer::isClient_(int id, bool *connected)
{
    vector<PrivateConnectionInfo_t> privateConnectionInfoList;
    int i, listSize;
    bool isClient = false;

    *connected = false;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    privateConnectionInfoList = shPrivateConnectionInfoList_;
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    listSize = privateConnectionInfoList.size();

    for (i = 0; i < listSize; i++)
    {
        if (privateConnectionInfoList[i].clientID == id)
        {
            isClient = true;

            if (privateConnectionInfoList[i].connectionThreadIsUp)
            {
                *connected = true;
                break;
            }
            else
            {
                // Client may have been disconnected. Continue to
                // look for connected one.
            }
        }
    }

    return isClient;
}

/**********************************************************************
 * $Log: RComServer.cc,v $
 * Revision 1.1  2008-07-15 22:34:00  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.2  2006/08/09 21:59:55  endo
 * gDebug -> gRComDebug
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.11  2006/04/14 05:17:29  endo
 * Modified to allow multiple sickets to use the same port number.
 *
 * Revision 1.10  2006/04/05 22:09:54  endo
 * Max queued data implemented.
 *
 * Revision 1.9  2006/04/02 23:24:09  endo
 * Minor clean up.
 *
 * Revision 1.8  2006/04/02 22:46:29  endo
 * Unregister_App() added to Reliable_Comm.
 *
 * Revision 1.7  2006/02/27 20:41:54  endo
 * rcomserver bug fixed.
 *
 * Revision 1.6  2006/02/25 22:10:23  endo
 * The user can now specify ports to be used.
 *
 * Revision 1.5  2006/02/25 16:30:59  endo
 * *** empty log message ***
 *
 * Revision 1.4  2006/02/25 16:26:00  endo
 * *** empty log message ***
 *
 * Revision 1.3  2006/02/25 16:17:49  endo
 * rcomserver added
 *
 * Revision 1.2  2006/02/24 11:50:34  endo
 * RComm_Controller checked in.
 *
 * Revision 1.1  2006/02/22 07:17:22  endo
 * RComServer and RComClient implemented.
 *
 **********************************************************************/
