/**********************************************************************
 **                                                                  **
 **                        Reliable_Comm.cpp                         **
 **                                                                  **
 **  Written by:  I.C. Envarli                                       **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef __Reliable_Comm_cpp
#define __Reliable_Comm_cpp

#include "RCom_Debug_Settings.h"
#include "Reliable_Comm.h"
#include <unistd.h>
#include <time.h>
#include <pthread.h>
#include <sys/time.h>//for secs()
#include <stdlib.h>


//#define USE_WTCP
 
//Threaded global-scope function that is used to call an application-layer callback when a message is received
void * Callback_Thread(void * ptr){
#ifdef DEBUG_RELCOMM_CALLBACK
   fprintf(stderr,"Beginning callback thread\n");
#endif
   callback_thread_info_t * temp = (callback_thread_info_t*) ptr;
   if(pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL)!=0) fprintf(stderr,"Error: Making the callback thread asynchronous failed\n");
   //boolean for letting Reliable_Comm class know that the callback function has terminated
   bool * local_pointer =temp->done_reading_data;
   while(1){
     if(*(local_pointer)==false){
       //calling the appropriate callback function with the appropriate parameters
       (temp->funcptr)(temp->app_info.id, temp->data_length, temp->reliable, temp->app_info.instance);
       //let Reliable_Comm know that this thread is (almost) done
       *(local_pointer)=true;
     }
     else usleep(1000);
   }

#ifdef DEBUG_RELCOMM_CALLBACK
     fprintf(stderr,"Ending callback thread\n");
#endif
   
   //return NULL;
   
}


//Constructor that initializes all the mutex variables and the statistics
Reliable_Comm::Reliable_Comm(){
  for (int i=0;i<RCOM_NUM_PATHS;i++){
    sendcount[i]=recvcount[i]=ackcount[i]=0;
    pthread_mutex_init(&recv_count_mutex[i],NULL);
    pthread_mutex_init(&send_count_mutex[i],NULL);
    pthread_mutex_init(&send_mutex[i],NULL);
    pthread_mutex_init(&fragment_mutex[i],NULL);
    pthread_mutex_init(&wtcp_mutex[i],NULL);
  }
  pthread_mutex_init(&reliable_send_mutex,NULL);
  pthread_mutex_init(&rtt_mutex,NULL);
  pthread_mutex_init(&register_mutex,NULL);
  pthread_mutex_init(&thruput_mutex,NULL);
  application_count=0;
  reliable_acked=false;
  
  for (int i=0;i<RCOM_NUM_PATHS;i++){
    rtt[i]=0;
    rttvar[i]=3;
    rto[i]=3;
    last_confirmed_timestamp[i]=last_outgoing_timestamp[i]=secs();

    //wtcp stuff
    loss_observed[i]=0;
    loss_expected[i]=0.05;
    loss_observed_reset[i]=last_received[i]=secs();
    dropped_packets_wtcp[i]=last_seq_loss_reset[i]=0;
    last_action_was_speedup[i]=false;
    confirmed_thruput[i]=0;
    outgoing_thruput[i]=0;
    cf_tput_queue_size[i]=og_tput_queue_size[i]=0;
    receive_offset[i]=0;
    inter_packet_time[i]=1500;//in microseconds
    if(i==2)inter_packet_time[i]*=2;//make the 900mhz slower
    missing_fragment_flag[i]=false;
    //wtcp stuff
    alpha_plus[i]=1.8;
    beta_plus[i]=2.1;
    alpha_minus[i]=1.5;
    beta_minus[i]=1.3;
    lavg_ratio[i]=1.6;
    savg_ratio[i]=1.6;
    sigma[i]=20;
    current_wtcp_phase[i]=WTCP_PHASE_MAINTAIN;
    highest_ack_seq_num_wtcp[i]=0;
    expected_msg_id[i]=0;
  }
  for (int i=0;i<RCOM_MAX_APPS;i++){
    data_read_by_app[i]=true;
  }
  h=0.25; g=0.125;//these numbers are used in RTT/RTO calculations

  T_loss=3.0;
  
}

void Reliable_Comm::Modify_Inter_Packet_Time(int path, int change_speed_usec){
  if(path>-1&&path<RCOM_NUM_PATHS&&(-change_speed_usec)<(int)inter_packet_time[path]){
    inter_packet_time[path]+=change_speed_usec;
  }
}
void Reliable_Comm::Set_Inter_Packet_Time(int path, int new_speed_usec){
  if(path>-1&&path<RCOM_NUM_PATHS&&new_speed_usec>0){
    inter_packet_time[path]=new_speed_usec;
  }
}

const char * Reliable_Comm::Get_Stats(){
  strcpy(stats_string,"");
  for(int i=0;i<RCOM_NUM_PATHS;i++){
  sprintf(temp_string,
	  "Link %d: Roundtrip= %f, Timeout= %f, RTTvariance= %f\nACKed Thruput=%lu bytes/s, Attempted Thruput=%lu bytes/s\nInter_packet time = %lu, Link Up: %s\n"
	  ,i,rtt[i],rto[i],rttvar[i],confirmed_thruput[i],outgoing_thruput[i],inter_packet_time[i],
	  verified_heartbeat[i]? "Yes":"No");
  strcat (stats_string,temp_string);
  sprintf(temp_string,"Received Packets (incl. acks): %d, Received Acks: %d, Sent packets(excl. acks): %d\n",recvcount[i],ackcount[i],sendcount[i]);
  strcat (stats_string,temp_string);
	  
  }
  return stats_string;
 }

//wtcpupdate
unsigned int Reliable_Comm::process_packet_wtcp(int path,unsigned int seq_no,unsigned int advertised_rate){
  unsigned int original_expected_msg_id=expected_msg_id[path];
   unsigned int result=advertised_rate;
   pthread_mutex_lock(&wtcp_mutex[path]);
   unsigned int missing_packets=0;
   double curr=secs();
   double delta_t=curr-last_received[path];
   if(delta_t>1)delta_t=1;
   double ratio=delta_t/((double)(advertised_rate)/1000000.0);
   if(expected_msg_id[path]<5){
     alpha_plus[path]=(unsigned int)((double)ratio*1.2);
     alpha_minus[path]=(unsigned int)((double)ratio*0.8);
     beta_plus[path]=(unsigned int)((double)ratio*1.1);
     beta_minus[path]=(unsigned int)((double)ratio*0.9);
   }
   else if(expected_msg_id[path]<200){
     alpha_plus[path] =(unsigned int)(0.9*(double) alpha_plus[path]+0.1*((double)ratio*1.2));
     alpha_minus[path]=(unsigned int)(0.9*(double)alpha_minus[path]+0.1*((double)ratio*0.8));
     beta_plus[path]  =(unsigned int)(0.9*(double)  beta_plus[path]+0.1*((double)ratio*1.1));
     beta_minus[path] =(unsigned int)(0.9*(double) beta_minus[path]+0.1*((double)ratio*0.9));
   }
   if(expected_msg_id==0)last_suggested_ipt[path]=advertised_rate;
   if (seq_no<expected_msg_id[path]){
     last_seq_loss_reset[path]=expected_msg_id[path]=seq_no;//synchronize receiver if sender was restarted
     ratio=lavg_ratio[path];
     original_expected_msg_id=0;//to avoid having weird ratio
   }
   else if (expected_msg_id[path]==0 && seq_no>0){
     last_seq_loss_reset[path]=expected_msg_id[path]=seq_no;//synchronize receiver if sender has been running for a while
     ratio=lavg_ratio[path];
   }
   missing_packets=seq_no-expected_msg_id[path];
 


   if(secs()-loss_observed_reset[path]>T_loss)
     {
       loss_observed[path]=0.0;
       loss_observed_reset[path]=secs();
       dropped_packets_wtcp[path]=0;
       last_seq_loss_reset[path]=expected_msg_id[path];
     }

   dropped_packets_wtcp[path]+=missing_packets;

   if(missing_packets){
     loss_observed[path]=1-(double)dropped_packets_wtcp[path]/(double)(seq_no-last_seq_loss_reset[path]+1);
     if(current_wtcp_phase[path]==WTCP_PHASE_MAINTAIN){
       loss_expected[path]=0.87*loss_expected[path]+0.13*loss_observed[path];
     }
     if(loss_observed[path]>loss_expected[path]){
       result=(unsigned int)((double)result*1.05);
       sigma[path]=1;
       current_wtcp_phase[path]=WTCP_PHASE_DECREASE;
     }
   }
   else{//if not missing packets in between
     if(original_expected_msg_id!=0){//avoid having weird ratio values for first packets
       lavg_ratio[path]=0.95*lavg_ratio[path]+0.05*ratio;
       savg_ratio[path]=0.7*savg_ratio[path]+0.3*ratio;
     }
     if(advertised_rate==last_suggested_ipt[path]){
       if(lavg_ratio[path]>=alpha_plus[path] && savg_ratio[path]>=beta_plus[path]){//increase
	 if(result>50)result-=50;
	 sigma[path]=1;
	 current_wtcp_phase[path]=WTCP_PHASE_INCREASE;
       }
       else if(lavg_ratio[path]<alpha_minus[path] || savg_ratio[path]<beta_minus[path]){
	 result=decrease_sending_rate_wtcp(path,advertised_rate);
       }
       else{//Maintain
	 sigma[path]=1;
	 current_wtcp_phase[path]=WTCP_PHASE_MAINTAIN;
       }
     }
     else if(last_action_was_speedup){
       if(lavg_ratio[path]<alpha_minus[path]&&savg_ratio[path]<beta_minus[path])
	 result=decrease_sending_rate_wtcp(path,advertised_rate);
     }
   }

#ifdef DEBUG_WTCP
   fprintf(stderr,"Path %d, missing packets %u,[lng ratio %f] [shrt ratio %f] [obs loss %f] [expct loss] %f [adv rate %u] [result %u]\n",path,missing_packets,lavg_ratio[path],savg_ratio[path],loss_observed[path],loss_expected[path],advertised_rate, result);
   fprintf(stderr,"ratio %f, alpha_plus %u,alpha_minus %u, betaplus %u, betaminus%f\n",ratio,alpha_plus[path],alpha_minus[path],beta_plus[path],beta_minus[path]);
#endif
   expected_msg_id[path]=seq_no+1;
   last_received[path]=curr;
   pthread_mutex_unlock(&wtcp_mutex[path]);
   if(result==0)result=50;
   if(result>5000)result=5000;
   last_suggested_ipt[path]=result;
   if(result<advertised_rate)last_action_was_speedup[path]=true;
   else last_action_was_speedup[path]=false;
   return result;
}

unsigned int Reliable_Comm::decrease_sending_rate_wtcp(int path,unsigned int adv_rate){
  //inter_packet_time[path]=inter_packet_time[path]*(100-sigma[path])/100;
  unsigned int result=adv_rate*(100+sigma[path])/100;
  sigma[path]=sigma[path]<<1;//multiply by 2
  if(sigma[path]>5)sigma[path]=5;
  current_wtcp_phase[path]=WTCP_PHASE_DECREASE;
  return result;
}

void Reliable_Comm::Print_Stats(){
  fprintf(stderr,"\n");
  for(int i=0;i<RCOM_NUM_PATHS;i++){
    fprintf(stderr,"Link %d: ",i);
    fprintf(stderr,"Roundtrip= %f, Timeout= %f, RTTvariance= %f\n",rtt[i],rto[i],rttvar[i]);
    fprintf(stderr,"ACKed Thruput=%lu bytes/s, Attempted Thruput=%lu bytes/s\n",confirmed_thruput[i],outgoing_thruput[i]);
    fprintf(stderr,"Inter_packet time = %lu, Link Up: %s\n",inter_packet_time[i],verified_heartbeat[i]? "Yes":"No");
    fprintf(stderr,"Received Packets (incl. acks): %d, Received Acks: %d, Sent packets(excl. acks): %d\n",recvcount[i],ackcount[i],sendcount[i]);
  }

}

//The function that looks at each packet header and decides what to do; the brain of the class.
void Reliable_Comm::Interpret_Packet_Header(byte packet[],unsigned int packetlength,unsigned int pathnumber){
	pthread_mutex_lock(&recv_count_mutex[pathnumber]);
	recvcount[pathnumber]++;
	pthread_mutex_unlock(&recv_count_mutex[pathnumber]);
	r_comm_header_t * pkt_header = (r_comm_header_t *) packet;//locate the header
	byte * payload = packet+sizeof(r_comm_header_t);//locate the payload
	//identify the type of the packet
#ifdef DEBUG_RELCOMM_MSGID
	//if(pkt_header->msg_type!=HEARTBEAT && pkt_header->msg_type!=HEARTBEATACK)
	if(pkt_header->msg_type==HEARTBEATACK||pkt_header->msg_type==RELIABLEACK||pkt_header->msg_type==UNRELIABLEACK)
	  fprintf (stderr,"Ack type: %d, Msg ID: %lu\n",pkt_header->msg_type,pkt_header->msg_id);
	else	  fprintf (stderr,"Msg type: %d, Msg ID: %lu\n",pkt_header->msg_type,pkt_header->msg_id);
#endif //DEBUG
	if(pkt_header->msg_type==UNRELIABLEACK){
#ifdef DEBUG_RTT_READINGS
		fprintf(stdout, "RTT for unreliableack %f\n", secs()-pkt_header->tstamp_sent);
#endif
		verified_heartbeat[pathnumber]=temp_verified_heartbeat[pathnumber]=true;//the heartbeat for this particular route was ACKed.
		//using the terrible hack: app_id contains the size of the original packet)
#ifdef USE_WTCP
		if(pkt_header->inter_packet_time!=inter_packet_time[pathnumber]){

#ifdef DEBUG_WTCP
		  fprintf(stdout,"Changing inter packet time from %u to %u\n",inter_packet_time[pathnumber],pkt_header->inter_packet_time);
#endif
		  inter_packet_time[pathnumber]=pkt_header->inter_packet_time;
		}
		
#endif //USE_WTCP
		UpdateRTT(pathnumber,secs()-pkt_header->tstamp_sent);
		UpdateThruput(pkt_header->app_id,pkt_header->tstamp_sent,true,pathnumber);

	}
	else if(pkt_header->msg_type==HEARTBEATACK){
		verified_heartbeat[pathnumber]=temp_verified_heartbeat[pathnumber]=true;//the heartbeat for this particular route was ACKed.
#ifdef DEBUG_RTT_READINGS
		fprintf(stdout, "RTT for heartbeatack %f\n", secs()-pkt_header->tstamp_sent);
#endif
		UpdateRTT(pathnumber,secs()-pkt_header->tstamp_sent);
		UpdateThruput(pkt_header->app_id,pkt_header->tstamp_sent,true,pathnumber);
#ifdef DEBUG_RELCOMM
		fprintf (stderr,"HEARTBEAT_ACK RECEIVED for %d\n", pathnumber);
#endif //DEBUG
	}
	else if(pkt_header->msg_type==HEARTBEAT){
		verified_heartbeat[pathnumber]=temp_verified_heartbeat[pathnumber]=true;//utilize heartbeats originated from other side
#ifdef DEBUG_RELCOMM
		fprintf (stderr,"HEARTBEAT RECEIVED for %d\n", pathnumber);
#endif //DEBUG
		Send_Ack(pathnumber,pkt_header->msg_id,HEARTBEATACK,pkt_header->tstamp_sent,packetlength,pkt_header->inter_packet_time); //send heartbeatack once
	}
	else if(pkt_header->msg_type==RELIABLEACK){
#ifdef DEBUG_RELCOMM
		fprintf (stderr,"RELIABLEACK RECEIVED for %d\n", pathnumber);
#endif //DEBUG
		reliable_acked=true;
		verified_heartbeat[pathnumber]=temp_verified_heartbeat[pathnumber]=true;//the heartbeat for this particular route was ACKed.
#ifdef DEBUG_RTT_READINGS
		fprintf(stdout, "RTT for reliableack %f\n", secs()-pkt_header->tstamp_sent);
#endif
		//using the terrible hack: app_id contains the size of the original packet)
		UpdateRTT(pathnumber,secs()-pkt_header->tstamp_sent);
		UpdateThruput(pkt_header->app_id,pkt_header->tstamp_sent,true,pathnumber);

	}	
	else if(pkt_header->msg_type==RELIABLEAPP){
	  Send_Ack(pathnumber,pkt_header->msg_id,RELIABLEACK,pkt_header->tstamp_sent,packetlength,pkt_header->inter_packet_time);
                //Find_Callback(pkt_header->app_id,payload,packetlength-sizeof(r_comm_header_t));
		receive_offset[pathnumber]=0;
                Find_Callback(pkt_header->app_id,payload,packetlength-sizeof(r_comm_header_t), true);
	}	
	else if(pkt_header->msg_type==UNRELIABLEAPP){
#ifdef DEBUG_RELCOMM
		fprintf (stderr,"UNRELIABLEAPP RECEIVED for %d\n", pathnumber);
#endif //DEBUG
		Send_Ack(pathnumber,pkt_header->msg_id,UNRELIABLEACK,pkt_header->tstamp_sent,packetlength,pkt_header->inter_packet_time);
		//Find_Callback(pkt_header->app_id,payload,packetlength-sizeof(r_comm_header_t));
		Find_Callback(pkt_header->app_id,payload,packetlength-sizeof(r_comm_header_t), false);
		receive_offset[pathnumber]=0;
	}
	else if(pkt_header->msg_type==UNRELIABLE_MISSING_FRAGMENT_NOTIFICATION){
	  missing_fragment_flag[pathnumber]=true;
	  missing_fragment_number[pathnumber]=pkt_header->msg_id;
	}
	else if(pkt_header->msg_type==UNRELIABLE_FRAGMENT){
	  Send_Ack(pathnumber,pkt_header->msg_id,UNRELIABLEACK,pkt_header->tstamp_sent,packetlength,pkt_header->inter_packet_time);
		//fprintf(stderr,"OFFSETS - packet: %d, buffer: %d\n",pkt_header->offset,receive_offset[pathnumber]);
		pthread_mutex_lock(&fragment_mutex[pathnumber]);
		if(receive_offset[pathnumber]==pkt_header->offset){
		  memcpy(fragmented_receive_buffer[pathnumber]+receive_offset[pathnumber],payload,packetlength-sizeof(r_comm_header_t));
		  receive_offset[pathnumber]+=packetlength-sizeof(r_comm_header_t);
		}
		else {//missing segment
		  pthread_mutex_lock(&send_mutex[pathnumber]);
		  r_comm_header_t packet;
		  packet.msg_type=UNRELIABLE_MISSING_FRAGMENT_NOTIFICATION;
		  packet.msg_id=receive_offset[pathnumber];
		  if(sendto(send_sock[pathnumber],&packet, sizeof(packet),0,(struct sockaddr *)
			    (DestAddr + pathnumber), sizeof(DestAddr[pathnumber]))!=sizeof(packet))
		    Display_Error("sendto() sent a different number of bytes than expected");
		  receive_offset[pathnumber]=0;
		  pthread_mutex_unlock(&send_mutex[pathnumber]);
		  
		}
		
		pthread_mutex_unlock(&fragment_mutex[pathnumber]);
	}
	else if(pkt_header->msg_type==UNRELIABLE_FRAGMENT_LAST){
	  Send_Ack(pathnumber,pkt_header->msg_id,UNRELIABLEACK,pkt_header->tstamp_sent,packetlength,pkt_header->inter_packet_time);
		pthread_mutex_lock(&fragment_mutex[pathnumber]);
		if(receive_offset[pathnumber]==pkt_header->offset){
		  memcpy(fragmented_receive_buffer[pathnumber]+receive_offset[pathnumber],payload,packetlength-sizeof(r_comm_header_t));
		  Find_Callback(pkt_header->app_id,fragmented_receive_buffer[pathnumber],receive_offset[pathnumber]+packetlength-sizeof(r_comm_header_t), false);
		  receive_offset[pathnumber]=0;	
		}
		else receive_offset[pathnumber]=0;
		pthread_mutex_unlock(&fragment_mutex[pathnumber]);
	}	
	else if(pkt_header->msg_type==PLAINTEXT){
		fprintf(stderr,"Received Text Message: %s\n",payload);
		//Send_Ack(pathnumber,pkt_header->msg_id,UNRELIABLEACK,pkt_header->tstamp_sent,packetlength);
	}	
	//else if(pkt_header->msg_type==){
	//}	

	else Display_Error("Unknown Packet Type Received");
}

//Function for sending an ack, used internally by program
void Reliable_Comm::Send_Ack(const unsigned int route,unsigned int orig_msg_id, MESSAGE_TYPE ack_type, double tstamp_sent,unsigned int msgsize,unsigned int received_rate){
  unsigned int new_send_rate=  process_packet_wtcp(route,orig_msg_id,received_rate);

  pthread_mutex_lock(&send_mutex[route]);
  if(route<0 || route >=RCOM_NUM_PATHS){
    Display_Error("Sending ACK on invalid route");
    return;
  }
  if(ack_type!=RELIABLEACK && ack_type!=HEARTBEATACK && ack_type!=UNRELIABLEACK){
    Display_Error("Invalid ACK type");
    return;
  }
  r_comm_header_t pkt;//the entire packet is just a header
  pkt.inter_packet_time=new_send_rate;
  pkt.msg_type=ack_type;
  //terrible hack to keep packet header small
  pkt.app_id=msgsize;
  pkt.tstamp_sent=tstamp_sent;//simply echo the timestamp for RTT calculations for the other end
  pkt.msg_id=orig_msg_id;//echo the sequence number as well, so that the other side knows what this ack is a response to.
  //send the packet
  if (sendto(send_sock[route], & pkt, sizeof(pkt), 0, (struct sockaddr *)
	     (DestAddr + route), sizeof(DestAddr[route])) != sizeof(pkt)){
    pthread_mutex_unlock(&send_mutex[route]);
    Display_Error("sendto() sent a different number of bytes than expected");
    return;
  }
  pthread_mutex_unlock(&send_mutex[route]);
}

//finds the appropriate application for which the message was received
//and passes control to Non_Blocking_Callback
void Reliable_Comm::Find_Callback(unsigned int app_number,byte * data, unsigned int datasize, bool reliable){
  int i;
  for(i=0;i<(int)application_count;i++){
    if(application_info[i].id==app_number) {
      Non_Blocking_Callback(i,data,datasize, application_info[i], reliable);
      return;
    }
  }
#ifdef DEBUG_RELCOMM_CALLBACK
  if(i==(int)application_count)
    Display_Error("Unable to find callback function for the specific application id");
#endif
}

//First checks to see if the previous callback function for the particular application
//has terminated. If it has, it copies the received data to the application's local
//buffer, and creates the thread that will run the callback function
void Reliable_Comm::Non_Blocking_Callback(unsigned int app_array_loc, byte * data, unsigned int datasize, Reliable_Comm_AppInfo app_info, bool reliable){
#ifdef DEBUG_RELCOMM_CALLBACK
   fprintf(stderr,"::Non_Blocking_Callback: app_array_loc: %d\n",app_array_loc);
#endif
	if(data_read_by_app[app_array_loc]){
		memcpy(local_buffers[app_array_loc],data,datasize);
		cti[app_array_loc].data_length=datasize;
		cti[app_array_loc].app_info = app_info;
		cti[app_array_loc].reliable = reliable;
		data_read_by_app[app_array_loc]=false;
	}
#ifdef DEBUG_RELCOMM_CALLBACK
	else fprintf(stderr,"Previous callback has not terminated\n");
#endif
}

void Reliable_Comm::Unregister_App(int app_id){
  int counter=0;
  while(counter<(int)application_count&&application_info[counter].id!=(unsigned int)app_id){
    counter++;
  }
  if(counter<(int)application_count){
    application_count--;
    pthread_cancel(app_callback_threads[counter]);
    if(application_count>0){
      app_callback_threads[counter]=app_callback_threads[application_count];
      application_info[counter].id=application_info[application_count].id;
      application_info[counter].instance=application_info[application_count].instance;
      rec_callbacks[counter]=rec_callbacks[application_count];//application_info[application_count].;
      local_buffers[counter]=local_buffers[application_count];
      cti[counter].funcptr=cti[application_count].funcptr;
      cti[counter].done_reading_data=cti[application_count].done_reading_data;
      
    }
  }
}
//The function that registers an application with Reliable_Comm
//Parameters: a function pointer of Receiver_Callback_p type, an integer for application id,
//and a local array for storing received data
//Returns: 1 if successful, 0 if an error happens.
bool Reliable_Comm::Register_App(Receiver_Callback_p callback_func, int app_id, byte * app_buffer, void *app_instance){
	pthread_mutex_lock(&register_mutex);
	if (application_count==RCOM_MAX_APPS) {
		Display_Error("Cannot register application, maximum number of applications in use");
		return 0;
		}
	if(app_id<=0){
		Display_Error("Invalid application number");
		return 0;
		}
	for(int i=0;i<(int)application_count;i++) {
		if(application_info[i].id==(unsigned int)app_id){
				Display_Error("Application id clash with an existing application");
				return 0;
			}
		}
	application_info[application_count].id=app_id;
	application_info[application_count].instance=app_instance;
	rec_callbacks[application_count]=callback_func;
	local_buffers[application_count]=app_buffer;
	cti[application_count].funcptr=rec_callbacks[application_count];
	cti[application_count].done_reading_data=(data_read_by_app)+application_count; //pointer arithmetic

	pthread_create(&app_callback_threads[application_count],NULL, Callback_Thread,(void*) & cti[application_count]);
	//detach the thread so that resources are properly returned to the system when it terminates
	pthread_detach(app_callback_threads[application_count]);
	application_count++;
	pthread_mutex_unlock(&register_mutex);
	return 1;
}


bool Reliable_Comm::Send_Packet_on_Route(MESSAGE_TYPE m_type,byte payload[], const unsigned int payload_len,const unsigned int route){
	int total_pkt_len=payload_len+sizeof(r_comm_header_t);
	if(total_pkt_len > RCOM_MAX_PAYLOAD_SIZE){
		Display_Error("Payload too big");
		return false;
	}
	if(route<0 || route>=RCOM_NUM_PATHS){
		Display_Error("Invalid route");
		return false;
	}
	r_comm_header_t * pkt_header = (r_comm_header_t *) ((byte *)xmit_buffer + (route * RCOM_MAX_PAYLOAD_SIZE));//locate the header of the correct xmit buffer

	pthread_mutex_lock(&send_mutex[route]);
	

	//copy the payload to the buffer
	memcpy(xmit_buffer[route]+sizeof(r_comm_header_t),payload,payload_len);
	//for (int j=0;j<payload_len;j++){
	//	xmit_buffer[route][j+sizeof(r_comm_header_t)]=payload[j];
	//}

	//modify the packet header
	pkt_header->msg_type=m_type;
	
	pthread_mutex_lock(&send_count_mutex[route]);
	pkt_header->msg_id=sendcount[route]++;
	pthread_mutex_unlock(&send_count_mutex[route]);
	pkt_header->inter_packet_time=inter_packet_time[route];
	
	
	pkt_header->tstamp_sent=secs();//insert the timestamp
	UpdateThruput(total_pkt_len,pkt_header->tstamp_sent,false,route);

	//send the packet
    if (sendto(send_sock[route], xmit_buffer[route], total_pkt_len, 0, (struct sockaddr *)
		(DestAddr + route), sizeof(DestAddr[route])) != total_pkt_len){
		pthread_mutex_unlock(&send_mutex[route]);
        Display_Error("sendto() sent a different number of bytes than expected");
		return false;
    }
    usleep(inter_packet_time[route]);
    pthread_mutex_unlock(&send_mutex[route]);
    return true;
}

void Reliable_Comm::Send_Packet_Unreliable(byte payload[], const unsigned int payload_len, const unsigned int recipient_id){
  int total_pkt_len=payload_len+sizeof(r_comm_header_t);
  if(total_pkt_len > RCOM_MAX_PAYLOAD_SIZE){
    Display_Error("Payload too big");
    return;
  }
  unsigned int curr_packet_offset=0;//for payloads that have to be fragmented
  int use_route=Pick_Widest_Path();		//pick the route with the highest throughput
  r_comm_header_t * pkt_header = (r_comm_header_t *) (xmit_buffer[use_route]);//locate the header of the correct xmit buffer
  
  pthread_mutex_lock(&send_mutex[use_route]);
  
  
  //copy the payload to the buffer
  //int copy_len=PACKET_SIZE;
  //if(payload_len<PACKET_SIZE)copy_len=payload_len;
  //	memcpy(xmit_buffer[use_route]+sizeof(r_comm_header_t),payload+curr_packet_offset,copy_len);
  
  pkt_header->msg_type=UNRELIABLEAPP;
  pkt_header->app_id=recipient_id;
  
  missing_fragment_flag[use_route]=false;
  bool confirmed_missing_fragment_flag=false;
  while((!confirmed_missing_fragment_flag) && curr_packet_offset<payload_len){
    if(missing_fragment_flag[use_route]){
      if(missing_fragment_number[use_route]<curr_packet_offset)confirmed_missing_fragment_flag=true;
      else missing_fragment_flag[use_route]=false;
    }
    unsigned int actual_pkt_len;
    pthread_mutex_lock(&send_count_mutex[use_route]);
    pkt_header->msg_id=sendcount[use_route]++;
    pthread_mutex_unlock(&send_count_mutex[use_route]);
    pkt_header->inter_packet_time=inter_packet_time[use_route];
    pkt_header->tstamp_sent=secs();
    if((int)payload_len>(PACKET_SIZE)){
      pkt_header->msg_type=UNRELIABLE_FRAGMENT;
      actual_pkt_len=PACKET_SIZE+sizeof(r_comm_header_t);
      if(curr_packet_offset+PACKET_SIZE>=payload_len){
	pkt_header->msg_type=UNRELIABLE_FRAGMENT_LAST;
	actual_pkt_len=(int)payload_len%(PACKET_SIZE)+sizeof(r_comm_header_t);
      }
      
    }
    else {
      actual_pkt_len=total_pkt_len;
    }
    memcpy(xmit_buffer[use_route]+sizeof(r_comm_header_t),payload+(int)curr_packet_offset,(int)actual_pkt_len);
    pkt_header->offset=curr_packet_offset;
    UpdateThruput(actual_pkt_len,pkt_header->tstamp_sent,false,use_route);
    //send the packet
    //fprintf(stderr,"offset %d,actual_pkt_len %d, MSGTYPE %d,payload_len %d\n",curr_packet_offset,actual_pkt_len,pkt_header->msg_type,payload_len);
    if (sendto(send_sock[use_route], (xmit_buffer[use_route]), actual_pkt_len, 0, (struct sockaddr *)
	       (DestAddr + use_route), sizeof(DestAddr[use_route])) != (int)actual_pkt_len){
      Display_Error("sendto() sent a different number of bytes than expected");
    }
    usleep(inter_packet_time[use_route]);
    curr_packet_offset+=PACKET_SIZE;
  }

  pthread_mutex_unlock(&send_mutex[use_route]);
  if(confirmed_missing_fragment_flag) fprintf(stderr,"Received a missing fragment notification. Not sending the rest of the data stream\n");
}
void Reliable_Comm::Send_Packet_Reliable(byte payload[], const unsigned int payload_len, const unsigned int recipient_id){
  int total_pkt_len=payload_len+sizeof(r_comm_header_t);
  if(payload_len > (unsigned int)PACKET_SIZE){
    Display_Error("Payload too big");
    return;
  }
  
  r_comm_header_t * pkt_header = (r_comm_header_t *) (reliable_xmit_buffer);//locate the header of the correct xmit buffer
  
  //copy the payload to the buffer
  memcpy(reliable_xmit_buffer+sizeof(r_comm_header_t),payload,payload_len);
  
  //modify the packet header
  pkt_header->msg_type=RELIABLEAPP;
  pkt_header->app_id=recipient_id;
  int use_route;
  //lock a mutex and wait until ack confirmation so that packets cannot be sent out of order.
  pthread_mutex_lock(&reliable_send_mutex);
  usleep(40);
  bool delivered=false;
  while(!delivered){
    use_route=Pick_Quickest_Path();		//pick the route with the lowest round trip time:
    pthread_mutex_lock(&send_count_mutex[use_route]);
    pkt_header->msg_id=sendcount[use_route]++;
    pthread_mutex_unlock(&send_count_mutex[use_route]);
    pkt_header->inter_packet_time=inter_packet_time[use_route];
    double packet_departure_time=pkt_header->tstamp_sent=secs();
    UpdateThruput(total_pkt_len,pkt_header->tstamp_sent,false,use_route);
    //send the packet
    if (sendto(send_sock[use_route], reliable_xmit_buffer, total_pkt_len, 0, (struct sockaddr *)
	       (DestAddr + use_route), sizeof(DestAddr[use_route])) != total_pkt_len){
      Display_Error("sendto() sent a different number of bytes than expected");
      packet_departure_time -= 5*rto[use_route];//we already know that the packet did not make it so we have to rexmit, just quit the loop as quickly as possible
    }
    //fprintf(stderr,"TIME1: %f sendcount: %d rto: %f\n",secs()-packet_departure_time,sendcount,rto[use_route]);
    while((secs()-packet_departure_time)<5*rto[use_route]){
      if(reliable_acked){
	packet_departure_time-=10*rto[use_route];
	delivered=true;
	reliable_acked=false;
      }
      else usleep(100);
      //do an else yield call here
    }
    //if(!delivered)fprintf(stderr,"Resending reliable\n");
    //fprintf(stderr,"TIME2: %f %d\n",secs()-packet_departure_time,sendcount);
  }
  pthread_mutex_unlock(&reliable_send_mutex);
#ifdef DEBUG_RELCOMM
  fprintf(stderr,"Reliable Packet (msgid: %d) Delivered on route %d\n",pkt_header->msg_id,use_route);
#endif
}


bool Reliable_Comm::Setup_Sender(int port,char destips[RCOM_NUM_PATHS][16]){
    fprintf(stderr, "Sender UDP buffer size: ");

    for (int i=0;i<RCOM_NUM_PATHS;i++){
        /* Create a datagram/UDP socket */
        if ((send_sock[i] = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0){
            Display_Error("Send socket() failed");
            return false;
        }
        /*Change the size of the system udp send buffer*/
        int bigbuffersize=200000;
        if(setsockopt(send_sock[i],SOL_SOCKET,SO_SNDBUF,(char*)&bigbuffersize,sizeof(bigbuffersize))
           !=0)
            Display_Error("Unable to modify send buffer size");
        int checksize;
        socklen_t * dummy;
        dummy = new(socklen_t);
        *dummy=sizeof(checksize);
        getsockopt(rec_sock[i],SOL_SOCKET,SO_SNDBUF,&checksize,dummy);
        //fprintf(stderr,"Link %d: sender UDP buffer size [%d bytes]\n", i, checksize);
        fprintf(stderr, " [Link %d: %d KB]", i, (int)(checksize/1000.0));
        /* Construct the server address structure */
        memset(&DestAddr[i], 0, sizeof(DestAddr[i]));    /* Zero out structure */
        DestAddr[i].sin_family = AF_INET;                 /* Internet addr family */
        DestAddr[i].sin_addr.s_addr = inet_addr(destips[i]);  /* Server IP address */
        DestAddr[i].sin_port   = htons(port+i);     /* Server port */
    }
    fprintf(stderr, "\n");
    return true;
}
//changes the channel of the specified route to the specified channel number.
//if the channel number is 0, the function determines an appropriate number itself
void Reliable_Comm::Change_Channel(unsigned int route, unsigned int newchannel){
}

//This function should be called by a threaded function. It periodically sends multiple heartbeat packets(for possibility of dropped packets)
//on all routes, and changes a channel if one of the links is down.
void Reliable_Comm::Run_Heartbeat_Loop(){
    bool report_link_status = true;
    bool link_is_up[RCOM_NUM_PATHS];
    bool link_was_up[RCOM_NUM_PATHS];
    memset(link_is_up, 0x0, sizeof(bool)*RCOM_NUM_PATHS);
    memset(link_was_up, 0x0, sizeof(bool)*RCOM_NUM_PATHS);
    for(;;){
    
        for (int j=0;j<10;j++) {
            //for each link, send a heartbeat
            for (unsigned int i=0;i<RCOM_NUM_PATHS;i++){
                temp_verified_heartbeat[i]=false;//do not change the main heartbeat status, change a temporary veariable instead
                unsigned int payload_i[1]={i};
#ifdef DEBUG_RELCOMM3
                fprintf(stderr,"Sending beat on link %d\n",i);
#endif
                Send_Packet_on_Route(HEARTBEAT,(byte*)payload_i,sizeof(i),i); //send heartbeat 5 times
	
            }
            //timing between each heartbeat
            usleep(200000);
        }
        //timing before next heartbeat cycle
        usleep(400000);
        //check for whether any links are down, and if yes, change channel
        for (int i=0;i<RCOM_NUM_PATHS;i++){
            if(temp_verified_heartbeat[i]==false) {
                Change_Channel(i,0);//let the function decide what channel to switch to
                verified_heartbeat[i]=false; //make the permanent change
                link_is_up[i] = false;
#ifdef DEBUG_RELCOMM
                fprintf(stderr, "Link %d Down\n", i);
#endif
            }
            else{
                link_is_up[i] = true;
#ifdef DEBUG_RELCOMM
                fprintf(stderr,"Link %d Up\n", i);
#endif
            }
            if (link_is_up[i] != link_was_up[i]) {
                // Status has changed. Report it to the user.
                report_link_status = true;
            }
            link_was_up[i] = link_is_up[i];
        }
    
        if (report_link_status) {
            fprintf(stderr, "RCom Status:");
            for (int i=0;i<RCOM_NUM_PATHS;i++){
                fprintf(stderr, " [Link %d: %s]", i, link_is_up[i]? " up " : "down");
            }
            fprintf(stderr, "\n");
            report_link_status = false;
        }
    }
}

bool Reliable_Comm::Setup_Receiver(int port){
    fprintf(stderr, "Receiver UDP buffer size: ");
	for (int i=0;i<RCOM_NUM_PATHS;i++){
		/* Create socket for sending/receiving datagrams */
		if ((rec_sock[i] = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0){
			Display_Error("Reliable_Comm::Setup_Receiver(). socket() failed");
			return false;
		}
		/* Construct local address structure */
		memset(&(OwnAddr[i]), 0, sizeof(OwnAddr[i]));   /* Zero out structure */
		OwnAddr[i].sin_family = AF_INET;                /* Internet address family */
		OwnAddr[i].sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
		OwnAddr[i].sin_port = htons(port+i);      /* Local port */
		int bigbuffersize=800000;
		if(setsockopt(rec_sock[i],SOL_SOCKET,SO_RCVBUF,(char*)&bigbuffersize,sizeof(bigbuffersize))
		   !=0)
            Display_Error("Unable to modify send buffer size");
		int checksize;
		socklen_t * dummy;
		dummy = new(socklen_t);
		*dummy=sizeof(checksize);
		getsockopt(rec_sock[i],SOL_SOCKET,SO_RCVBUF,&checksize,dummy);
		//fprintf(stderr,"Link %d: receiver UDP buffer size [%d bytes]\n", i, checksize);
        fprintf(stderr, " [Link %d: %d KB]", i, (int)(checksize/1000.0));
		
		/* Bind to the local address */
		if (bind(rec_sock[i], (struct sockaddr *) (OwnAddr+i), sizeof(OwnAddr[i])) < 0){
			Display_Error("Reliable_Comm::Setup_Receiver(). bind() failed.");
			return false;
		}
		//fprintf(stderr,"Opened port %d\n",port+i);

	}
    fprintf(stderr, "\n");
	return true;
}

//Returns the array location of the alive link with the highest throughput.
//If no links are alive, returns -1.
int Reliable_Comm::Pick_Widest_Path(){
	int result =rand()%RCOM_NUM_PATHS;
	double max_delivery_ratio_in_bytes=0;
	for (int i=0;i<RCOM_NUM_PATHS;i++){
		double delivery_ratio_in_bytes;
		if(outgoing_thruput[i]==0)delivery_ratio_in_bytes=1.0;
		else delivery_ratio_in_bytes=(double)confirmed_thruput[i]/(double)outgoing_thruput[i];
		if(verified_heartbeat[i]&&inter_packet_time[i]<10000 &&
		   ((delivery_ratio_in_bytes>max_delivery_ratio_in_bytes)||((max_delivery_ratio_in_bytes-delivery_ratio_in_bytes<0.01)&&(rtt[i]<rtt[result])))
		   ){
			result=i;
			max_delivery_ratio_in_bytes=delivery_ratio_in_bytes;
		}
		
	}
	return result;
}

//Returns the array location of the alive link with the lowest round trip time.
//If no links are alive, returns -1.
int Reliable_Comm::Pick_Quickest_Path(){
  int result =-1;
  double min_rtt=1000000;
  for (int i=0;i<RCOM_NUM_PATHS;i++){
    if(verified_heartbeat[i]&&rtt[i]<min_rtt){
      result=i;
      min_rtt=rtt[i];
    }
  }
  return result;
}


//Infinite loop that performs blocking receive. Whenever a message is received,
//Interpret_Packet_Header() is called.
void Reliable_Comm::Run_Receiver_Loop(int pathnumber){
  if(pathnumber>=RCOM_NUM_PATHS || pathnumber < 0) return; // do not run it if the pathnumber is incorrect
  else{
	  int recvMsgSize;   // Size of received message
	  unsigned int cliAddrLen;    // Length of incoming message
	  for (;;)           // Run forever 
	  {
		  /* Set the size of the in-out parameter */
		  cliAddrLen = sizeof(NullAddr[pathnumber]);
		  /* Block until message received from a client */
		  if ((recvMsgSize = recvfrom(rec_sock[pathnumber], (received_packet_buffer[pathnumber]), RCOM_MAX_PAYLOAD_SIZE, 0,
			  (struct sockaddr *) (NullAddr+pathnumber), (socklen_t*)&cliAddrLen)) < 0)
			  Display_Error("Receiving failed");
		  //Code for simulating a packet loss. default value for PACKET_LOSS_RATE is 0, which means no packet loss.
		  else {
			  if(!(((rand()%100)+1)<PACKET_LOSS_RATE))
			    Interpret_Packet_Header((received_packet_buffer[pathnumber]),recvMsgSize,pathnumber);
		  }
	  }
  }
}

//Returns the system time in double format
double Reliable_Comm::secs(){
	struct timeval t;
	gettimeofday(&t, (struct timezone *)0);
	return(t.tv_sec+ t.tv_usec*1.e-6);
}

void Reliable_Comm::Display_Error (char * errmsg){
	fprintf(stderr,"Error: %s\n",errmsg);
}

//Updates rtt estimates using the current reading
//and also calculates a timeout value rto
void Reliable_Comm::UpdateRTT(unsigned int route,double rtt_reading){
#ifdef DEBUG_RELCOMM
  fprintf(stderr,"rtt reading %f\n",rtt_reading);
#endif
  
  pthread_mutex_lock(&rtt_mutex);
  ackcount[route]++;
  rtt[route] = (1-g)*rtt[route] + g*rtt_reading;
  delta[route] = rtt_reading - rtt[route];
  if (delta[route] < 0 ) delta[route] = -delta[route];
  rttvar[route] = (1-h)*rttvar[route] + h * (delta[route] - rttvar[route]);
  rto[route] = rtt[route] + 4*rttvar[route];  /* may want to force it > 1 */
  pthread_mutex_unlock(&rtt_mutex);
}
#if 0
void Reliable_Comm::UpdateThruput(unsigned int datasize, double tstamp_sent, bool ackbased, unsigned int route){
  
  if(ackbased&&cf_tput_queue_size[route]<THRUPUT_QUEUE_SIZE){
    confirmed_thruput_queue[route][cf_tput_queue_size[route]].tstamp=tstamp_sent;
    confirmed_thruput_queue[route][cf_tput_queue_size[route]].size_bytes=datasize;
    
  }
  else if(og_tput_queue_size[route]<THRUPUT_QUEUE_SIZE){
    outgoing_thruput_queue[route][og_tput_queue_size[route]].tstamp=tstamp_sent;
      }
}
#endif 
//for each packet about to be sent and for each ack received, update the appropriate throughput.
//void Reliable_Comm::UpdateThruput_Calculate(unsigned int datasize, double tstamp_sent,bool ackbased,unsigned int route){
void Reliable_Comm::UpdateThruput(unsigned int datasize, double tstamp_sent, bool ackbased, unsigned int route){
  pthread_mutex_lock(&thruput_mutex);
  //fprintf(stderr,"updatetput size%d\n",datasize);
  if (ackbased){
    confirmed_thruput[route]=(unsigned long int)((1-g)*(double)confirmed_thruput[route]+g*(double)datasize/(tstamp_sent-last_confirmed_timestamp[route]));
    if(tstamp_sent-last_confirmed_timestamp[route]<0)fprintf(stderr,"RComm Warning: Timestamp is earlier than last timestamp\n");
    last_confirmed_timestamp[route]=tstamp_sent;
  }
  else{
    //fprintf(stderr,"timestamp: %f, ltimestamp: %f\n",tstamp_sent,last_outgoing_timestamp[route]);
    //fprintf(stderr,"diff: %f\n",tstamp_sent-last_outgoing_timestamp[route]);
    outgoing_thruput[route]=(unsigned long int)((1-g)*(double)outgoing_thruput[route]+g*(double)datasize/(tstamp_sent-last_outgoing_timestamp[route]));
    if(tstamp_sent-last_outgoing_timestamp[route]<0)fprintf(stderr,"RComm Warning: Timestamp is earlier than last timestamp\n");
    last_outgoing_timestamp[route]=tstamp_sent;
    //fprintf(stderr,"UPDATE THRUPUT: %f\n",g*(double)datasize/(tstamp-last_outgoing_timestamp[route]));
    //fprintf(stderr,"UPDATE THRUPUT: %ul\n",g*(double)datasize/(tstamp-last_outgoing_timestamp[route]));
  }
  pthread_mutex_unlock(&thruput_mutex);

}
void Reliable_Comm::UpdateThruputConstantly(){
  while (1){
    for (int i=0;i<RCOM_NUM_PATHS;i++){
      if(secs() - last_outgoing_timestamp[i] >1.0){
	//fprintf(stderr,"sdf\n");
	UpdateThruput(0,secs(),false,i);
      }
      if(secs() - last_confirmed_timestamp[i] > 1.0) {
	UpdateThruput(0,secs(),true,i);
	//fprintf(stderr,"sdf\n");
      }
      usleep(80000);
    }
#ifdef DEBUG_PRINT_STATS_AFTER_EACH_MESSAGE
    Print_Stats();
#endif
  }
}
#endif //__Reliable_Comm_cpp

/**********************************************************************
 * $Log: Reliable_Comm.cpp,v $
 * Revision 1.1  2008-07-15 22:33:58  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.2  2007/01/02 20:49:56  endo
 * Debug info clean-up.
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.48  2006/04/16 18:48:34  envarli
 * *** empty log message ***
 *
 * Revision 1.47  2006/04/16 18:14:01  envarli
 * *** empty log message ***
 *
 * Revision 1.46  2006/04/16 18:09:47  envarli
 * *** empty log message ***
 *
 * Revision 1.45  2006/04/16 05:53:21  envarli
 * *** empty log message ***
 *
 * Revision 1.44  2006/04/16 05:15:25  envarli
 * *** empty log message ***
 *
 * Revision 1.43  2006/04/16 04:55:53  envarli
 * *** empty log message ***
 *
 * Revision 1.42  2006/04/16 04:34:13  envarli
 * *** empty log message ***
 *
 * Revision 1.41  2006/04/16 03:37:12  envarli
 * *** empty log message ***
 *
 * Revision 1.40  2006/04/16 03:31:51  envarli
 * *** empty log message ***
 *
 * Revision 1.39  2006/04/16 03:25:22  envarli
 * *** empty log message ***
 *
 * Revision 1.38  2006/04/16 03:01:14  envarli
 * *** empty log message ***
 *
 * Revision 1.37  2006/04/16 02:49:19  envarli
 * *** empty log message ***
 *
 * Revision 1.36  2006/04/16 02:27:03  envarli
 * *** empty log message ***
 *
 * Revision 1.35  2006/04/16 02:17:23  envarli
 * *** empty log message ***
 *
 * Revision 1.34  2006/04/16 00:34:11  envarli
 * *** empty log message ***
 *
 * Revision 1.33  2006/04/16 00:20:01  envarli
 * *** empty log message ***
 *
 * Revision 1.32  2006/04/16 00:16:18  envarli
 * *** empty log message ***
 *
 * Revision 1.31  2006/04/15 23:48:34  envarli
 * *** empty log message ***
 *
 * Revision 1.30  2006/04/15 23:46:43  envarli
 * *** empty log message ***
 *
 * Revision 1.29  2006/04/15 23:42:31  envarli
 * *** empty log message ***
 *
 * Revision 1.28  2006/04/15 23:09:43  envarli
 * *** empty log message ***
 *
 * Revision 1.27  2006/04/15 23:05:28  envarli
 * *** empty log message ***
 *
 * Revision 1.26  2006/04/15 21:34:00  envarli
 * Created the backbone for WTCP
 *
 * Revision 1.25  2006/04/15 01:59:19  envarli
 * *** empty log message ***
 *
 * Revision 1.24  2006/04/14 22:14:52  envarli
 * *** empty log message ***
 *
 * Revision 1.23  2006/04/14 04:39:36  envarli
 * *** empty log message ***
 *
 * Revision 1.22  2006/04/13 18:30:11  envarli
 * *** empty log message ***
 *
 * Revision 1.21  2006/04/11 05:27:12  envarli
 * *** empty log message ***
 *
 * Revision 1.20  2006/04/11 05:21:09  envarli
 * *** empty log message ***
 *
 * Revision 1.19  2006/04/08 17:01:30  envarli
 * Added a new thread for thruput calculations, not currently used but will be soon
 *
 * Revision 1.18  2006/04/06 14:23:40  envarli
 * *** empty log message ***
 *
 * Revision 1.17  2006/04/05 23:32:03  envarli
 * Can - Made sure unregisterapp works. Changed the way callback threads are handled.
 *
 * Revision 1.16  2006/04/05 21:29:22  envarli
 * *** empty log message ***
 *
 * Revision 1.15  2006/04/03 00:27:31  endo
 * PACKET_SIZE changed from #define to const int.
 *
 * Revision 1.14  2006/04/02 23:41:38  endo
 * payload_len casted as int when being compared to PACKET_SIZE
 *
 * Revision 1.13  2006/04/02 23:24:09  endo
 * Minor clean up.
 *
 * Revision 1.12  2006/04/02 22:46:29  endo
 * Unregister_App() added to Reliable_Comm.
 *
 * Revision 1.11  2006/03/29 05:27:02  envarli
 * *** empty log message ***
 *
 * Revision 1.10  2006/03/26 22:54:13  envarli
 * *** empty log message ***
 *
 * Revision 1.9  2006/03/26 17:28:12  envarli
 * *** empty log message ***
 *
 * Revision 1.8  2006/03/14 05:17:03  envarli
 * Fixed the threading-related problem where the callback were not working properly after sending a certain number of messages.
 * Now an infinite number of messages can be received properly.
 *
 * Revision 1.7  2006/03/13 20:08:54  endo
 * Debugging
 *
 * Revision 1.6  2006/03/09 18:29:09  envarli
 * Implemented application layer fragmentation. Works. It's awesome. Improved the quality of the code here and there. Made it even more thread-safe.
 *
 * Revision 1.4  2006/02/28 18:32:21  envarli
 *
 * Removed some unnecessary lines, added comments.
 *
 * Revision 1.3  2006/02/27 20:41:54  endo
 * rcomserver bug fixed.
 *
 * Revision 1.2  2006/02/25 22:10:22  endo
 * The user can now specify ports to be used.
 *
 * Revision 1.1  2006/02/24 11:50:33  endo
 * RComm_Controller checked in.
 *
 **********************************************************************/
