/**********************************************************************
 **                                                                  **
 **                        Reliable_Comm.h                           **
 **                                                                  **
 **  Written by:  I.C. Envarli                                       **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef __Reliable_Comm_h
#define __Reliable_Comm_h

#include <sys/socket.h> /* for socket(), connect(), sendto(), and recvfrom() */
#include <arpa/inet.h>  /* for sockaddr_in and inet_addr() */
#include <stdio.h>
#include <string.h>     /* for memset() */
#include <pthread.h>

#include "RComDefaultSetups.h"
//These three variables have been moved to the file above
//#define RCOM_NUM_PATHS 3
//#define RCOM_MAX_APPS 15 //The maximum number of applications that can be using reliable communication simultaneously
//#define RCOM_MAX_PAYLOAD_SIZE 8000
#define THRUPUT_QUEUE_SIZE 10
#define PACKET_LOSS_RATE 0 // 5 means 5% drop rate. 0 means no simulated packet drops.
//#define PACKET_SIZE 1434
const int PACKET_SIZE = 1434;

typedef unsigned char byte;
// Pointer to a callback function for receiving a message.
// The arguments are the recipient ID, length of the payload, reliability flag, and pointer to the application instance.
//typedef void (*Receiver_Callback_p)(int);
typedef void (*Receiver_Callback_p)(int, int, bool, void *);

typedef enum {
  UNRELIABLEACK=1,
  RELIABLEACK=2,     //the ack for a reliable message is a different type since dropped control packets have to be rexmitted
  HEARTBEATACK=3,
  HEARTBEAT=4,
  UNRELIABLEAPP=5,
  RELIABLEAPP=6,
  PLAINTEXT=9,//Debugging purposes, receiver prints the message on its screen
  UNRELIABLE_FRAGMENT=10,
  UNRELIABLE_FRAGMENT_LAST=11,
  UNRELIABLE_MISSING_FRAGMENT_NOTIFICATION=12
} MESSAGE_TYPE;

#define WTCP_PHASE_DECREASE 0
#define WTCP_PHASE_MAINTAIN 1
#define WTCP_PHASE_INCREASE 2


typedef struct r_comm_header{
  MESSAGE_TYPE msg_type;
  unsigned long int msg_id;
  double tstamp_sent;    //time stamp for RTT calculations
  unsigned int inter_packet_time;//used by WTCP
  unsigned int app_id; //application id for app messages
  unsigned int offset;//for fragmented packets
} 
r_comm_header_t;

typedef struct Reliable_Comm_AppInfo {
    unsigned int id;
    void *instance;
};



typedef struct callback_thread_info{
    Receiver_Callback_p funcptr;
    bool * done_reading_data;
    unsigned int data_length;
    Reliable_Comm_AppInfo app_info;
    bool reliable;
} callback_thread_info_t;

typedef struct thruput_entry{
  double tstamp;
  unsigned int size_bytes;
} thruput_entry_t;

class Reliable_Comm{


public:
	Reliable_Comm();
	
	unsigned int Recv_Count(unsigned int path){ if(path<RCOM_NUM_PATHS){pthread_mutex_lock(&recv_count_mutex[path]);unsigned int temp=recvcount[path]; pthread_mutex_unlock(&recv_count_mutex[path]);return temp;}else return 0;}
	unsigned int Send_Count(unsigned int path){ if(path<RCOM_NUM_PATHS){pthread_mutex_lock(&send_count_mutex[path]);unsigned int temp=sendcount[path]; pthread_mutex_unlock(&send_count_mutex[path]);return temp;}else return 0;}
	unsigned int Ack_Count(unsigned int path){if (path<RCOM_NUM_PATHS)return ackcount[path];else return 0;}
	void Print_Stats();
	const char * Get_Stats();


	void Modify_Inter_Packet_Time(int path, int change_speed_usec);
	void Unregister_App(int app_id);
	bool Register_App(Receiver_Callback_p callback_func, int app_id, byte * app_buffer, void *app_instance);
	bool Send_Packet(MESSAGE_TYPE m_type,byte payload[], const unsigned int payload_len);
	void Send_Packet_Reliable(byte payload[], const unsigned int payload_len, const unsigned int recipient_id);
	void Send_Packet_Unreliable(byte payload[],const unsigned int payload_len, const unsigned int recipient_id);
	bool Send_Packet_on_Route(MESSAGE_TYPE m_type,byte payload[], const unsigned int payload_len,const unsigned int route);
	
	

	void UpdateThruputConstantly();
	void Send_Heartbeat_on_All();
	void Send_Ack(const unsigned int route,unsigned int orig_msg_id, MESSAGE_TYPE ack_type, double tstamp, unsigned int msgsize,unsigned int received_rate);
	
	bool Setup_Receiver(int port);
	bool Setup_Sender(int port,char destips[RCOM_NUM_PATHS][16]);
	void Change_Channel(unsigned int route,unsigned int newchannel);
	
	void Interpret_Packet_Header(byte packet[],unsigned int packetlength,unsigned int pathnumber);
	void Run_Receiver_Loop(int pathnumber);
	void Run_Heartbeat_Loop();
	byte received_packet_buffer[RCOM_NUM_PATHS][RCOM_MAX_PAYLOAD_SIZE];

 private:
	pthread_mutex_t recv_count_mutex[RCOM_NUM_PATHS],send_count_mutex[RCOM_NUM_PATHS],send_mutex[RCOM_NUM_PATHS], reliable_send_mutex, rtt_mutex,register_mutex,thruput_mutex,fragment_mutex[RCOM_NUM_PATHS];
	pthread_mutex_t wtcp_mutex[RCOM_NUM_PATHS];
	unsigned int sendcount[RCOM_NUM_PATHS],ackcount[RCOM_NUM_PATHS],recvcount[RCOM_NUM_PATHS];
	
	//The transmit buffers for each paths
	byte xmit_buffer[RCOM_NUM_PATHS][RCOM_MAX_PAYLOAD_SIZE];
	byte reliable_xmit_buffer[RCOM_MAX_PAYLOAD_SIZE];//separate buffer for reliable send
	byte fragmented_receive_buffer[RCOM_NUM_PATHS][RCOM_MAX_PAYLOAD_SIZE];
	int Pick_Widest_Path();//returns the path that is not down and that has the highest throughput
	int Pick_Quickest_Path();//returns the path that is not down and that has the lowest round trip time
	struct sockaddr_in DestAddr[RCOM_NUM_PATHS]; /* Address of other side */
	struct sockaddr_in OwnAddr[RCOM_NUM_PATHS];     /* Address of self */
	struct sockaddr_in NullAddr[RCOM_NUM_PATHS];     //dummy variable used when receiving messages
	int rec_sock[RCOM_NUM_PATHS];      /* Sockets */
	int send_sock[RCOM_NUM_PATHS];     /* Sockets */
	bool verified_heartbeat[RCOM_NUM_PATHS];//set to false only if temporary variable is still false after the allowed response time
	bool temp_verified_heartbeat[RCOM_NUM_PATHS];//temporary variable set to false before heartbeat requests 
	void Display_Error(char*);
	unsigned int receive_offset[RCOM_NUM_PATHS];
	bool reliable_acked;
	int current_route;
	byte callback_buffer[RCOM_MAX_PAYLOAD_SIZE];
	//Members for intra-pc communication with applications
	Receiver_Callback_p rec_callbacks[RCOM_MAX_APPS];
	//unsigned int application_id[RCOM_MAX_APPS];
	Reliable_Comm_AppInfo application_info[RCOM_MAX_APPS];
	bool data_read_by_app[RCOM_MAX_APPS];
	byte * local_buffers[RCOM_MAX_APPS];
	unsigned int application_count;
	//void Find_Callback(unsigned int app_number,byte * data, unsigned int datasize);
	void Find_Callback(unsigned int app_number,byte * data, unsigned int datasize, bool reliable);
	//void Non_Blocking_Callback(unsigned int app_array_loc, byte * data, unsigned int datasize);
	void Non_Blocking_Callback(unsigned int app_array_loc, byte * data, unsigned int datasize, Reliable_Comm_AppInfo app_info, bool reliable);
	pthread_t app_callback_threads[RCOM_MAX_APPS];
	callback_thread_info_t cti[RCOM_MAX_APPS];
	char stats_string[1000];
	char temp_string[300];
	//Network statistics variables/functions
	unsigned long int inter_packet_time[RCOM_NUM_PATHS];
	unsigned long int outgoing_thruput[RCOM_NUM_PATHS];
	unsigned long int confirmed_thruput[RCOM_NUM_PATHS];
	double last_outgoing_timestamp[RCOM_NUM_PATHS];
	double last_confirmed_timestamp[RCOM_NUM_PATHS];
	double rto[RCOM_NUM_PATHS],delta[RCOM_NUM_PATHS],rttvar[RCOM_NUM_PATHS];
	double h,g;
	double rtt[RCOM_NUM_PATHS];
	//queues for thruput calculations
	thruput_entry_t outgoing_thruput_queue[RCOM_NUM_PATHS][THRUPUT_QUEUE_SIZE];
	thruput_entry_t confirmed_thruput_queue[RCOM_NUM_PATHS][THRUPUT_QUEUE_SIZE];
	unsigned int og_tput_queue_size[RCOM_NUM_PATHS];
	unsigned int cf_tput_queue_size[RCOM_NUM_PATHS];
	
	bool missing_fragment_flag[RCOM_NUM_PATHS];
	unsigned long int missing_fragment_number[RCOM_NUM_PATHS];

	//Roundtrip time and timeout calculations
	void UpdateRTT(unsigned int route,double rtt_reading);
	void UpdateThruput(unsigned int datasize, double tstamp,bool ackbased,unsigned int route);
	//returns the system time in seconds used for round trip time calculations
	double secs();
	//WTCP functions
	unsigned int process_packet_wtcp(int path, unsigned int seq_no,unsigned int advertised_rate);
	unsigned int  decrease_sending_rate_wtcp(int path,unsigned int adv_rate);
	void Set_Inter_Packet_Time(int path, int new_speed_usec);
	//WTCP variables
	unsigned int sigma[RCOM_NUM_PATHS];
	double alpha_plus[RCOM_NUM_PATHS];
	double alpha_minus[RCOM_NUM_PATHS];
	double beta_plus[RCOM_NUM_PATHS];
	double beta_minus[RCOM_NUM_PATHS];
	double loss_observed_reset[RCOM_NUM_PATHS];
	double loss_observed[RCOM_NUM_PATHS];
	unsigned int dropped_packets_wtcp[RCOM_NUM_PATHS];
	unsigned int last_seq_loss_reset[RCOM_NUM_PATHS];
	double loss_expected[RCOM_NUM_PATHS];
	unsigned long int expected_msg_id[RCOM_NUM_PATHS];
	byte current_wtcp_phase[RCOM_NUM_PATHS];
	unsigned long int highest_ack_seq_num_wtcp[RCOM_NUM_PATHS];
	unsigned int last_suggested_ipt[RCOM_NUM_PATHS];
	bool last_action_was_speedup[RCOM_NUM_PATHS];
	double last_received[RCOM_NUM_PATHS];
	double lavg_ratio[RCOM_NUM_PATHS];
	double savg_ratio[RCOM_NUM_PATHS];
	double T_loss;

};


#endif //__Reliable_Comm_h

/**********************************************************************
 * $Log: Reliable_Comm.h,v $
 * Revision 1.1  2008-07-15 22:33:58  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.19  2006/04/16 18:48:34  envarli
 * *** empty log message ***
 *
 * Revision 1.18  2006/04/16 02:17:23  envarli
 * *** empty log message ***
 *
 * Revision 1.17  2006/04/16 00:16:18  envarli
 * *** empty log message ***
 *
 * Revision 1.16  2006/04/15 23:42:31  envarli
 * *** empty log message ***
 *
 * Revision 1.15  2006/04/15 23:05:28  envarli
 * *** empty log message ***
 *
 * Revision 1.14  2006/04/15 21:34:00  envarli
 * Created the backbone for WTCP
 *
 * Revision 1.13  2006/04/14 22:14:52  envarli
 * *** empty log message ***
 *
 * Revision 1.12  2006/04/14 04:39:35  envarli
 * *** empty log message ***
 *
 * Revision 1.11  2006/04/11 05:21:09  envarli
 * *** empty log message ***
 *
 * Revision 1.10  2006/04/08 17:01:30  envarli
 * Added a new thread for thruput calculations, not currently used but will be soon
 *
 * Revision 1.9  2006/04/05 23:32:03  envarli
 * Can - Made sure unregisterapp works. Changed the way callback threads are handled.
 *
 * Revision 1.8  2006/04/03 00:27:31  endo
 * PACKET_SIZE changed from #define to const int.
 *
 * Revision 1.7  2006/04/02 22:46:29  endo
 * Unregister_App() added to Reliable_Comm.
 *
 * Revision 1.6  2006/03/29 05:27:02  envarli
 * *** empty log message ***
 *
 * Revision 1.5  2006/03/26 22:54:13  envarli
 * *** empty log message ***
 *
 * Revision 1.4  2006/03/09 18:29:09  envarli
 * Implemented application layer fragmentation. Works. It's awesome. Improved the quality of the code here and there. Made it even more thread-safe.
 *
 * Revision 1.2  2006/02/25 22:10:22  endo
 * The user can now specify ports to be used.
 *
 * Revision 1.1  2006/02/24 11:50:33  endo
 * RComm_Controller checked in.
 *
 **********************************************************************/
