/**********************************************************************
 **                                                                  **
 **                            rcomserver.cc                         **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This file contains code to build a rcomserver.                  **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: rcomserver.cc,v 1.1 2008-07-15 22:33:58 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <unistd.h>
#include <time.h>
#include <getopt.h>
#include <signal.h>
#include <string>
#include <vector>
#include "RComAPI.h"
#include "RComAPIFactory.h"
#include "RComIDs.h"
#include "RComDefaultSetups.h"

using std::string;
using std::vector;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const struct timespec MAIN_SLEEP_TIME_NSEC = {1, 0}; // 1 second sleep

//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------
bool gRComDebug = false;

//-----------------------------------------------------------------------
// Other variables
//-----------------------------------------------------------------------
int myID = RCOM_USER_ID_UCS_RCOM;
bool done = false;
string setupString = "";
bool useDefault = true;

//-----------------------------------------------------------------------
// Forward declaration
//-----------------------------------------------------------------------
void run(void);

//-----------------------------------------------------------------------
// The callback function that will be called when a new data is arrived
// to rcomserver itself.
//-----------------------------------------------------------------------
void cbProcessRecievedData(
    int recipientID,
    const void *data,
    size_t dataSize,
    bool reliableEnforced,
    void *extraArg)
{
    if (gRComDebug)
    {
        fprintf(stderr, "rcomserver:\n");
        fprintf(stderr, " Data delivered to RCom itself.\n");
        fprintf(stderr, " Recipient ID: %d\n", recipientID);
        fprintf(stderr, " Data Size: %d\n", dataSize);
        fprintf(stderr, " Enforced Reliability: %d\n", reliableEnforced);
        fprintf(stderr, " Extra Arg from initialize(): %s\n", (extraArg)? "Specified" : "None");
        fprintf(stderr, " Data: %s\n", (char *)data);
    }
}

//---------------------------------------------------------------------
// This function catches the signal.
//---------------------------------------------------------------------
void signalHandler(int sig)
{
    switch (sig) {

    case SIGABRT:
        fprintf(stderr, "Caught SIGABRT. Exiting the program...\n");
	done = true;
        break;

    case SIGBUS:
        fprintf(stderr, "Caught SIGBUS. Exiting the program...\n");
	done = true;
        break;

    case SIGHUP:
        fprintf(stderr, "Caught SIGHUP. Exiting the program...\n");
	done = true;
        break;

    case SIGSEGV:
        fprintf(stderr, "Caught SIGSEGV. Exiting the program...\n");
	done = true;
        break;

    case SIGPIPE:
        fprintf(stderr, "Caught SIGPIPE. Exiting the program...\n");
	done = true;
        break;

    case SIGINT:
	done = true;
        break;

    default:
        fprintf(stderr, "Caught Signal(%d). Exiting the program...\n", sig);
	done = true;
        break;
    }

    if (done)
    {
        exit(0);
    }
}

//---------------------------------------------------------------------
// Run rcomserver.
//---------------------------------------------------------------------
void run(void)
{
    RComAPI *appHandler = NULL;
    RComAPI *rcomHandler = NULL;
    string::size_type index;
    char peerRComIP[RCOM_NUM_PATHS][IP_ARRAY_SIZE];
    int localPort, interRComPort;
    int i;

    if (useDefault)
    {
        localPort = DEFAULT_RCOM_SERVER_PORT;
        interRComPort = DEFAULT_INTER_RCOM_PORT;

        for (i = 0; i < RCOM_NUM_PATHS; i++)
        {
            strcpy(peerRComIP[i], DEFAULT_PEER_RCOM_IP[i]);
        }
    }
    else
    {
        index = setupString.find_first_of(':');

        if (index != string::npos)
        {
            // Get the string before ':'.
            localPort = atoi((setupString.substr(0, index+1)).c_str());

            if (gRComDebug)
            {
                fprintf(stderr, "rcomserver: Local port = %d\n", localPort);
            }

            // Get the string after ':'.
            setupString = setupString.substr(index+1, string::npos);

            index = setupString.find_first_of(':');

            if (index != string::npos)
            {
                // Get the string before ':'.
                interRComPort = atoi((setupString.substr(0, index)).c_str());

                if (gRComDebug)
                {
                    fprintf(stderr, "rcomserver: Inter-RCom port = %d\n", interRComPort);
                }

                // Get the string after ':'.
                setupString = setupString.substr(index+1, string::npos);

                index = setupString.find_first_of(':');

                for (i = 0; i < (RCOM_NUM_PATHS-1); i++)
                {
                    if (index != string::npos)
                    {
                        // Get the string before ':'.
                        strcpy(peerRComIP[i], (setupString.substr(0, index)).c_str());

                        if (gRComDebug)
                        {
                            fprintf(stderr, "rcomserver: Peer-RCom IP[%d] = %s\n", i, peerRComIP[i]);
                        }

                        // Get the string after ':'.
                        setupString = setupString.substr(index+1, string::npos);

                        index = setupString.find_first_of(':');
                    }
                    else
                    {
                        fprintf(
                            stderr,
                            "Error(rcomserver): Invalid argument.\n");
                        exit(0);
                    }
                }

                strcpy(peerRComIP[RCOM_NUM_PATHS-1], setupString.c_str());

                if (gRComDebug)
                {
                    fprintf(stderr, "rcomserver: Peer-RCom IP[%d] = %s\n", RCOM_NUM_PATHS-1, peerRComIP[RCOM_NUM_PATHS-1]);
                }
            }
            else
            {
                fprintf(
                    stderr,
                    "Error(rcomserver): Invalid argument.\n");
                exit(0);
            }
        }
        else
        {
            fprintf(
                stderr,
                "Error(rcomserver): Invalid argument.\n");
            exit(0);
        }
    }

    // Create the application handler through RComAPI.
    appHandler = RComAPIFactory::createRComAPI(RCOMAPI_TYPE_RCOM_SERVER);

    if (appHandler == NULL)
    {
        fprintf(
            stderr,
            "Error(rcomserver): RComServer could not be created.\n");
        exit(0);
    }

    // Set the RComServer-to-RComClient port
    ((RComServer *)appHandler)->setLocalPort(localPort);

    // Set the RCom-to-RCom port and IP addresses
    rcomHandler = ((RComServer *)appHandler)->getMainRComHandler();

    if (rcomHandler != NULL)
    {
        ((RComInterthread *)rcomHandler)->setInterRComPort(interRComPort);
        ((RComInterthread *)rcomHandler)->setPeerRComIPs(peerRComIP);
    }
    else
    {
        fprintf(
            stderr,
            "Error(rcomserver): RCom Handler is NULL.\n");
        exit(0);
    }

    // Initialize. Supply the callback function for arriving data fom RCom.
    if (!(appHandler->initialize(myID, cbProcessRecievedData, NULL)))
    {
        fprintf(
            stderr,
            "Error(rcomserver): RComServer could not be initialized.\n");
        exit(0);
    }

    while (!done)
    {
        // Don't kill the main process.
        nanosleep(&MAIN_SLEEP_TIME_NSEC, NULL);
    }
}

//---------------------------------------------------------------------
// Main function for rcomserver.
//---------------------------------------------------------------------
int main(int argc, char **argv)
{
    int c;

    // Set the signal handler.
    signal(SIGINT, signalHandler);
    signal(SIGABRT, signalHandler);
    signal(SIGTERM, signalHandler);
    signal(SIGBUS, signalHandler);
    signal(SIGHUP, signalHandler);
    signal(SIGSEGV, signalHandler);
    signal(SIGPIPE, signalHandler);

    if (argc > 1)
    {
        while ((c = getopt(argc, argv, "ds:")) != -1)
        {
            switch (c) {

            case 'd':
                gRComDebug = true;
                break;

            case 's': // Argument Format: "-s localPort:interRComPort:IP1:IP2:IP3"
                setupString = optarg;
                useDefault = false;
                break;
            }
        }
    }

    run();

    return 0;
}

/**********************************************************************
 * $Log: rcomserver.cc,v $
 * Revision 1.1  2008-07-15 22:33:58  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.2  2006/08/09 21:59:55  endo
 * gDebug -> gRComDebug
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.8  2006/04/05 23:48:47  endo
 * Singal handler implemented.
 *
 * Revision 1.7  2006/04/02 23:24:09  endo
 * Minor clean up.
 *
 * Revision 1.6  2006/03/31 22:53:28  endo
 * *** empty log message ***
 *
 * Revision 1.5  2006/03/08 01:19:41  jinhlee
 * original directories for image compress
 *
 * Revision 1.4  2006/02/27 20:41:54  endo
 * rcomserver bug fixed.
 *
 * Revision 1.3  2006/02/25 22:10:23  endo
 * The user can now specify ports to be used.
 *
 * Revision 1.2  2006/02/25 16:30:59  endo
 * *** empty log message ***
 *
 * Revision 1.1  2006/02/25 16:17:49  endo
 * rcomserver added
 *
 * Revision 1.1  2006/02/24 11:50:34  endo
 * RComm_Controller checked in.
 *
 * Revision 1.1  2006/02/22 07:17:22  endo
 * RComServer and RComClient implemented.
 *
 **********************************************************************/
