/**********************************************************************
 **                                                                  **
 **                         FrameGrabber.c                           **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Video framegrabber control for HServer                          **
 **                                                                  **
 **  Copyright 2000 - 2008, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: frame_grabber.c,v 1.1 2008-07-15 22:33:48 endo Exp $ */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "HServerTypes.h"
#include "hserver.h"
#include "message.h"
#include "statusbar.h"
#include "video.h"
#include "frame_grabber.h"
#include "xwindow.h"
#include "robot_config.h"
#include "sensors.h"

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const int FrameGrabber::DEFAULT_CHANNEL_ = 1;
//const int FrameGrabber::DEFAULT_HEIGHT_ = 120;
//const int FrameGrabber::DEFAULT_WIDTH_ = 160;
const int FrameGrabber::DEFAULT_HEIGHT_ = 240;
const int FrameGrabber::DEFAULT_WIDTH_ = 320;

//-----------------------------------------------------------------------
// Contructor
//-----------------------------------------------------------------------
FrameGrabber::FrameGrabber(
    Video **a,
    string portString,
    bool showImage) :
    Video(a, HS_VIDEO_TYPE_FRAMEGRABBER, HS_MODULE_NAME_VIDEO_FRAMEGRABBER, showImage),
    portString_(portString),
    mappedImage_(NULL),
    channel_(DEFAULT_CHANNEL_),
    shCurrentImageFd_(-1)
{
    int width, height;

    updateStatusBar_();

    // Initialize the frame grabber.
    if (!initV4Linux_(DEFAULT_WIDTH_, DEFAULT_HEIGHT_))
    {
        printfTextWindow("Error: FrameGrabber could not be initialized. Exiting FrameGrabber.\n");
        delete this;
        a = NULL;
        return;
    }

    // Start the reader thread.
    pthread_create(&readerThread_, NULL, &startReaderThread_, (void*)this);

    // Start the watchdog.
    if (gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(
            NAME_,
            WATCHDOG_CHECK_INTERVAL_SEC_,
            getpid(),
            readerThread_);
    }

    if (showImage_)
    {
        pthread_mutex_lock(&currentImageMutex_);
        width = shCurrentImageWidth_;
        height = shCurrentImageHeight_;
        pthread_mutex_unlock(&currentImageMutex_);

        new Xwindow(
            &xwindow_,
            (char *)(NAME_.c_str()),
            0,
            0,
            width,
            height);

        // Start the display thread.
        pthread_create(&displayThread_, NULL, &startDisplayThread_, (void*)this);
    }

    refreshScreen();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
FrameGrabber::~FrameGrabber()
{
    if (displayThreadIsUp_)
    {
        displayThreadIsUp_ = false;
        pthread_cancel(displayThread_);
        pthread_join(displayThread_, NULL);
    }

    if (readerThreadIsUp_)
    {
        readerThreadIsUp_ = false;
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    if (shCurrentImage_ != NULL)
    {
        munmap((char*)shCurrentImage_, shCurrentImageSize_);
    }

    if (shCurrentImageFd_ > 0)
    {
        close(shCurrentImageFd_);
    }

    printTextWindow("FrameGrabber disconnected");
}

//-----------------------------------------------------------------------
// This function initializes V4Linux.
//-----------------------------------------------------------------------
bool FrameGrabber::initV4Linux_(int width, int height)
{
    struct video_mbuf videoBuffers = {2*0x151000, 0, {0,0x151000}};
    int i;
    int imageSize;
    bool hadMmapError = false;

    shCurrentImageFd_ = open(portString_.c_str(), O_RDWR);

    if (shCurrentImageFd_ <= 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber device open error");
        }
        return false;
    }
  
    if (ioctl(shCurrentImageFd_, VIDIOCGCAP, &videoCapability_) < 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber VIDIOCGCAP error");
        }
        return false;
    }
  
    if (report_level >= HS_REP_LEV_DEBUG)
    {
        printfTextWindow("FrameGrabber Capture Device Name: %s", videoCapability_.name);

        printfTextWindow(
            "   channels %d  audios %d  width %d %d height %d %d", 
            videoCapability_.channels,
            videoCapability_.audios,
            videoCapability_.minwidth,
            videoCapability_.maxwidth, 
            videoCapability_.minheight,
            videoCapability_.maxheight);

        rstr[0] = 0;

        if (videoCapability_.type & VID_TYPE_CAPTURE)
        {
            strcat(rstr, "Can capture, ");
        }

        if (videoCapability_.type & VID_TYPE_TUNER)
        {
            strcat(rstr, "Can tune, ");
        }

        if (videoCapability_.type & VID_TYPE_TELETEXT)
        {
            strcat(rstr, "Does teletext, ");
        }

        if (videoCapability_.type & VID_TYPE_OVERLAY)
        {
            strcat(rstr, "Overlay onto frame buffer, ");
        }

        if (videoCapability_.type & VID_TYPE_CLIPPING)
        {
            strcat(rstr, "Can clip, ");
        }

        if (videoCapability_.type & VID_TYPE_FRAMERAM)
        {
            strcat(rstr, "Uses frame buffer memory, ");
        }

        if (videoCapability_.type & VID_TYPE_SCALES)
        {
            strcat(rstr, "Scalable, ");
        }

        if (videoCapability_.type & VID_TYPE_MONOCHROME)
        {
            strcat(rstr, "Monchrome, ");
        }

        if (videoCapability_.type & VID_TYPE_SUBCAPTURE)
        {
            strcat(rstr, "Capture subareas.");
        }

        printTextWindow(rstr);

        for(i = 0; i < videoCapability_.channels; i++)
        {
            videoChannel_.channel = i;

            if(ioctl(shCurrentImageFd_, VIDIOCGCHAN, &videoChannel_) < 0)
            {
                if (report_level >= HS_REP_LEV_ERR)
                {
                    printTextWindow("FrameGrabber VIDIOCGCHAN error");
                }

                return false;
            }

            printfTextWindow(
                "FrameGrabber Source (%d) Name : %s  tuners: %d  flags: %d  norm %d", 
                i,
                videoChannel_.name,
                videoChannel_.tuners,
                videoChannel_.flags,
                videoChannel_.norm);
        }
    }
  
    shCurrentImageWidth_ = width;
    shCurrentImageHeight_ = height;

    imageSize = height*width*3;
    shCurrentImageSize_ = imageSize;
    shCurrentImage_ = new unsigned char[imageSize];

    if(ioctl(shCurrentImageFd_, VIDIOCGMBUF, &videoBuffers) < 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber VIDIOCGMBUF error");
        }

        return false;
    }

    mappedImage_ = (unsigned char *)mmap(
        0,
        videoBuffers.size,
        PROT_READ | PROT_WRITE,
        MAP_SHARED,
        shCurrentImageFd_,
        0);

    if (((int)mappedImage_) == -1)
    {
        shCurrentImage_ = NULL;
        shCurrentImageSize_ = 0;
        shCurrentImageWidth_ = 0;
        shCurrentImageHeight_ = 0;
        hadMmapError = true;
    }

    if (hadMmapError)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber mmap error (1)");
        }

        return false;
    }
  
    videoChannel_.channel = channel_;
    videoChannel_.norm = 1;

    if(ioctl(shCurrentImageFd_, VIDIOCSCHAN, &videoChannel_) < 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber VIDIOCSCHAN error");
        }
        return false;
    }

    if (report_level >= HS_REP_LEV_DEBUG)
    {
        if(ioctl(shCurrentImageFd_, VIDIOCGPICT, &videoPicture_) < 0)
        {
            if (report_level >= HS_REP_LEV_ERR)
            {
                printTextWindow("FrameGrabber VIDIOCGPICT error");
            }
            return false;
        }

        printfTextWindow(
            "Picture bight %d hue %d color %d contrast %d whiteness %d depth %d palette %d", 
            videoPicture_.brightness,
            videoPicture_.hue,
            videoPicture_.colour,
            videoPicture_.contrast, 
            videoPicture_.whiteness,
            videoPicture_.depth,
            videoPicture_.palette);

        videoAudio_.audio = 0;

        if(ioctl(shCurrentImageFd_, VIDIOCGAUDIO, &videoAudio_) < 0)
        {
            if (report_level >= HS_REP_LEV_ERR)
            {
                printTextWindow("FrameGrabber VIDIOCGAUDIO error");
            }
            return false;
        }

        printfTextWindow(
            "Audio volume %d bass %d treble %d flags %d name %s mode %d balance %d step %d", 
            videoAudio_.volume,
            videoAudio_.bass,
            videoAudio_.treble,
            videoAudio_.flags,
            videoAudio_.name, 
            videoAudio_.mode,
            videoAudio_.balance,
            videoAudio_.step);
    }

    videoMmap_.frame = 0;
    videoMmap_.height = height;
    videoMmap_.width = width;
    videoMmap_.format = VIDEO_PALETTE_RGB24;

    if(ioctl(shCurrentImageFd_, VIDIOCMCAPTURE, &videoMmap_) < 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber VIDIOCMCAPTURE error");
        }
        return false;
    }

    if(ioctl(shCurrentImageFd_, VIDIOCSYNC, &(videoMmap_.frame)) < 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber VIDIOSYNC error (1)");
        }
        return false;
    }

    memcpy(shCurrentImage_, mappedImage_, shCurrentImageSize_);

    return true;
}
  

//-----------------------------------------------------------------------
// This function sets the channel.
//-----------------------------------------------------------------------
void FrameGrabber::setChannel(int channel)
{
    if ((videoChannel_.channel != channel) || (channel_ != channel))
    {
        videoChannel_.channel = channel;
        videoChannel_.norm = 1;
        channel_ = channel;

        pthread_mutex_lock(&currentImageMutex_);

        if(ioctl(shCurrentImageFd_, VIDIOCSCHAN, &videoChannel_) < 0)
        {
            if (report_level >= HS_REP_LEV_ERR)
            {
                printTextWindow("FrameGrabber VIDIOCSCHAN error");
            }
        }

        pthread_mutex_unlock(&currentImageMutex_);
    }
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void FrameGrabber::updateStatusBar_(void)
{
    char buf[100];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[0] = 0;
    strcat(statusStr, "FrameGrabber: ");

    if (frameRate_ > 0)
    {
        sprintf(
            buf,
            "%c %.2f frames/sec",
            statusbarSpinner_->getStatus(),
            frameRate_);
    }
    else
    {
        sprintf(
            buf,
            "%c Connecting...",
            statusbarSpinner_->getStatus());
    }

    if (DISPLAY_FREQUENCY)
    {
        getFrequency_(frequency);
        sprintf(buf, "%s : %.2f Hz", buf, frequency);
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

//-----------------------------------------------------------------------
// This function reads the video data from the video card (Bt848).
//-----------------------------------------------------------------------
bool FrameGrabber::readVideoData_(void)
{
    double currentTime;
    bool hadError = false;

    currentTime = getCurrentEpochTime();

    pthread_mutex_lock(&currentImageMutex_);

    if(ioctl(shCurrentImageFd_, VIDIOCMCAPTURE, &videoMmap_) < 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber VIDIOCMCAPTURE error");
        }
        return false;
    }

    if(ioctl(shCurrentImageFd_, VIDIOCSYNC, &(videoMmap_.frame)) < 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber VIDIOSYNC error (1)");
        }
        return false;
    }

    memcpy(shCurrentImage_, mappedImage_, shCurrentImageSize_);
    shCurrentImageTimeStamp_ = currentTime;

    pthread_mutex_unlock(&currentImageMutex_);

    if (hadError)
    {
        return false;
    }
    
    if ((lastCapturedTime_ != 0) && (currentTime != lastCapturedTime_))
    {
        frameRate_ = 1.0 / (currentTime - lastCapturedTime_);
    }

    lastCapturedTime_ = currentTime;

    return true;
}

//-----------------------------------------------------------------------
// This function sets the size.
//-----------------------------------------------------------------------
void FrameGrabber::setSize_(int height, int width)
{
    int imageSize;
    bool hadMmapError = false;

    imageSize = height * width * 3;

    pthread_mutex_lock(&currentImageMutex_);

    // Unmap first.
    if (shCurrentImage_ != NULL)
    {
        munmap((char*)shCurrentImage_, shCurrentImageSize_);
    }

    shCurrentImageSize_ = imageSize;
    shCurrentImageWidth_ = width;
    shCurrentImageHeight_ = height;

    shCurrentImage_ = (unsigned char*)mmap(
        0,
        imageSize,
        PROT_READ | PROT_WRITE, MAP_SHARED,
        shCurrentImageFd_,
        0);

    if (((int)shCurrentImage_) < 0)
    {
        shCurrentImage_ = NULL;
        shCurrentImageSize_ = 0;
        shCurrentImageWidth_ = 0;
        shCurrentImageHeight_ = 0;
        hadMmapError = true;
    }

    videoMmap_.width = shCurrentImageWidth_;
    videoMmap_.height = shCurrentImageHeight_;

    pthread_mutex_unlock(&currentImageMutex_);

    if (hadMmapError)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("FrameGrabber mmap error (2)");
        }

        return;
    }

    printfTextWindow("FrameGrabber: Capture resolution set to %dx%d", width, height);

    if (showImage_)
    {
        if (xwindow_ != NULL)
        {
            delete xwindow_;
            xwindow_ = NULL;
        }

        new Xwindow(
            &xwindow_,
            (char *)(NAME_.c_str()),
            0,
            0,
            width,
            height);
    }
}

//-----------------------------------------------------------------------
// This function sets up the user interface for the FrameGrabber.
//-----------------------------------------------------------------------
void FrameGrabber::control(void)
{
    bool bDone = false;
    int msgData[2];
    int c;

    msgData[0] = 0;
    msgData[1] = 0;

    messageDrawWindow(
        EnMessageType_VIDEO_CONTROL,
        EnMessageErrType_NONE,
        msgData);

    do
    {
        c = getch();

        switch(c) {

        case 'd':
            delete this;
            return;
            break;

        case 'r':
            refreshScreen();
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            bDone = true;
            break;

        case 'g':
            //capture();
            break;

        case 's':
            /*
            streaming = !streaming;
            updateStatusBar_();
            msgData[0] = streaming;
            messageDrawWindow(EnMessageType_VIDEO_CONTROL, EnMessageErrType_NONE, msgData);
            */
            break;
        case 'a':
        case 'b':
        case 'c':
            setChannel(c-'a');
            break;

        case '!':
            setSize_(120, 160);
            break;

        case '@':
            setSize_(240, 320);
            break;

        case '#':
            setSize_(360, 480);
            break;

        case '$':
            setSize_(480, 640);
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;
        }
    } while (!bDone);

    messageHide();
}

/**********************************************************************
# $Log: frame_grabber.c,v $
# Revision 1.1  2008-07-15 22:33:48  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.5  2008/04/12 20:17:24  endo
# OpenCV face detection implemented.
#
# Revision 1.4  2008/04/11 20:22:44  endo
# FrameGrabber fixed.
#
# Revision 1.3  2008/01/21 19:14:15  endo
# Regular backup.
#
# Revision 1.2  2006/12/07 20:30:22  endo
# Regular backup.
#
# Revision 1.1  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.2  2006/11/20 16:43:37  endo
# Xwindow improved.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2006/05/15 06:01:03  endo
# gcc-3.4 upgrade
#
# Revision 1.1.1.1  2005/02/06 22:59:43  endo
# AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
#
# Revision 1.6  2003/04/06 13:32:37  endo
# gcc 3.1.1
#
# Revision 1.5  2001/03/23 21:32:46  blee
# altered to use a config file
#
# Revision 1.4  2000/12/12 23:23:24  blee
# Changed #defines for the user interface to enumerated types.
#
# Revision 1.3  2000/09/19 03:51:29  endo
# RCS log added.
#
#
#**********************************************************************/
