/**********************************************************************
 **                                                                  **
 **                          gazebo-robot.c                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for Gazebo robot (Player/Stage)                       **
 **                                                                  **
 **  Copyright 2006, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: gazebo-robot.c,v 1.9 2009/09/29 20:16:57 zkira Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <pthread.h>
#include <math.h>
#include <string.h>

#include "HServerTypes.h"
#include "hserver.h"
#include "sensors.h"
#include "robot_config.h"
#include "message.h"
#include "statusbar.h"
#include "gazebo-robot.h"
#include "Watchdog.h"
#include "LogManager.h"

//-----------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t GazeboRobot::gazeboPoseOffsetMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t GazeboRobot::latestRawPoseMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const GazeboRobot::RobotSpec_t GazeboRobot::GAZEBO_ROBOT_SPECS_[NUM_GAZEBO_ROBOT_CLASSES] =
{
    {
      GAZEBO_ROBOT_CLASS_HELICOPTER,
      "Helicopter", // className
    },
    {
      GAZEBO_ROBOT_CLASS_ATRVJR,
      "ATRVJr", // className
    },
    {
      GAZEBO_ROBOT_CLASS_P2AT,
      "P2AT", // className
    }, 
    {
      GAZEBO_ROBOT_CLASS_P2DX,
      "P2DX", // className
    }
};

const string GazeboRobot::DEFAULT_GAZEBO_ROBOT_NAME_ = "robot1";
const string GazeboRobot::DEFAULT_GAZEBO_ROBOT_CLASS_ = "P2DX";
const float GazeboRobot::VIRTUAL_BUMPER_DISTANCE_ = 0.5;
const int GazeboRobot::MAIN_THREAD_USLEEP_ = 20000;
const int GazeboRobot::SKIP_STATUSBAR_UPDATE_ = 10;
const int GazeboRobot::WATCHDOG_CHECK_INTERVAL_SEC_ = 2;
const int GazeboRobot::DEFAULT_GAZEBO_SERVER_ID_ = 0;
const int GazeboRobot::DEFAULT_GAZEBO_CLIENT_ID_ = 0;
const bool GazeboRobot::REPORT_SONARS_ = false;
const bool GazeboRobot::REPORT_VIRTUAL_BUMPER_ = false;
const bool GazeboRobot::USE_GROUND_TRUTH_ = true;

//-----------------------------------------------------------------------
// Contructor
//-----------------------------------------------------------------------
GazeboRobot::GazeboRobot(
  Robot** a,
  const char *gazeboRobotName,
  const char *gazeboRobotClass,
  const int clientID,
  const int serverID
  ) : 
    Robot(a, HS_ROBOT_TYPE_GAZEBO, HS_MODULE_NAME_ROBOT_GAZEBO),
    sensorSonar_(NULL),
    partXyt_(NULL),
    partSonar_(NULL),
    gazeboClient_(NULL),
    gazeboPosition_(NULL),
    gazeboTruth_(NULL),
    gazeboSonar_(NULL),
    gazeboRobotName_(gazeboRobotName),
    sonarAngles_(NULL),
    sonarPos_(NULL),
    gazeboServerID_(serverID),
    gazeboClientID_(clientID),
    numSonars_(0),
    readerThreadIsUp_(false),
    robotSpecIndex_(-1)
{
    gazeboRobotName_ = (gazeboRobotName != NULL)?
        gazeboRobotName : DEFAULT_GAZEBO_ROBOT_NAME_;

    gazeboRobotClass_ = (gazeboRobotClass != NULL)?
        gazeboRobotClass : DEFAULT_GAZEBO_ROBOT_CLASS_;

    pthread_mutex_init(&gazeboPoseOffsetMutex_, NULL);
    pthread_mutex_init(&latestRawPoseMutex_, NULL);

    resetPose_();
    updateStatusBar_();

    if (!initializeGazebo_())
    {
        printfTextWindow("Error: GazeboRobot could not be initialized. Exiting Gazebo.\n");
        delete this;
        return;
    }

    addSensors_();

    pthread_create(&readerThread_, NULL, &startReaderThread_, (void*)this);

    if (gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(NAME_, WATCHDOG_CHECK_INTERVAL_SEC_, getpid(), readerThread_);
    }

    refreshScreen();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
GazeboRobot::~GazeboRobot(void)
{
    if (readerThreadIsUp_)
    {
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    // Close the interface and free object for sonar.
    if (gazeboSonar_ != NULL)
    {
        gz_sonar_close(gazeboSonar_);
        gz_sonar_free(gazeboSonar_);
        gazeboSonar_ = NULL;

        if (sonarPos_ != NULL)
        {
            delete [] sonarPos_;
            sonarPos_ = NULL;
        }

        if (sonarAngles_ != NULL)
        {
            delete [] sonarAngles_;
            sonarAngles_ = NULL;
        }
    }

    // Close the interface and free object for position.
    if (gazeboPosition_ != NULL)
    {
        gz_position_close(gazeboPosition_);
        gz_position_free(gazeboPosition_);
        gazeboPosition_ = NULL;
    }

    // Close the interface and free object for truth
    if (gazeboTruth_ != NULL)
    {
        gz_truth_close(gazeboTruth_);
        gz_truth_free(gazeboTruth_);
        gazeboTruth_ = NULL;
    }

    // Close the connection and free the client object
    if (gazeboClient_ != NULL)
    {
        gz_client_disconnect(gazeboClient_);
        gz_client_free(gazeboClient_);
        gazeboClient_ = NULL;
    }

    delete partXyt_;
}

//-----------------------------------------------------------------------
// This function initialize gazebo
//-----------------------------------------------------------------------
bool GazeboRobot::initializeGazebo_(void)
{
    float x, y, dist;
    int i;

    // Convert class name of robot to an integer type

    for (i = 0; i < NUM_GAZEBO_ROBOT_CLASSES; i++)
    {
        if (GAZEBO_ROBOT_SPECS_[i].className == gazeboRobotClass_)
        {
            robotSpecIndex_ = i;
            break;
        }
    }

    if (robotSpecIndex_ < 0)
    {
        robotSpecIndex_ = 0;

        printfTextWindow(
            "Warning: GazeboRobot::%s(). Robot spect %d not found.\n",
            __FUNCTION__,
            
            GAZEBO_ROBOT_SPECS_[robotSpecIndex_].className.c_str());
    }

    
    // Print all errors
    gz_error_init(1, 9);

    // Create a client object
    gazeboClient_ = gz_client_alloc();  

    // Connect to the server
    if (gz_client_connect_wait(gazeboClient_, gazeboServerID_, gazeboClientID_) != 0)
    {
        printfTextWindow("Error: GazeboRobot could not connect to the client.\n");
        gz_client_free(gazeboClient_);
        gazeboClient_ = NULL;
        return false;
    }

    // Create a position interface
    gazeboPosition_ = gz_position_alloc();
   
    // Connect to device on server
    if (gz_position_open(gazeboPosition_, gazeboClient_, gazeboRobotName_.c_str()) != 0)
    {
        printfTextWindow("Error: GazeboRobot could not open the position device.\n");
        gz_position_free(gazeboPosition_);
        gazeboPosition_ = NULL;
        return false;
    }

    if (USE_GROUND_TRUTH_) {
      gazeboTruth_ = gz_truth_alloc();
      char name[10000];
      sprintf(name, "truth_%s", gazeboRobotName_.c_str());
      printfTextWindow("Connecting to %s", name);
      if (gz_truth_open(gazeboTruth_, gazeboClient_, name ) != 0)
      {
        printfTextWindow("Error: GazeboRobot could not open the ground truth device. (Make sure it's called truth_%s in world file)\n", gazeboRobotName_.c_str());
        gz_truth_free(gazeboTruth_);
        gazeboTruth_ = NULL;
        return false;
      }
    }

    // Finally, create a sonar interface
    gazeboSonar_ = gz_sonar_alloc();

    // Connect to device on server
    if (gz_sonar_open(gazeboSonar_, gazeboClient_, gazeboRobotName_.c_str()) != 0)
    {
        // The sonar device not mounted on this robot. Proceed without it.
        printfTextWindow("Warning: GazeboRobot could not open the sonar device.\n");
        gz_sonar_free(gazeboSonar_);
        gazeboSonar_ = NULL;
    }
    else
    {
        memset(sonarLoc_, 0x0, sizeof(sonarLoc_));

        // Get the angles and the positions.
        gz_sonar_lock(gazeboSonar_, 1);

        numSonars_ = gazeboSonar_->data->sonar_count;

        sonarAngles_ = new float[numSonars_];
        sonarPos_ = new float[numSonars_];

        for (i = 0; i < numSonars_; i++)
        {
            sonarAngles_[i] = GAZEBO_ROBOT_RAD2DEG(gazeboSonar_->data->sonar_rot[i][2]); // yaw

            x = gazeboSonar_->data->sonar_pos[i][0];
            y = gazeboSonar_->data->sonar_pos[i][1];
            dist = sqrt(pow(x, 2) + pow(y, 2));

            sonarPos_[i] = dist;
        }

        gz_sonar_unlock(gazeboSonar_);
    }

    return true;
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void GazeboRobot::updateStatusBar_(void)
{
    HSPose_t pose;
    char buf[ 100 ];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[ 0 ] = 0;
    strcat(statusStr, "GazeboRobot [");

    getPose(pose);

    sprintf(
        buf,
        "%s]: %c x: %.2f  y: %.2f  t: %.2f",
        gazeboRobotName_.c_str(),
        statusbarSpinner_->getStatus(),
        pose.loc.x,
        pose.loc.y,
        pose.rot.yaw);

    if (DISPLAY_FREQUENCY)
    {
        getFrequency_(frequency);
        sprintf(buf, "%s : %.2f Hz", buf, frequency);
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

//-----------------------------------------------------------------------
// This function starts the reader thread.
//-----------------------------------------------------------------------
void *GazeboRobot::startReaderThread_(void *gazeboInstance)
{
    ((GazeboRobot*) gazeboInstance)->readerLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to move GazeboRobot.
//-----------------------------------------------------------------------
void GazeboRobot::readerLoop_(void)
{
    int count = 0;
    
    readerThreadIsUp_ = true;

    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Read the pose data from the gazebo.
        readPositionData_();

        // Check for the pthread termination again.
        pthread_testcancel();

        if (gazeboSonar_ != NULL)
        {
            // Read the sonar data from the gazebo.
            readSonarData_();

            // Check for the pthread termination again.
            pthread_testcancel();
        }

        // Compute the control speed.
        compControlSpeed_();

        // Control the robot.
        controlRobot_();

        // Compute the frequency.
        compFrequency_();

        // Compute the velocity.
        updateVelocity_();

        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        if (gLogManager != NULL)
        {
            manageLog_(readerThread_);
        }

        if (watchdog_ != NULL)
        {
            watchdog_->notifyUpdate();
        }

        usleep(MAIN_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function reads the position data from gazebo.
//-----------------------------------------------------------------------
void GazeboRobot::readPositionData_(void)
{
    HSPose_t pose, lastPose;
    double curTime, dist, dt;
    HSSpeed_t controlSpeed;
    static double lastTime = 0;

    memset((HSPose_t *)&pose, 0x0, sizeof(HSPose_t));

    if (USE_GROUND_TRUTH_) {
      gz_truth_lock(gazeboTruth_, 1);

      pose.loc.x = gazeboTruth_->data->pos[0];
      pose.loc.y = gazeboTruth_->data->pos[1];
      pose.loc.z = gazeboTruth_->data->pos[2];

      double euler[3];
      gz_truth_euler_from_quatern(gazeboTruth_, euler, gazeboTruth_->data->rot);
      pose.rot.yaw = euler[2]; // - (M_PI/2.0);
      pose.rot.pitch = euler[1];
      pose.rot.roll = euler[0];

      gz_truth_unlock(gazeboTruth_);
    } else {
      gz_position_lock(gazeboPosition_, 1);

      pose.loc.x = gazeboPosition_->data->pos[0];
      pose.loc.y = gazeboPosition_->data->pos[1];
      pose.loc.z = gazeboPosition_->data->pos[2];
      pose.rot.yaw = gazeboPosition_->data->rot[2];
      pose.rot.pitch = gazeboPosition_->data->rot[1];
      pose.rot.roll = gazeboPosition_->data->rot[0];

      gz_position_unlock(gazeboPosition_);
    }

    curTime = getCurrentEpochTime();
    pose.rot.yaw = ROBOT_RAD2DEG(pose.rot.yaw);
    //ROBOT_CRAMPDEG(pose.rot.yaw, 0.0, 360.0);
    ROBOT_CRAMPDEG(pose.rot.yaw, -180.0, 180.0);
    pose.extra.time = curTime;
    pose.loc.extra.time = curTime;
    pose.rot.extra.time = curTime;

    // ENDO - TO BE REVISED
    getLatestRawPose_(lastPose);
    dist = sqrt(pow((pose.loc.x - lastPose.loc.x), 2.0) + pow((pose.loc.y - lastPose.loc.y), 2.0));
    addTravelDist_(dist);

    switch(robotSpecIndex_) {
     case GAZEBO_ROBOT_CLASS_HELICOPTER:
      if (USE_GROUND_TRUTH_ == false) {
        // No odometry info available. Simulate it.
      
        if (lastTime > 0)
        {
          dt = curTime - lastTime;
          Robot::getControlSpeed_(controlSpeed);
          Robot::advancePose2D_(controlSpeed, dt);
          Robot::getPose(pose);
        } 
        lastTime = curTime;
        break; 
      }
     case GAZEBO_ROBOT_CLASS_P2AT:
     case GAZEBO_ROBOT_CLASS_P2DX:
     case GAZEBO_ROBOT_CLASS_ATRVJR:
      break;
     default:
      printfTextWindow("Error: Invalid gazebo robot class %d.\n", robotSpecIndex_);
    }
    
    setLatestRawPose_(pose);

    updatePoseFromGazeboPoseOffset_();
}

//-----------------------------------------------------------------------
// This function reads the sonar data from gazebo.
//-----------------------------------------------------------------------
void GazeboRobot::readSonarData_(void)
{
    HSPose_t pose;
    int i;

    memset(sonar_, 0x0, sizeof(sonar_));

    gz_sonar_lock(gazeboSonar_, 1);

    for (i = 0; i < numSonars_; i++)
    {
        sonar_[i] = (float)(gazeboSonar_->data->sonar_ranges[i]);
    }

    gz_sonar_unlock(gazeboSonar_);

    getPose(pose);

    for (i = 0; i < numSonars_; i++)
    {
        sonar_[i] += sonarPos_[i];

        sonarLoc_[i][0] = pose.loc.x;
        sonarLoc_[i][1] = pose.loc.y;
        sonarLoc_[i][2] = pose.rot.yaw;

        if (sonar_[i] <= VIRTUAL_BUMPER_DISTANCE_)
        {
            virtualBumper_[i] = 1.0;
        }
        else
        {
            virtualBumper_[i] = 0.0;
        }
    }
}

//-----------------------------------------------------------------------
// This function sets the offset between the raw gazebo data and the
// current robot pose.
//-----------------------------------------------------------------------
void GazeboRobot::setGazeboPoseOffset_(HSPose_t poseOffset)
{
    pthread_mutex_lock(&gazeboPoseOffsetMutex_);
    shGazeboPoseOffset_ = poseOffset;
    pthread_mutex_unlock(&gazeboPoseOffsetMutex_);
}

//-----------------------------------------------------------------------
// This function returns the offset between the raw gazebo data and the
// current robot pose.
//-----------------------------------------------------------------------
void GazeboRobot::getGazeboPoseOffset_(HSPose_t &poseOffset)
{
    pthread_mutex_lock(&gazeboPoseOffsetMutex_);
    poseOffset = shGazeboPoseOffset_;
    pthread_mutex_unlock(&gazeboPoseOffsetMutex_);
}

//-----------------------------------------------------------------------
// This function saves the latest raw pose data.
//-----------------------------------------------------------------------
void GazeboRobot::setLatestRawPose_(HSPose_t pose)
{
    pthread_mutex_lock(&latestRawPoseMutex_);
    shLatestRawPose_ = pose;
    pthread_mutex_unlock(&latestRawPoseMutex_);
}

//-----------------------------------------------------------------------
// This function returns the saved latest raw pose data.
//-----------------------------------------------------------------------
void GazeboRobot::getLatestRawPose_(HSPose_t &pose)
{
    pthread_mutex_lock(&latestRawPoseMutex_);
    pose = shLatestRawPose_;
    pthread_mutex_unlock(&latestRawPoseMutex_);
}

//-----------------------------------------------------------------------
// This function updates the robot pose from the offset.
//-----------------------------------------------------------------------
void GazeboRobot::updatePoseFromGazeboPoseOffset_(void)
{
    HSPose_t pose;
    HSPose_t poseOffset;
    double x, y, angle;

    getLatestRawPose_(pose);
    getGazeboPoseOffset_(poseOffset);

    x = pose.loc.x;
    y = pose.loc.y;

    angle = poseOffset.rot.yaw;
    ROBOT_CRAMPDEG(angle, -180.0, 180.0);
    angle = ROBOT_DEG2RAD(angle);

    pose.loc.x = (x*cos(angle)) - (y*sin(angle)) + poseOffset.loc.x;
    pose.loc.y = (x*sin(angle)) + (y*cos(angle)) + poseOffset.loc.y;
    pose.rot.yaw += poseOffset.rot.yaw;
    //ROBOT_CRAMPDEG(pose.rot.yaw, 0.0, 360.0);
    ROBOT_CRAMPDEG(pose.rot.yaw, -180.0, 180.0);

    setPose_(pose);
}

//-----------------------------------------------------------------------
// This function controls the robot with the control speed.
//-----------------------------------------------------------------------
void GazeboRobot::controlRobot_(void)
{
    HSSpeed_t controlSpeed;

    getControlSpeed_(controlSpeed);

    gz_position_lock(gazeboPosition_, 1);

    gazeboPosition_->data->cmd_vel_pos[0] = controlSpeed.driveSpeed.value;
    gazeboPosition_->data->cmd_vel_pos[1] = controlSpeed.driveSpeed.value;//0;
    gazeboPosition_->data->cmd_vel_pos[2] = controlSpeed.altitudeSpeed.value*20;
    gazeboPosition_->data->cmd_vel_rot[0] = 0;
    gazeboPosition_->data->cmd_vel_rot[1] = 0;
    gazeboPosition_->data->cmd_vel_rot[2] = ROBOT_DEG2RAD(controlSpeed.steerSpeed.value);

    gz_position_unlock(gazeboPosition_);
}

//-----------------------------------------------------------------------
// This function adds the sensors.
//-----------------------------------------------------------------------
void GazeboRobot::addSensors_(void)
{
    partXyt_ = new RobotPartXyt();

    if (REPORT_SONARS_ && (gazeboSonar_ != NULL))
    {
        partSonar_ = new RobotPartSonar(numSonars_, (float *)sonarAngles_);

        sensorSonar_ = new Sensor(
            SENSOR_SONAR,
            sizeof(sonarLoc_),
            (char*)sonarLoc_, 
            numSonars_,
            (float *)sonarAngles_,
            numSonars_,
            sonar_,
            NULL);
    }

    if (REPORT_VIRTUAL_BUMPER_ && (gazeboSonar_ != NULL))
    {
        sensorVirtualBumper_ = new Sensor(
            SENSOR_BUMPER,
            sizeof(sonarLoc_),
            (char*)sonarLoc_, 
            numSonars_,
            (float *)sonarAngles_,
            numSonars_,
            virtualBumper_,
            NULL);
    }
}

//-----------------------------------------------------------------------
// This function sets up the user interface for the robot option.
//-----------------------------------------------------------------------
void GazeboRobot::control(void)
{
    bool done = false;
    int msgData[2];

    msgData[0] = 0; // "speed_factor" is obsolete.
    msgData[1] = 0; // "angular_speed_factor" is obsolete.

    messageDrawWindow(
        EnMessageType_GAZEBO_ROBOT_CONTROL,
        EnMessageErrType_NONE,
        msgData);

    do
    {
        int c = getch();

        switch(c) {

        case 'd':
            messageHide();
            delete this;
            return;
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;

        }
    } while (!done);

    messageHide();
}

/**********************************************************************
# $Log: gazebo-robot.c,v $
# Revision 1.9  2009/09/29 20:16:57  zkira
# Attempt to fix PoseCalc hanging issue (not fixed yet)
# Added a file output flag to indicate to scripts when it has started successfully
#
# Revision 1.8  2009/01/27 20:23:37  zkira
# Turned sonar reporting off
#
# Revision 1.7  2008/10/29 02:25:22  ts2883
# Increased multiplier for altitude speed
#
# Revision 1.6  2008/10/29 00:17:37  ts2883
# Fixed hclient "error in update sensor" by disabling bumper: There is no corresponding
# BumperPart right now
#
# Revision 1.5  2008/09/24 00:29:40  ts2883
# Added GroundTruth device
#
# Revision 1.4  2008/09/18 04:15:44  ts2883
# Fixed gazebo code to compile/run
# Still does not work with scarab, but works with multiple robots/hservers otherwise
#
# Revision 1.3  2008/09/16 16:50:40  zkira
# Added class name, client ID, and server ID as paramters/constructor
# Added advancePose() odometry simulation to helicopter
# Gazebo files are untested/uncompiled so far!
#
# Revision 1.2  2008/09/09 20:42:04  zkira
# Crude altitude control (untested)
#
# Revision 1.1  2008/07/15 22:33:49  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.11  2008/01/13 00:35:40  endo
# Regular backup.
#
# Revision 1.10  2007/04/09 17:36:10  endo
# TravelDist and its logging added.
#
# Revision 1.9  2006/12/16 22:27:40  endo
# Regular backup.
#
# Revision 1.8  2006/12/12 20:45:15  endo
# Regular update.
#
# Revision 1.7  2006/12/07 20:30:22  endo
# Regular backup.
#
# Revision 1.6  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.5  2006/11/16 18:20:26  endo
# The virtual bumper added.
#
# Revision 1.4  2006/11/15 23:45:58  endo
# GazeboLaser can now accept the second laser.
#
# Revision 1.3  2006/11/14 01:04:01  endo
# Regular backup.
#
# Revision 1.2  2006/09/15 19:19:19  endo
# Regular backups.
#
# Revision 1.1  2006/07/31 22:01:14  endo
# Gazebo -> GazeboRobot
#
# Revision 1.1  2006/07/21 23:35:01  endo
# Gazebo integrated into HServer
#
#**********************************************************************/
