/**********************************************************************
 **                                                                  **
 **                       HServerConfigurator.c                      **
 **                                                                  **
 **                                                                  **
 **  Written by:  Brian Lee                                          **
 **                                                                  **
 **  Copyright 2002 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: HServerConfigurator.c,v 1.11 2010-05-03 22:12:57 zkira Exp $ */

#include <limits>
#include <ctype.h>

#include "HServerConfigurator.h"
#include "../robot.h"
#include "../gps.h"
#include "../compass.h"
#include "../gyro.h"
#include "../PoseCalculatorTypes.h"

using std::numeric_limits;

const double ClHServerConfigurator::HSCONFIG_HUGE_VAL_ = numeric_limits<double>::infinity();


ClHServerConfigurator::ClHServerConfigurator(ClHServerConfig& clConfig)
{
    m_pclConfig = &clConfig;
}

ClHServerConfigurator::~ClHServerConfigurator()
{
}

SuCognachromeInitData ClHServerConfigurator::GetCognachromeOptions(int iSection)
{
    SuCognachromeInitData suData;
    string strValue;

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnCognachromeValueType_PORT_STRING,
                                strValue) && 
         (strValue != ""))
    {
        suData.strPortString = strValue;
    }

    return suData;
}

SuCameraInitData ClHServerConfigurator::GetCameraOptions(int iSection)
{
    SuCameraInitData suData;
    string strValue;

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnCameraValueType_PORT_STRING,
                                strValue) && 
         (strValue != ""))
    {
        suData.strPortString = strValue;
    }

    return suData;
}

SuNomadInitData ClHServerConfigurator::GetNomadOptions(int iSection)
{
    SuNomadInitData suData;
    suData.enType = Nomad::EnNomadType_150;
    suData.strPortString = "";
    suData.strHost = "";

    string strValue;

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnNomadValueType_TYPE,
                                strValue) && 
         (strValue != ""))
    {
        if (strValue == "150")
        {
            suData.enType = Nomad::EnNomadType_150;
        }
        else if (strValue == "200")
        {
            suData.enType = Nomad::EnNomadType_200;
        }
        else
        {
            string strName;
            m_pclConfig->GetSectionName(iSection, strName);
            fprintf(stderr, "hserver error: bad nomad type \"%s\" for\n", strValue.c_str());
            fprintf(stderr, "    nomad section labeled: %s\n", strName.c_str());
            fprintf(stderr, "    Legal values are \"150\" and \"200\".\n");
            exit(-1);
        }
    }
    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnNomadValueType_PORT_STRING,
                                strValue) && 
         (strValue != ""))
    {
        suData.strPortString = strValue;
    }
    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnNomadValueType_HOST_NAME,
                                strValue) && 
         (strValue != ""))
    {
        suData.strHost = strValue;
    }

    return suData;
}

SuPioneerInitData ClHServerConfigurator::GetPioneerOptions(int iSection)
{
    SuPioneerInitData suData;
    suData.strPortString = "";

    string strValue;

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnPioneerValueType_PORT_STRING,
                                strValue) && 
         (strValue != ""))
    {
        suData.strPortString = strValue;
    }

    return suData;
}

SuPioneer2InitData ClHServerConfigurator::GetPioneer2Options(int iSection)
{
    SuPioneer2InitData suData;
    suData.strPortString = "";

    string strValue;

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnPioneer2ValueType_PORT_STRING,
                                strValue) && 
         (strValue != ""))
    {
        suData.strPortString = strValue;
    }

    return suData;
}

SuAmigoBotInitData ClHServerConfigurator::GetAmigoBotOptions(int iSection)
{
    SuAmigoBotInitData suData;
    suData.strPortString = "";

    string strValue;

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnAmigoBotValueType_PORT_STRING,
                                strValue) && 
         (strValue != ""))
    {
        suData.strPortString = strValue;
    }

    return suData;
}

SuFrameGrabberInitData ClHServerConfigurator::GetFrameGrabberOptions(int iSection)
{
    SuFrameGrabberInitData suData;
    string strValue;

    suData.strPortString = "";
    suData.showImage = false;


    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnFrameGrabberValueType_PORT_STRING,
                                strValue) && 
         (strValue != ""))
    {
        suData.strPortString = strValue;
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnFrameGrabberValueType_SHOW_IMAGE, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.showImage = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.showImage = false;
        }
        else
        {
            fprintf(stderr, "hserver error: a framegrabber's \"show_image\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    return suData;
}

SuCricketInitData ClHServerConfigurator::GetCricketOptions(int iSection)
{
    SuCricketInitData suData;
    suData.ID = 1;
    suData.IP = "127.0.0.1";

    string strValue;

    if (m_pclConfig->GetValue(iSection,
                                ClHServerConfig::EnCricketValueType_IP,
                                strValue) &&
         (strValue != ""))
    {
        suData.IP = strValue;
    }
    if (m_pclConfig->GetValue(iSection,
                                ClHServerConfig::EnCricketValueType_ID,
                                strValue) &&
         (strValue != ""))
    {
        suData.ID = atoi(strValue.c_str());
    }
    return suData;
}

SuGpsInitData ClHServerConfigurator::GetGpsOptions(int iSection)
{
    SuGpsInitData suData;
    suData.strPortString = "";
    suData.bUseBase = false;
    suData.dBaseLat = DEFAULT_GPS_BASE_LAT;
    suData.dBaseLong = DEFAULT_GPS_BASE_LON;
    suData.dXDiff = DEFAULT_GPS_BASE_X;
    suData.dYDiff = DEFAULT_GPS_BASE_Y;
    suData.dMPerLat = DEFAULT_GPS_M_PER_LON;
    suData.dMPerLong = DEFAULT_GPS_M_PER_LAT;

    string strValue;
    char* pcEnd;

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnGpsValueType_USE_BASE,
                                strValue) && 
         (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
             (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.bUseBase = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                  (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.bUseBase = false;
        }
        else
        {
            fprintf(stderr, "hserver error: a GPS's \"use_base\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", \"0\"\n");
            exit(-1);
        }
    }
    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnGpsValueType_PORT_STRING,
                                strValue) && 
         (strValue != ""))
    {
        suData.strPortString = strValue;
    }
    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnGpsValueType_BASE_LAT,
                                strValue) && 
         (strValue != ""))
    {
        suData.dBaseLat = strtod(strValue.c_str(), &pcEnd);
        if (*pcEnd != '\0')
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"base_lat\" value, %s, was not valid.\n", 
                     strValue.c_str());
            exit(-1);
        }
        if ((suData.dBaseLat == HSCONFIG_HUGE_VAL_) || (suData.dBaseLat == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"base_lat\" value, %s, caused an overflow.\n", 
                     strValue.c_str());
            exit(-1);
        }
    }
    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnGpsValueType_BASE_LONG,
                                strValue) && 
         (strValue != ""))
    {
        suData.dBaseLong = strtod(strValue.c_str(), &pcEnd);
        if (*pcEnd != '\0')
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"base_long\" value, %s, was not valid.\n", 
                     strValue.c_str());
            exit(-1);
        }
        if ((suData.dBaseLong == HSCONFIG_HUGE_VAL_) || (suData.dBaseLong == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"base_long\" value, %s, caused an overflow.\n", 
                     strValue.c_str());
            exit(-1);
        }
    }
    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnGpsValueType_X_DIFF,
                                strValue) && 
         (strValue != ""))
    {
        suData.dXDiff = strtod(strValue.c_str(), &pcEnd);
        if (*pcEnd != '\0')
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"x_diff\" value, %s, was not valid.\n", 
                     strValue.c_str());
            exit(-1);
        }
        if ((suData.dXDiff == HSCONFIG_HUGE_VAL_) || (suData.dXDiff == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"x_diff\" value, %s, caused an overflow.\n", 
                     strValue.c_str());
            exit(-1);
        }
    }
    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnGpsValueType_Y_DIFF,
                                strValue) && 
         (strValue != ""))
    {
        suData.dYDiff = strtod(strValue.c_str(), &pcEnd);
        if (*pcEnd != '\0')
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"y_diff\" value, %s, was not valid.\n", 
                     strValue.c_str());
            exit(-1);
        }
        if ((suData.dYDiff == HSCONFIG_HUGE_VAL_) || (suData.dYDiff == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"y_diff\" value, %s, caused an overflow.\n", 
                     strValue.c_str());
            exit(-1);
        }
    }
    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnGpsValueType_M_PER_LAT,
                                strValue) && 
         (strValue != ""))
    {
        suData.dMPerLat = strtod(strValue.c_str(), &pcEnd);
        if (*pcEnd != '\0')
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"m_per_lat\" value, %s, was not valid.\n", 
                     strValue.c_str());
            exit(-1);
        }
        if ((suData.dMPerLat == HSCONFIG_HUGE_VAL_) || (suData.dMPerLat == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"m_per_lat\" value, %s, caused an overflow.\n", 
                     strValue.c_str());
            exit(-1);
        }
    }
    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnGpsValueType_M_PER_LONG,
                                strValue) && 
         (strValue != ""))
    {
        suData.dMPerLong = strtod(strValue.c_str(), &pcEnd);
        if (*pcEnd != '\0')
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"m_per_long\" value, %s, was not valid.\n", 
                     strValue.c_str());
            exit(-1);
        }
        if ((suData.dBaseLong == HSCONFIG_HUGE_VAL_) || (suData.dBaseLong == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, 
                     "hserver error: a GPS's \"m_per_long\" value, %s, caused an overflow.\n", 
                     strValue.c_str());
            exit(-1);
        }
    }

    return suData;
}

SuJboxInitData ClHServerConfigurator::GetJboxOptions(int iSection)
{
    SuJboxInitData suData;
    suData.jboxID = 0;
    suData.disableNetworkQuery = false;
    suData.gpsSection = "";

    string strValue;

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnJboxValueType_JBOX_ID,
            strValue) && 
        (strValue != ""))
    {
        suData.jboxID = atoi(strValue.c_str());
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnJboxValueType_DISABLE_NETWORK_QUERY,
            strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
             (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.disableNetworkQuery = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                  (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.disableNetworkQuery = false;
        }
        else
        {
            fprintf(stderr, "hserver error: a jbox's \"disable_network_query\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnJboxValueType_GPS_SECTION,
            strValue) && 
        (strValue != ""))
    {
        suData.gpsSection = strValue;
    }

    return suData;
}

SuCompassInitData ClHServerConfigurator::GetCompassOptions(int iSection)
{
    SuCompassInitData suData;
    string strValue;
    string strName;

    suData.portString = "";
    suData.type = 0;

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnCompassValueType_PORT_STRING,
                                strValue) && 
         (strValue != ""))
    {
        suData.portString = strValue;
    }

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnCompassValueType_TYPE,
                                strValue) && 
         (strValue != ""))
    {
        if (strValue == "KVH-C100")
        {
            suData.type = Compass::TYPE_KVHC100;
        }
        else if (strValue == "3DM-G")
        {
            suData.type = Compass::TYPE_3DMG;
        }
        else
        {
            m_pclConfig->GetSectionName(iSection, strName);
            fprintf(stderr, "hserver error: bad compass type \"%s\" for\n", strValue.c_str());
            fprintf(stderr, "    compass section labeled: %s\n", strName.c_str());
            fprintf(stderr, "    Legal values are \"KVH-C100\" and \"3DM-G\".\n");
            exit(-1);
        }
    }

    return suData;
}

SuGyroInitData ClHServerConfigurator::GetGyroOptions(int iSection)
{
    SuGyroInitData suData;
    string strValue;
    string strName;

    suData.portString = "";
    suData.type = 0;

    if (m_pclConfig->GetValue(iSection, 
                                ClHServerConfig::EnGyroValueType_TYPE,
                                strValue) && 
         (strValue != ""))
    {
        if (strValue == "DMU-VGX")
        {
            suData.type = Gyro::TYPE_DMUVGX;
        }
        else
        {
            m_pclConfig->GetSectionName(iSection, strName);
            fprintf(stderr, "hserver error: bad gyro type \"%s\" for\n", strValue.c_str());
            fprintf(stderr, "    gyro section labeled: %s\n", strName.c_str());
            fprintf(stderr, "    Legal value is \"DMU-VGX\".\n");
            exit(-1);
        }
    }

    return suData;
}

SuPoseCalcInitData ClHServerConfigurator::GetPoseCalcOptions(int iSection)
{
    SuPoseCalcInitData suData;
    string strValue;
    string strName;
    char* pcEnd;

    memset((SuPoseCalcInitData *)&suData, 0x0, sizeof(SuPoseCalcInitData));

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnPoseCalcValueType_FUSER_TYPE,
            strValue) && 
        (strValue != ""))
    {
        if (strValue == "maxconf")
        {
            suData.fuserType = POSECALC_FUSER_TYPE_MAXMUM_CONFIDENCE;
        }
        else if (strValue == "ekf")
        {
            suData.fuserType = POSECALC_FUSER_TYPE_EXTENDED_KALMAN_FILTER;
        }
        else if (strValue == "pfilter")
        {
            suData.fuserType = POSECALC_FUSER_TYPE_PARTICLE_FILTER;
        }
        else
        {
            m_pclConfig->GetSectionName(iSection, strName);
            fprintf(stderr, "hserver error: bad posecal fuser type \"%s\" for\n", strValue.c_str());
            fprintf(stderr, "    posecalc section labeled: %s\n", strName.c_str());
            fprintf(stderr, "    Legal values are \"maxconf\", \"ekf\" and \"pfilter\".\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnPoseCalcValueType_VARIANCE_SCHEME,
            strValue) && 
        (strValue != ""))
    {
        if (strValue == "dynamic")
        {
            suData.varianceScheme = POSECALC_VARIANCE_SCHEME_DYNAMIC;
        }
        else if (strValue == "static")
        {
            suData.varianceScheme = POSECALC_VARIANCE_SCHEME_STATIC;
        }
        else
        {
            m_pclConfig->GetSectionName(iSection, strName);
            fprintf(stderr, "hserver error: bad posecal variance scheme \"%s\" for\n", strValue.c_str());
            fprintf(stderr, "    posecalc section labeled: %s\n", strName.c_str());
            fprintf(stderr, "    Legal values are \"dynamic\", and \"static\".\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnPoseCalcValueType_COMPASS2MLAB_HEADING_FACTOR,
            strValue) && 
        (strValue != ""))
    {
        suData.compass2MlabHeadingFacor = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (compass2mlab_headingFacor).\n");
            exit(-1);
        }
        if ((suData.compass2MlabHeadingFacor == HSCONFIG_HUGE_VAL_) || (suData.compass2MlabHeadingFacor == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (compass2mlab_headingFacor).\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnPoseCalcValueType_COMPASS2MLAB_HEADING_OFFSET,
            strValue) && 
        (strValue != ""))
    {
        suData.compass2MlabHeadingOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (compass2mlab_headingFacor).\n");
            exit(-1);
        }
        if ((suData.compass2MlabHeadingOffset == HSCONFIG_HUGE_VAL_) || (suData.compass2MlabHeadingOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (compass2mlab_headingFacor).\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnPoseCalcValueType_COMPASS_MAX_VALID_ANGSPEED4HEADING,
            strValue) && 
        (strValue != ""))
    {
        suData.compassMaxValidAngSpeed4Heading = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (compass_max_valid_angspeed4heading).\n");
            exit(-1);
        }
        if ((suData.compassMaxValidAngSpeed4Heading == HSCONFIG_HUGE_VAL_) || (suData.compassMaxValidAngSpeed4Heading == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (compass_max_valid_angspeed4heading).\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnPoseCalcValueType_GPS2MLAB_HEADING_FACTOR,
            strValue) && 
        (strValue != ""))
    {
        suData.gps2MlabHeadingFacor = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (gps2mlab_headingFacor).\n");
            exit(-1);
        }
        if ((suData.gps2MlabHeadingFacor == HSCONFIG_HUGE_VAL_) || (suData.gps2MlabHeadingFacor == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (gps2mlab_headingFacor).\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnPoseCalcValueType_GPS2MLAB_HEADING_OFFSET,
            strValue) && 
        (strValue != ""))
    {
        suData.gps2MlabHeadingOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (gps2mlab_headingFacor).\n");
            exit(-1);
        }
        if ((suData.gps2MlabHeadingOffset == HSCONFIG_HUGE_VAL_) || (suData.gps2MlabHeadingOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (gps2mlab_headingFacor).\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnPoseCalcValueType_GPS_MIN_VALID_TRANSSPEED4HEADING,
            strValue) && 
        (strValue != ""))
    {
        suData.gpsMinValidTransSpeed4Heading = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (gps_min_valid_transspeed4heading).\n");
            exit(-1);
        }
        if ((suData.gpsMinValidTransSpeed4Heading == HSCONFIG_HUGE_VAL_) || (suData.gpsMinValidTransSpeed4Heading == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (gps_min_valid_transspeed4heading).\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnPoseCalcValueType_GPS_MAX_VALID_ANGSPEED4HEADING,
            strValue) && 
        (strValue != ""))
    {
        suData.gpsMaxValidAngSpeed4Heading = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (gps_max_valid_angspeed4heading).\n");
            exit(-1);
        }
        if ((suData.gpsMaxValidAngSpeed4Heading == HSCONFIG_HUGE_VAL_) || (suData.gpsMaxValidAngSpeed4Heading == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, "hserver error: invalid value in the RC file (gps_max_valid_angspeed4heading).\n");
            exit(-1);
        }
    }

    return suData;
}

SuLaserInitData ClHServerConfigurator::GetLaserOptions(int iSection)
{
    SuLaserInitData suData;
    suData.strPort = "";
    suData.bSend = false;
    suData.bListen = false;
    suData.fAngleOffset = 0.0;
    suData.fXOffset = 0.0;
    suData.fYOffset = 0.0;
    suData.strName = "";
    suData.strStreamAddress = "";

    string strValue = "";

    char* pcEnd;

    if (m_pclConfig->GetValue(iSection, 
                              ClHServerConfig::EnLaserValueType_NAME, 
                              strValue) && 
        (strValue != ""))
    {
        suData.strName = strValue;
    }

    if (m_pclConfig->GetValue(iSection, 
                              ClHServerConfig::EnLaserValueType_PORT_STRING, 
                              strValue) && 
        (strValue != ""))
    {
        suData.strPort = strValue;
    }

    if (m_pclConfig->GetValue(iSection, 
                              ClHServerConfig::EnLaserValueType_ANGLE_OFFSET, 
                              strValue) && 
        (strValue != ""))
    {
        suData.fAngleOffset = strtod(strValue.c_str(), &pcEnd);
        if (*pcEnd != '\0')
        {
            fprintf(stderr, 
                    "hserver error: a laser's \"angle_offset\" value, %s, was not valid.\n", 
                    strValue.c_str());
            exit(-1);
        }
        if ((suData.fAngleOffset == HSCONFIG_HUGE_VAL_) || (suData.fAngleOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, 
                    "hserver error: a laser's \"angle_offset\" value, %s, caused an overflow.\n", 
                    strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(iSection, 
                              ClHServerConfig::EnLaserValueType_X_OFFSET, 
                              strValue) && 
        (strValue != ""))
    {
        suData.fXOffset = strtod(strValue.c_str(), &pcEnd);
        if (*pcEnd != '\0')
        {
            fprintf(stderr, 
                    "hserver error: a laser's \"x_offset\" value, %s, was not valid.\n", 
                    strValue.c_str());
            exit(-1);
        }
        if ((suData.fXOffset == HSCONFIG_HUGE_VAL_) || (suData.fXOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, 
                    "hserver error: a laser's \"x_offset\" value, %s, caused an overflow.\n", 
                    strValue.c_str());
            exit(-1);
        }
        suData.fXOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(iSection, 
                              ClHServerConfig::EnLaserValueType_Y_OFFSET, 
                              strValue) && 
        (strValue != ""))
    {
        suData.fYOffset = strtod(strValue.c_str(), &pcEnd);
        if (*pcEnd != '\0')
        {
            fprintf(stderr, 
                    "hserver error: a laser's \"y_offset\" value, %s, was not valid.\n", 
                    strValue.c_str());
            exit(-1);
        }
        if ((suData.fYOffset == HSCONFIG_HUGE_VAL_) || (suData.fYOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(stderr, 
                    "hserver error: a GPS's \"y_offset\" value, %s, caused an overflow.\n", 
                    strValue.c_str());
            exit(-1);
        }
        suData.fYOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(iSection, 
                              ClHServerConfig::EnLaserValueType_LISTEN_IPT, 
                              strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.bListen = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.bListen = false;
        }
        else
        {
            fprintf(stderr, "hserver error: a laser's \"listen_ipt\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnLaserValueType_SEND_IPT, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.bSend = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.bSend = false;
        }
        else
        {
            fprintf(stderr, "hserver error: a laser's \"send_ipt\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(iSection, 
                              ClHServerConfig::EnLaserValueType_STREAM_HOST, 
                              strValue) && 
        (strValue != ""))
    {
        suData.strStreamAddress = strValue;
    }

    return suData;
}

SuRobotInitData ClHServerConfigurator::GetRobotOptions(int iSection)
{
    SuRobotInitData suData;
    string strValue, strName;
    char* pcEnd = NULL;

    suData.minPointTurnAngle 	  = Robot::DEFAULT_MIN_POINT_TURN_ANGLE_;
    suData.maxSpeed 		  = Robot::DEFAULT_MAX_SPEED_;
    suData.maxSteerSpeed 	  = Robot::DEFAULT_MAX_STEER_SPEED_;
    suData.maxSteerSpeedAngle 	  = Robot::DEFAULT_MAX_STEER_SPEED_ANGLE_;
    suData.maxAltitudeSpeed 	  = Robot::DEFAULT_MAX_ALTITUDE_SPEED_;
    suData.maxNoCommandTimeSec 	  = Robot::DEFAULT_MAX_NO_COMMAND_TIME_SEC_;
    suData.minSteerSpeed 	  = Robot::DEFAULT_MIN_STEER_SPEED_;
    suData.stoppingSpeedThreshold = Robot::DEFAULT_STOPPING_SPEED_THRESHOLD_;
    suData.numVelocitySamples 	  = Robot::DEFAULT_NUM_VELOCITY_SAMPLES_;

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRobotValueType_MIN_POINT_TURN_ANGLE, 
            strValue) && 
        (strValue != ""))
    {
        suData.minPointTurnAngle = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Robot's \"min_point_turn_angle\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.minPointTurnAngle == HSCONFIG_HUGE_VAL_) || (suData.minPointTurnAngle == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Robot's \"min_point_turn_angle\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection,
            ClHServerConfig::EnRobotValueType_MAX_SPEED,
            strValue) &&
        (strValue != ""))
    {
        suData.maxSpeed = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max_speed\" value, %s, was not valid.\n",
                strValue.c_str());
            exit(-1);
        }

        if ((suData.maxSpeed == HSCONFIG_HUGE_VAL_) || (suData.maxSpeed == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max_speed\" value, %s, caused an overflow.\n",
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection,
            ClHServerConfig::EnRobotValueType_MAX_STEER_SPEED,
            strValue) &&
        (strValue != ""))
    {
        suData.maxSteerSpeed = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max_speed\" value, %s, was not valid.\n",
                strValue.c_str());
            exit(-1);
        }

        if ((suData.maxSteerSpeed == HSCONFIG_HUGE_VAL_) || (suData.maxSteerSpeed == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max__speed\" value, %s, caused an overflow.\n",
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection,
            ClHServerConfig::EnRobotValueType_MAX_STEER_SPEED_ANGLE,
            strValue) &&
        (strValue != ""))
    {
        suData.maxSteerSpeedAngle = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max_steer_speed_angle\" value, %s, was not valid.\n",
                strValue.c_str());
            exit(-1);
        }

        if ((suData.maxSteerSpeedAngle == HSCONFIG_HUGE_VAL_) || (suData.maxSteerSpeedAngle == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max_steer_speed_angle\" value, %s, caused an overflow.\n",
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection,
            ClHServerConfig::EnRobotValueType_MAX_ALTITUDE_SPEED,
            strValue) &&
        (strValue != ""))
    {
        suData.maxAltitudeSpeed = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max_altitude_speed\" value, %s, was not valid.\n",
                strValue.c_str());
            exit(-1);
        }

        if ((suData.maxAltitudeSpeed == HSCONFIG_HUGE_VAL_) || (suData.maxAltitudeSpeed == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max_altitude_speed\" value, %s, caused an overflow.\n",
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection,
            ClHServerConfig::EnRobotValueType_MAX_NO_COMMAND_TIME_SEC,
            strValue) &&
        (strValue != ""))
    {
        suData.maxNoCommandTimeSec = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max_no_command_time_sec\" value, %s, was not valid.\n",
                strValue.c_str());
            exit(-1);
        }

        if ((suData.maxNoCommandTimeSec == HSCONFIG_HUGE_VAL_) || (suData.maxNoCommandTimeSec == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"max_no_command_time_sed\" value, %s, caused an overflow.\n",
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection,
            ClHServerConfig::EnRobotValueType_MIN_STEER_SPEED,
            strValue) &&
        (strValue != ""))
    {
        suData.minSteerSpeed = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"min_steer_speed\" value, %s, was not valid.\n",
                strValue.c_str());
            exit(-1);
        }

        if ((suData.minSteerSpeed == HSCONFIG_HUGE_VAL_) || (suData.minSteerSpeed == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"min_steer_speed\" value, %s, caused an overflow.\n",
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection,
            ClHServerConfig::EnRobotValueType_STOPPING_SPEED_THRESHOLD,
            strValue) &&
        (strValue != ""))
    {
        suData.stoppingSpeedThreshold = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"stopping_speed_threshold\" value, %s, was not valid.\n",
                strValue.c_str());
            exit(-1);
        }

        if ((suData.stoppingSpeedThreshold == HSCONFIG_HUGE_VAL_) || (suData.stoppingSpeedThreshold == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr,
                "hserver error: a Robot's \"stopping_speed_threshold\" value, %s, caused an overflow.\n",
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRobotValueType_NUM_VELOCITY_SAMPLES,
            strValue) && 
        (strValue != ""))
    {
        if (!isdigit(strValue[0]))
        {
            fprintf(
                stderr, 
                "hserver error: a robot's \"numVelocitySamples\" value, %s, is not a number.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.numVelocitySamples = atoi(strValue.c_str());

        if ((suData.numVelocitySamples == HSCONFIG_HUGE_VAL_) || (suData.numVelocitySamples == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a robot's \"numVelocitySamples\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    return suData;
}

SuGazeboInitData ClHServerConfigurator::GetGazeboOptions(int iSection)
{
    SuGazeboInitData suData;
    string strValue, strName;
    char* pcEnd = NULL;

    suData.name = "";
    suData.truthName = "";
    suData.robot_class = "";
    suData.type = -1;
    suData.clientID = 0;
    suData.serverID = 0;
    suData.angleOffset = 0;
    suData.xOffset = 0;
    suData.yOffset = 0;
    suData.showImage = false;
    suData.laserPitchAngle = 0;

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnGazeboValueType_NAME, 
            strValue) && 
        (strValue != ""))
    {
        suData.name = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnGazeboValueType_TYPE, 
            strValue) && 
        (strValue != ""))
    {
        if (strValue == "robot")
        {
            suData.type = HS_GAZEBO_TYPE_ROBOT;
        }
        else if (strValue == "laser")
        {
            suData.type = HS_GAZEBO_TYPE_LASER;
        }
        else if (strValue == "video")
        {
            suData.type = HS_GAZEBO_TYPE_VIDEO;
        }
        else
        {
            m_pclConfig->GetSectionName(iSection, strName);
            fprintf(stderr, "hserver error: bad Gazebo type \"%s\" for\n", strValue.c_str());
            fprintf(stderr, "    Gazebo section labeled: %s\n", strName.c_str());
            fprintf(stderr, "    Legal values are \"robot\", and \"laser\".\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnGazeboValueType_ROBOT_CLASS,
            strValue) && 
        (strValue != ""))
    {
        suData.robot_class = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnGazeboValueType_CLIENT_ID,
            strValue) && 
        (strValue != ""))
    {
        if (!isdigit(strValue[0]))
        {
            fprintf(
                stderr, 
                "hserver error: a gazebo's \"clientID\" value, %s, is not a number.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.clientID == HSCONFIG_HUGE_VAL_) || (suData.clientID == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a gazebo's \"clientID\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.clientID = atoi(strValue.c_str());
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnGazeboValueType_SERVER_ID,
            strValue) && 
        (strValue != ""))
    {
        if (!isdigit(strValue[0]))
        {
            fprintf(
                stderr, 
                "hserver error: a gazebo's \"serverID\" value, %s, is not a number.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.serverID == HSCONFIG_HUGE_VAL_) || (suData.serverID == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a gazebo's \"port\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.serverID = atoi(strValue.c_str());
    }  
    
    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnGazeboValueType_ANGLE_OFFSET, 
            strValue) && 
        (strValue != ""))
    {
        suData.angleOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a gazebo's \"angle_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.angleOffset == HSCONFIG_HUGE_VAL_) || (suData.angleOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a gazebo's \"angle_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnGazeboValueType_X_OFFSET, 
            strValue) && 
        (strValue != ""))
    {
        suData.xOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a gazebo's \"x_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.xOffset == HSCONFIG_HUGE_VAL_) || (suData.xOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a gazebo's \"x_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.xOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnGazeboValueType_Y_OFFSET, 
            strValue) && 
        (strValue != ""))
    {
        suData.yOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a gazebo's \"y_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.yOffset == HSCONFIG_HUGE_VAL_) || (suData.yOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a GPS's \"y_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.yOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(
            iSection,
            ClHServerConfig::EnGazeboValueType_LASER_PITCH_ANGLE,
            strValue) &&
        (strValue != ""))
    {
        suData.laserPitchAngle = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr,
                "hserver error: a usarsim's \"laserPitchAngle\" value, %s, was not valid.\n",
                strValue.c_str());
            exit(-1);
        }

        if ((suData.laserPitchAngle == HSCONFIG_HUGE_VAL_) || (suData.laserPitchAngle == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr,
                "hserver error: a usarsim's \"laserPitchAngle\" value, %s, caused an overflow.\n",
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnGazeboValueType_SHOW_IMAGE, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.showImage = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.showImage = false;
        }
        else
        {
            fprintf(stderr, "hserver error: Gazebo's \"show_image\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    return suData;
}

SuRovioInitData ClHServerConfigurator::GetRovioOptions(int iSection)
{
    SuRovioInitData suData;
    string strValue, strName;
    char* pcEnd = NULL;

    suData.host = "";
    suData.name = "";
    suData.robot_class = "";
    suData.port = -1;
    suData.type = -1;
    suData.angleOffset = 0;
    suData.pos_x = 0;
    suData.pos_y = 0;
    suData.pos_z = 0;
    suData.rot_yaw = 0;
    suData.rot_pitch = 0;
    suData.rot_roll = 0;
    suData.xOffset = 0;
    suData.yOffset = 0;
    suData.showImage = false;

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_HOST,
            strValue) && 
        (strValue != ""))
    {
        suData.host = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_NAME,
            strValue) && 
        (strValue != ""))
    {
        suData.name = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_ROBOT_CLASS,
            strValue) && 
        (strValue != ""))
    {
        suData.robot_class = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_PORT,
            strValue) && 
        (strValue != ""))
    {
        if (!isdigit(strValue[0]))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"port\" value, %s, is not a number.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.port = atoi(strValue.c_str());

        if ((suData.port == HSCONFIG_HUGE_VAL_) || (suData.port == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"port\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_TYPE, 
            strValue) && 
        (strValue != ""))
    {
        if (strValue == "robot")
        {
            suData.type = HS_ROVIO_TYPE_ROBOT;
        }
        else if (strValue == "video")
        {
            suData.type = HS_ROVIO_TYPE_VIDEO;
        }
        else
        {
            m_pclConfig->GetSectionName(iSection, strName);
            fprintf(stderr, "hserver error: bad Rovio type \"%s\" for\n", strValue.c_str());
            fprintf(stderr, "    Rovio section labeled: %s\n", strName.c_str());
            fprintf(stderr, "    Legal values are \"robot\", and \"laser\".\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_ANGLE_OFFSET, 
            strValue) && 
        (strValue != ""))
    {
        suData.angleOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"angle_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.angleOffset == HSCONFIG_HUGE_VAL_) || (suData.angleOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"angle_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_POS_X, 
            strValue) && 
        (strValue != ""))
    {
        suData.pos_x = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"pos_x\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.pos_x == HSCONFIG_HUGE_VAL_) || (suData.pos_x == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"pos_x\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_POS_Y, 
            strValue) && 
        (strValue != ""))
    {
        suData.pos_y = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"pos_y\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.pos_y == HSCONFIG_HUGE_VAL_) || (suData.pos_y == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"pos_y\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_POS_Z, 
            strValue) && 
        (strValue != ""))
    {
        suData.pos_z = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"pos_z\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.pos_z == HSCONFIG_HUGE_VAL_) || (suData.pos_z == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"pos_z\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_ROT_YAW, 
            strValue) && 
        (strValue != ""))
    {
        suData.rot_yaw = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"rot_yaw\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.rot_yaw == HSCONFIG_HUGE_VAL_) || (suData.rot_yaw == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"rot_yaw\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_ROT_PITCH, 
            strValue) && 
        (strValue != ""))
    {
        suData.rot_pitch = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"rot_pitch\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.rot_pitch == HSCONFIG_HUGE_VAL_) || (suData.rot_pitch == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"rot_pitch\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_ROT_ROLL, 
            strValue) && 
        (strValue != ""))
    {
        suData.rot_roll = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a RoviRovioo's \"rot_roll\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.rot_roll == HSCONFIG_HUGE_VAL_) || (suData.rot_roll == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"rot_roll\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_X_OFFSET,
            strValue) && 
        (strValue != ""))
    {
        suData.xOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"x_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.xOffset == HSCONFIG_HUGE_VAL_) || (suData.xOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"x_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.xOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnRovioValueType_Y_OFFSET, 
            strValue) && 
        (strValue != ""))
    {
        suData.yOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Rovio's \"y_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.yOffset == HSCONFIG_HUGE_VAL_) || (suData.yOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a GPS's \"y_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.yOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnRovioValueType_SHOW_IMAGE, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.showImage = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.showImage = false;
        }
        else
        {
            fprintf(stderr, "hserver error: Rovio's \"show_image\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

	 
    return suData;
}



SuSkybotixInitData ClHServerConfigurator::GetSkybotixOptions(int iSection)
{
    SuSkybotixInitData suData;
    string strValue, strName;
    char* pcEnd = NULL;

    suData.host = "";
    suData.name = "";
    suData.robot_class = "";
    suData.port = -1;
    suData.type = -1;
    suData.angleOffset = 0;
    suData.pos_x = 0;
    suData.pos_y = 0;
    suData.pos_z = 0;
    suData.rot_yaw = 0;
    suData.rot_pitch = 0;
    suData.rot_roll = 0;
    suData.xOffset = 0;
    suData.yOffset = 0;
    suData.showImage = false;

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_HOST,
            strValue) && 
        (strValue != ""))
    {
        suData.host = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_NAME,
            strValue) && 
        (strValue != ""))
    {
        suData.name = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_ROBOT_CLASS,
            strValue) && 
        (strValue != ""))
    {
        suData.robot_class = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_PORT,
            strValue) && 
        (strValue != ""))
    {
        if (!isdigit(strValue[0]))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"port\" value, %s, is not a number.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.port = atoi(strValue.c_str());

        if ((suData.port == HSCONFIG_HUGE_VAL_) || (suData.port == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"port\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_TYPE, 
            strValue) && 
        (strValue != ""))
    {
        if (strValue == "robot")
        {
            suData.type = HS_ROVIO_TYPE_ROBOT;
        }
        else if (strValue == "video")
        {
            suData.type = HS_ROVIO_TYPE_VIDEO;
        }
        else
        {
            m_pclConfig->GetSectionName(iSection, strName);
            fprintf(stderr, "hserver error: bad Skybotix type \"%s\" for\n", strValue.c_str());
            fprintf(stderr, "    Skybotix section labeled: %s\n", strName.c_str());
            fprintf(stderr, "    Legal values are \"robot\", and \"laser\".\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_ANGLE_OFFSET, 
            strValue) && 
        (strValue != ""))
    {
        suData.angleOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"angle_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.angleOffset == HSCONFIG_HUGE_VAL_) || (suData.angleOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"angle_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_POS_X, 
            strValue) && 
        (strValue != ""))
    {
        suData.pos_x = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"pos_x\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.pos_x == HSCONFIG_HUGE_VAL_) || (suData.pos_x == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"pos_x\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_POS_Y, 
            strValue) && 
        (strValue != ""))
    {
        suData.pos_y = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"pos_y\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.pos_y == HSCONFIG_HUGE_VAL_) || (suData.pos_y == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"pos_y\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_POS_Z, 
            strValue) && 
        (strValue != ""))
    {
        suData.pos_z = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"pos_z\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.pos_z == HSCONFIG_HUGE_VAL_) || (suData.pos_z == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"pos_z\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_ROT_YAW, 
            strValue) && 
        (strValue != ""))
    {
        suData.rot_yaw = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"rot_yaw\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.rot_yaw == HSCONFIG_HUGE_VAL_) || (suData.rot_yaw == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"rot_yaw\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_ROT_PITCH, 
            strValue) && 
        (strValue != ""))
    {
        suData.rot_pitch = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"rot_pitch\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.rot_pitch == HSCONFIG_HUGE_VAL_) || (suData.rot_pitch == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"rot_pitch\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_ROT_ROLL, 
            strValue) && 
        (strValue != ""))
    {
        suData.rot_roll = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a RoviSkybotixo's \"rot_roll\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.rot_roll == HSCONFIG_HUGE_VAL_) || (suData.rot_roll == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"rot_roll\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_X_OFFSET,
            strValue) && 
        (strValue != ""))
    {
        suData.xOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"x_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.xOffset == HSCONFIG_HUGE_VAL_) || (suData.xOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"x_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.xOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnSkybotixValueType_Y_OFFSET, 
            strValue) && 
        (strValue != ""))
    {
        suData.yOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a Skybotix's \"y_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.yOffset == HSCONFIG_HUGE_VAL_) || (suData.yOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a GPS's \"y_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.yOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnSkybotixValueType_SHOW_IMAGE, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.showImage = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.showImage = false;
        }
        else
        {
            fprintf(stderr, "hserver error: Skybotix's \"show_image\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

	 
    return suData;
}


SuUSARSimInitData ClHServerConfigurator::GetUSARSimOptions(int iSection)
{
    SuUSARSimInitData suData;
    string strValue, strName;
    char* pcEnd = NULL;

    suData.host = "";
    suData.name = "";
    suData.robot_class = "";
    suData.port = -1;
    suData.type = -1;
    suData.angleOffset = 0;
    suData.pos_x = 0;
    suData.pos_y = 0;
    suData.pos_z = 0;
    suData.rot_yaw = 0;
    suData.rot_pitch = 0;
    suData.rot_roll = 0;
    suData.xOffset = 0;
    suData.yOffset = 0;
    suData.showImage = false;

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_HOST,
            strValue) && 
        (strValue != ""))
    {
        suData.host = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_NAME,
            strValue) && 
        (strValue != ""))
    {
        suData.name = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_ROBOT_CLASS,
            strValue) && 
        (strValue != ""))
    {
        suData.robot_class = strValue;
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_PORT,
            strValue) && 
        (strValue != ""))
    {
        if (!isdigit(strValue[0]))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"port\" value, %s, is not a number.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.xOffset == HSCONFIG_HUGE_VAL_) || (suData.xOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"x_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.port = atoi(strValue.c_str());
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_TYPE, 
            strValue) && 
        (strValue != ""))
    {
        if (strValue == "robot")
        {
            suData.type = HS_USARSIM_TYPE_ROBOT;
        }
        else if (strValue == "laser")
        {
            suData.type = HS_USARSIM_TYPE_LASER;
        }
        else if (strValue == "video")
        {
            suData.type = HS_USARSIM_TYPE_VIDEO;
        }
        else
        {
            m_pclConfig->GetSectionName(iSection, strName);
            fprintf(stderr, "hserver error: bad USARSim type \"%s\" for\n", strValue.c_str());
            fprintf(stderr, "    USARSim section labeled: %s\n", strName.c_str());
            fprintf(stderr, "    Legal values are \"robot\", and \"laser\".\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_ANGLE_OFFSET, 
            strValue) && 
        (strValue != ""))
    {
        suData.angleOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"angle_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.angleOffset == HSCONFIG_HUGE_VAL_) || (suData.angleOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"angle_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_POS_X, 
            strValue) && 
        (strValue != ""))
    {
        suData.pos_x = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"pos_x\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.pos_x == HSCONFIG_HUGE_VAL_) || (suData.pos_x == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"pos_x\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_POS_Y, 
            strValue) && 
        (strValue != ""))
    {
        suData.pos_y = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"pos_y\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.pos_y == HSCONFIG_HUGE_VAL_) || (suData.pos_y == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"pos_y\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_POS_Z, 
            strValue) && 
        (strValue != ""))
    {
        suData.pos_z = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"pos_z\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.pos_z == HSCONFIG_HUGE_VAL_) || (suData.pos_z == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"pos_z\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_ROT_YAW, 
            strValue) && 
        (strValue != ""))
    {
        suData.rot_yaw = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"rot_yaw\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.rot_yaw == HSCONFIG_HUGE_VAL_) || (suData.rot_yaw == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"rot_yaw\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_ROT_PITCH, 
            strValue) && 
        (strValue != ""))
    {
        suData.rot_pitch = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"rot_pitch\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.rot_pitch == HSCONFIG_HUGE_VAL_) || (suData.rot_pitch == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"rot_pitch\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_ROT_ROLL, 
            strValue) && 
        (strValue != ""))
    {
        suData.rot_roll = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"rot_roll\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.rot_roll == HSCONFIG_HUGE_VAL_) || (suData.rot_roll == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"rot_roll\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_X_OFFSET,
            strValue) && 
        (strValue != ""))
    {
        suData.xOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"x_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.xOffset == HSCONFIG_HUGE_VAL_) || (suData.xOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"x_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.xOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_Y_OFFSET, 
            strValue) && 
        (strValue != ""))
    {
        suData.yOffset = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"y_offset\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.yOffset == HSCONFIG_HUGE_VAL_) || (suData.yOffset == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a GPS's \"y_offset\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }

        suData.yOffset /= 100.0; // convert cm to meters
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnUSARSimValueType_SHOW_IMAGE, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.showImage = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.showImage = false;
        }
        else
        {
            fprintf(stderr, "hserver error: USARSim's \"show_image\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(
            iSection, 
            ClHServerConfig::EnUSARSimValueType_LASER_PITCH_ANGLE, 
            strValue) && 
        (strValue != ""))
    {
        suData.laserPitchAngle = strtod(strValue.c_str(), &pcEnd);

        if (*pcEnd != '\0')
        {
            fprintf(
                stderr, 
                "hserver error: a USARSim's \"laserPitchAngle\" value, %s, was not valid.\n", 
                strValue.c_str());
            exit(-1);
        }

        if ((suData.laserPitchAngle == HSCONFIG_HUGE_VAL_) || (suData.laserPitchAngle == -HSCONFIG_HUGE_VAL_))
        {
            fprintf(
                stderr, 
                "hserver error: a usarsim's \"laserPitchAngle\" value, %s, caused an overflow.\n", 
                strValue.c_str());
            exit(-1);
        }
    }
    return suData;
}

SuOpenCVInitData ClHServerConfigurator::GetOpenCVOptions(int iSection)
{
    SuOpenCVInitData suData;
    string strValue, strName;

    suData.trackBlobs = false;
    suData.detectFace = false;
    suData.showImage = false;

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnOpenCVValueType_TRACK_BLOBS, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.trackBlobs = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.trackBlobs = false;
        }
        else
        {
            fprintf(stderr, "hserver error: OpenCV's \"track_blobs\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnOpenCVValueType_DETECT_FACE, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.detectFace = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.detectFace = false;
        }
        else
        {
            fprintf(stderr, "hserver error: OpenCV's \"detect_face\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnOpenCVValueType_DETECT_OBJECTS, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.detectObjects = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.detectObjects = false;
        }
        else
        {
            fprintf(stderr, "hserver error: OpenCV's \"detect_objects\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnOpenCVValueType_DIRECT_GRAB, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.directGrab = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.directGrab = false;
        }
        else
        {
            fprintf(stderr, "hserver error: OpenCV's \"direct_grab\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    if (m_pclConfig->GetValue(iSection, ClHServerConfig::EnOpenCVValueType_SHOW_IMAGE, strValue) && 
        (strValue != ""))
    {
        if ((strcasecmp(strValue.c_str(), "true") == 0) || 
            (strcasecmp(strValue.c_str(), "1") == 0))
        {
            suData.showImage = true;
        }
        else if ((strcasecmp(strValue.c_str(), "false") == 0) || 
                 (strcasecmp(strValue.c_str(), "0") == 0))
        {
            suData.showImage = false;
        }
        else
        {
            fprintf(stderr, "hserver error: OpenCV's \"show_image\" value had an invalid value.\n");
            fprintf(stderr, "  Valid values: \"true\", \"false\", \"1\", and \"0\"\n");
            exit(-1);
        }
    }

    return suData;
}

/**********************************************************************
# $Log: HServerConfigurator.c,v $
# Revision 1.11  2010-05-03 22:12:57  zkira
# Addition of rovio & cricket
#
# Revision 1.9  2009/08/28 19:48:16  zkira
# Bug in Rovio configuration parsing
#
# Revision 1.8  2009/04/17 17:50:05  zkira
# Added max_speed to differentiate it from max_steer_speed
#
# Revision 1.7  2009/04/13 02:58:54  zkira
# max altitude speed & other robot parameters
#
# Revision 1.6  2009/03/02 18:47:07  zkira
# Added ability to change low level robot parameters (e.g. max speed) in the .hserverrc file.
# Added Rovio
#
# Revision 1.5  2009/01/27 20:21:17  zkira
# Added gazebo laser_pitch parameter
#
# Revision 1.4  2008/11/20 03:53:57  ts2883
# The pitch angle information for the laser can now be parsed and read by the
# hserver using .hserverrc file.
#
# Revision 1.3  2008/09/16 16:47:37  zkira
# Added Gazebo robot class, clientID, and serverID to configuration file
#
# Revision 1.2  2008/07/15 22:34:01  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.10  2008/04/13 00:33:42  endo
# Regular backup.
#
# Revision 1.9  2008/04/12 20:17:24  endo
# OpenCV face detection implemented.
#
# Revision 1.8  2008/01/13 00:35:40  endo
# Regular backup.
#
# Revision 1.7  2008/01/09 22:54:03  endo
# USARSimRobot added.
#
# Revision 1.6  2006/12/01 21:14:54  endo
# Regular backup.
#
# Revision 1.5  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.4  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
# Revision 1.3  2006/11/15 23:45:58  endo
# GazeboLaser can now accept the second laser.
#
# Revision 1.2  2006/08/08 21:33:49  endo
# GazeboLaser added.
#
# Revision 1.1.1.1  2006/07/21 17:42:16  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:57  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/05 21:35:54  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2006/05/17 16:46:15  endo
# HUGE_VAL replaced by HSCONFIG_HUGE_VAL
#
# Revision 1.1.1.1  2005/02/06 22:59:45  endo
# AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
#
# Revision 1.3  2003/04/26 18:31:06  ebeowulf
# Checked in Pioneer2-dxe module.
#
# Revision 1.2  2002/02/18 13:52:47  endo
# AmigoBot added.
#
#
#**********************************************************************/
