/**********************************************************************
 **                                                                  **
 **                         opencv_handler.h                         **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Copyright 2006 - 2008 Georgia Tech Research Corporation         **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: opencv_handler.h,v 1.1 2008-07-15 22:33:51 endo Exp $ */

#ifndef OPENCV_HANDLER_H
#define OPENCV_HANDLER_H

#include <pthread.h>
#include <cvaux.h>
#include <highgui.h>
#include <string>

#include "module.h"
#include "video.h"
#include "sensors.h"
#include "xwindow.h"

using std::string;

class OpenCVHandler : public Module {

    typedef enum ImageProcessTypes_t
    {
        IMAGE_PROCESS_TYPE_BLOB_TRACK,
        IMAGE_PROCESS_TYPE_FACE_DETECT,
        IMAGE_PROCESS_TYPE_LETTER_RECOG,
        IMAGE_PROCESS_TYPE_OBJECTS_DETECT,
        NUM_IMAGE_PROCESS_TYPES
    };

    typedef enum ObjectsDetectObjectTypes_t
    {
        OBJECTS_DETECT_OBJECT_TYPE_001,
        OBJECTS_DETECT_OBJECT_TYPE_002,
        OBJECTS_DETECT_OBJECT_TYPE_003,
        OBJECTS_DETECT_OBJECT_TYPE_004,
        NUM_OBJECTS_DETECT_OBJECT_TYPES
    };

    typedef enum FGDetectorTypes_t
    {
        FG_DETECTOR_ACM_MM2003,
        FG_DETECTOR_ACM_MM2003_SIMPLE,
        FG_DETECTOR_CVPR1999,
        NUM_FG_DETECTOR_TYPES
    };

    typedef enum BlobDetectorTypes_t
    {
        BLOB_DETECTOR_CC,
        BLOB_DETECTOR_SIMPLE,
        NUM_BLOB_DETECTOR_TYPES
    };

    typedef enum BlobTrackerTypes_t
    {
        BLOB_TRACKER_CCMSPF,
        BLOB_TRACKER_CC,
        BLOB_TRACKER_MS,
        BLOB_TRACKER_MSFG,
        BLOB_TRACKER_MSPF,
        NUM_BLOB_TRACKER_TYPES
    };

    typedef enum BlobTrajectoryAnalyzerTypes_t
    {
        BLOB_TRAJECTORY_ANALYZER_NONE,
        BLOB_TRAJECTORY_ANALYZER_HISTPVS,
        BLOB_TRAJECTORY_ANALYZER_HISTP,
        BLOB_TRAJECTORY_ANALYZER_HISTPV,
        BLOB_TRAJECTORY_ANALYZER_HISTSS,
        BLOB_TRAJECTORY_ANALYZER_TRACKDIST,
        BLOB_TRAJECTORY_ANALYZER_IOR,
        NUM_BLOB_TRAJECTORY_ANALYZER_TYPES
    };

    typedef enum RGBColorTypes_t
    {
        RGB_COLOR_RED,
        RGB_COLOR_GREEN,
        RGB_COLOR_BLUE,
        NUM_RGB_COLOR_TYPES
    };

    typedef enum SensorTypes_t
    {
        OPENCV_SENSOR_BLOB_SIZE,
        OPENCV_SENSOR_BLOB_COLOR,
        OPENCV_SENSOR_FACE_SIZE,
        OPENCV_SENSOR_OBJECT_001,
        OPENCV_SENSOR_OBJECT_002,
        OPENCV_SENSOR_OBJECT_003,
        OPENCV_SENSOR_OBJECT_004,
        NUM_OPENCV_SENSOR_TYPES
    };

    typedef struct SensorExtraBlobSize_t
    {
        int numBlobSizeTypes;
    };

    typedef struct SensorExtraBlobColor_t
    {
        int numBlobColorTypes;
    };


    typedef struct SensorExtraFaceSize_t
    {
        int numFaceSizeTypes;
    };

    typedef struct SensorExtraObject_t
    {
        int numObjectTypes;
    };

    typedef struct RGBColor255_t
    {
        int red;
        int green;
        int blue;
    };

protected:
    Xwindow *xwindows_[NUM_IMAGE_PROCESS_TYPES];
    Sensor *sensors_[NUM_OPENCV_SENSOR_TYPES];
    IplImage *rawImage_;
    CvBlobTrackerAuto *blobTracker_;
    CvBlobTrackerAutoParam1 blobTrackerParam_;
	CvCapture *directGrabSource_;
    CvFont blobTrackerFont_;
    CvFont faceDetectorFont_;
    CvFont objectDetectorsFont_;
    CvHaarClassifierCascade *faceDetector_;
    CvHaarClassifierCascade **objectDetectors_;
    CvMemStorage *faceMemStorage_;
    CvMemStorage *objectMemStorages_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    SensorExtraBlobSize_t sensorExtraBlobSize_;
    SensorExtraBlobColor_t sensorExtraBlobColor_;
    SensorExtraFaceSize_t sensorExtraFaceSize_;
    SensorExtraObject_t sensorExtraObject_;
    pthread_t handlerThread_;
    pthread_t displayThread_;
    IplImage *shOutputImages_[NUM_IMAGE_PROCESS_TYPES];
    double fieldOfView_;
    double rawImageTimeStamp_;
    float *blobSizeAngles_;
    float *blobSizeReadings_;
    float *blobColorAngles_;
    float *blobColorReadings_;
    float *faceSizeAngles_;
    float *faceSizeReadings_;
    float *objectAngles_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    float *objectReadings_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    bool enableImageProcess_[NUM_IMAGE_PROCESS_TYPES];
    bool resetImageProcess_[NUM_IMAGE_PROCESS_TYPES];
    bool directGrab_;
    bool showImage_;
    bool handlerThreadIsUp_;
    bool displayThreadIsUp_;
    bool videoIsAvailable_;
    
    static pthread_mutex_t outputImagesMutex_[NUM_IMAGE_PROCESS_TYPES];
    static pthread_mutex_t sensorDataMutex_[NUM_OPENCV_SENSOR_TYPES];

    static const struct timespec HANDLER_SLEEPTIME_NSEC_;
    static const struct timespec DISPLAY_SLEEPTIME_NSEC_;
    static const RGBColor255_t FACE_DETECT_DISPLAY_COLOR_;
    static const RGBColor255_t OBJECTS_DETECT_DISPLAY_COLOR_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    static const string IMAGE_PROCESS_TYPE_NAMES_[NUM_IMAGE_PROCESS_TYPES];
    static const string FG_DETECTOR_NICKNAMES_[NUM_FG_DETECTOR_TYPES];
    static const string BLOB_DETECTOR_NICKNAMES_[NUM_BLOB_DETECTOR_TYPES];
    static const string BLOB_TRACKER_NICKNAMES_[NUM_BLOB_TRACKER_TYPES];
    static const string BLOB_TRAJECTORY_ANALYZER_NICKNAMES_[NUM_BLOB_TRAJECTORY_ANALYZER_TYPES];
    static const string FACE_DETECT_DATA_XML_FILE_;
    static const string FACE_DETECT_LABEL_;
    static const string OBJECTS_DETECT_DATA_XML_FILES_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    static const string OBJECTS_DETECT_LABELS_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    static const char RGB_CHARS_[NUM_RGB_COLOR_TYPES];
    static const double BLOB_TRACKER_RESET_TIME_SEC_;
    static const double DEFAULT_FIELD_OF_VIEW_;
    static const float DEFAULT_FONT_HSCALE_;
    static const float DEFAULT_FONT_VSCALE_;
    static const float DEFAULT_FONT_ITALICSCALE_;
    static const float DEFAULT_OBJECTS_DETECT_READING_VALUE_;
    static const float FACE_DETECT_SCALE_FACTOR_;
    static const float OBJECTS_DETECT_SCALE_FACTORS_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    static const int SKIP_STATUSBAR_UPDATE_;
    static const int WATCHDOG_CHECK_INTERVAL_SEC_;
    static const int DEFAULT_DIRECT_GRAB_CAM_INDEX_;
    static const int DEFAULT_DIRECT_GRAB_HEIGHT_;
    static const int DEFAULT_DIRECT_GRAB_WIDTH_;
    static const int DEFAULT_FG_TRAIN_FRAMES_;
    static const int DEFAULT_FG_DETECTOR_TYPE_;
    static const int DEFAULT_BLOB_DETECTOR_TYPE_;
    static const int DEFAULT_BLOB_TRACKER_TYPE_;
    static const int DEFAULT_BLOB_TRAJECTORY_ANALYZER_TYPE_;
    static const int DEFAULT_FONT_THICKNESS_;
    static const int DEFAULT_FONT_LINETYPE_;
    static const int FACE_DETECT_DISPLAY_THICKNESS_;
    static const int FACE_DETECT_MIN_HEIGHT_;
    static const int FACE_DETECT_MIN_NUMBER_;
    static const int FACE_DETECT_MIN_WIDTH_;
    static const int OBJECTS_DETECT_DISPLAY_THICKNESS_;
    static const int OBJECTS_DETECT_MIN_NUMBER_;
    static const int OBJECTS_DETECT_MIN_HEIGHTS_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    static const int OBJECTS_DETECT_MIN_WIDTHS_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    static const int MAX_REPORTING_BLOBS_;
    static const int MAX_REPORTING_FACES_;
    static const int MAX_REPORTING_OBJECTS_[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    static const bool OBJECTS_DETECT_REPORT_SIZE_;
    static const bool SORT_REPORTING_BLOB_SIZE_;
    static const bool SORT_REPORTING_FACE_SIZE_;
    static const bool SORT_REPORTING_0BJECTS_;

    CvFGDetector *createOpenCVFGDetector_(int fgDetectorType);
    CvBlobDetector *createOpenCVBlobDetector_(int blobDetectorType);
    CvBlobTracker *createOpenCVBlobTracker_(int blobTrackerType);
    CvBlobTrackAnalysis *createBlobTrajectoryAnalyzer_(int blobTrajectoryAnalyzerType);
    CvScalar getBlobColor_(const IplImage *rawImage, const CvBlob *blob);
    void addSensors_(void);
    void handlerLoop_(void);
    void displayLoop_(void);
    void processImage_(void);
    void processBlobTrack_(IplImage *rawImage);
    void processFaceDetect_(IplImage *rawImage);
    void processObjectsDetect_(IplImage *rawImage);
    void updateStatusBar_(void);
    void showOutputImages_(void);
    void hideOutputImages_(void);
    bool initializeRawImage_(int width, int height);
    bool initializeBlobTracker_(IplImage *rawImage);
    bool initializeFaceDetector_(IplImage *rawImage);
    bool initializeObjectDetectors_(IplImage *rawImage);

    static void *startHandlerThread_(void *openCVHandlerInstance);
    static void *startDisplayThread_(void *videoInstance);

public:
    OpenCVHandler(
        OpenCVHandler** a,
        bool trackBlobs,
        bool detectFace,
        bool detectObjects,
        bool directGrab,
        bool showImage);
    ~OpenCVHandler(void);
    void control(void);
};

extern OpenCVHandler *gOpenCVHandler;

#endif
/**********************************************************************
# $Log: opencv_handler.h,v $
# Revision 1.1  2008-07-15 22:33:51  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.19  2008/04/27 02:41:10  endo
# Regular backup.
#
# Revision 1.18  2008/04/23 18:01:12  endo
# Regular backup.
#
# Revision 1.17  2008/04/18 21:46:36  endo
# Regular backup.
#
# Revision 1.16  2008/04/17 01:29:06  endo
# Pioneer2 cleaned up a bit.
#
# Revision 1.15  2008/04/15 00:07:20  endo
# Regular backup.
#
# Revision 1.14  2008/04/13 00:33:42  endo
# Regular backup.
#
# Revision 1.13  2008/04/12 20:39:52  endo
# Regular backup.
#
# Revision 1.12  2008/04/12 20:17:24  endo
# OpenCV face detection implemented.
#
# Revision 1.11  2007/01/19 21:07:54  endo
# Regular backup.
#
# Revision 1.10  2006/12/17 04:42:54  endo
# SENSOR_OPENCV_BLOB_COLOR added.
#
# Revision 1.9  2006/12/07 20:30:23  endo
# Regular backup.
#
# Revision 1.8  2006/12/01 21:14:54  endo
# Regular backup.
#
# Revision 1.7  2006/11/30 17:51:55  endo
# Regular update.
#
# Revision 1.6  2006/11/29 23:13:28  endo
# OpenCV improved.
#
# Revision 1.5  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.4  2006/11/24 23:17:14  endo
# OpenCV BlobTrack improved.
#
# Revision 1.3  2006/11/22 22:24:28  endo
# OpenCV BlobTrack implemented.
#
# Revision 1.2  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.1  2006/11/20 22:55:52  endo
# OpenCV added.
#
#**********************************************************************/
