/**********************************************************************
 **                                                                  **
 **                             pioneer2.c                           **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **  Modified by: Eric Martinson (for DXE robot)                     **
 **                                                                  **
 **  Pioneer2 robot driver for HServer                               **
 **                                                                  **
 **  Copyright 2000 - 2008, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: pioneer2.c,v 1.2 2008-07-15 22:33:52 endo Exp $ */

#include <stdio.h>
#include <termio.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <ctype.h>

#include "HServerTypes.h"
#include "message.h"
#include "pioneer2.h"
#include "hserver.h"
#include "camera.h"
#include "statusbar.h"
#include "robot_config.h"
#include "sensors.h"

//---------------------------------------------------------------------
// Constants
//---------------------------------------------------------------------
#define AngleConvFactor      0.0057       // radians per encoder count diff 
#define DAngleConvFactor     0.0879       // deg per encoder count diff 
//#define AngleConvFactor      0.0057       // radians per encoder count diff 
//#define DAngleConvFactor     0.3267       // deg per encoder count diff 
#define DistConvFactor       0.969        // (mm/count)
#define RobotRadius          330.0        // radius in mm
#define RobotDiagonal        120.0        // half-height to diagonal of octagon
#define Holonomic            1            // turns in own radius 
#define MaxRVelocity         100.0        // degrees per second
#define MaxVelocity          500.0        // mm per second
#define RangeConvFactor      0.268        // sonar range mm per 2 usec tick

#define ShaftDegPerRealDeg   180.0/160.0  // fudge factor

#define PIONEER2_PTZ 42
#define PIONEER2_ARGINT  0x3B
#define PIONEER2_ARGNINT 0x1B
#define PIONEER2_ARGSTR  0x2B

#define VISIONIO_COM 100
#define VISIONIOpac 0x80
#define VISIONpac 0x50
#define FRAME24pac 0xb0
#define DISTpac 0xd0

#define BLOB_MODE 0
#define BLOB_BB_MODE 2
#define LINE_MODE 1

#define TURNACCURACY 2
#define SONARMAX 10000
#define SONARHOLDTIME 10
#define IR_DISTANCE 290

//float pioneer2_range_angle[] = {90, 50, 30, 10, -10, -30, -50, -90, -90, -130, -150, -170, 170, 150, 130, 90 }; 

float pioneer2_range_angle_IR[] = { 60, 300 };

//float pioneer2_range_loc[PIONEER2_NUM_SONAR] = 
//{  0.18, 0.18, 0.18, 0.18, 0.18, 0.18, 0.18, 0.18, 0.20, 0.23, 0.24, 0.24, 0.24, 0.24, 0.23, 0.20 };

int numIRs2 = 0;
// bool gbAlreadySending = false;

const float Pioneer2::MAX_SONAR_READABLE_DISTANCE_ = 5.0;
const float Pioneer2::SONAR_ANGLES_[] = {90, 50, 30, 10, -10, -30, -50, -90, -90, -130, -150, -170, 170, 150, 130, 90};
const float Pioneer2::SONAR_POS_[] = {0.18, 0.18, 0.18, 0.18, 0.18, 0.18, 0.18, 0.18, 0.20, 0.23, 0.24, 0.24, 0.24, 0.24, 0.23, 0.20};
const int Pioneer2::CMD_SYNC0_ = 0;
const int Pioneer2::CMD_SYNC1_ = 1;
const int Pioneer2::CMD_SYNC2_ = 2;
const int Pioneer2::CMD_PULSE_ = 0;
const int Pioneer2::CMD_OPEN_ = 1;
const int Pioneer2::CMD_CLOSE_ = 2;
const int Pioneer2::CMD_POLLING_ = 3;
const int Pioneer2::CMD_ENABLE_ = 4;
const int Pioneer2::CMD_SETO_ = 7;
const int Pioneer2::CMD_VEL_ = 11;
const int Pioneer2::CMD_HEAD_ = 12;
const int Pioneer2::CMD_DHEAD_ = 13;
const int Pioneer2::CMD_SAY_ = 15;
const int Pioneer2::CMD_RVEL_ = 21;
const int Pioneer2::CMD_DIGOUT_ = 30;
const int Pioneer2::MAX_ENCODER_VALUE_ = 65535;
const int Pioneer2::READER_LOOP_USLEEP_ = 1000;
const int Pioneer2::READ_BACKOFF_USEC_ = 100;
const int Pioneer2::SKIP_STATUSBAR_UPDATE_ = 10;
const int Pioneer2::SYNCER_LOOP_USLEEP_ = 500000;
const bool Pioneer2::START_SONAR_ = true;
const bool Pioneer2::ENABLE_IR_ = false;
const bool Pioneer2::ENABLE_COMPASS_ = false;

//---------------------------------------------------------------------
// Constructor
//---------------------------------------------------------------------
Pioneer2::Pioneer2(
    Robot** a,
    const string &strPortString) :
    Robot(a, HS_ROBOT_TYPE_PIONEER2, HS_MODULE_NAME_ROBOT_PIONEER2),
    sensorSonar(NULL),
    sensorIr(NULL),
    partSonar(NULL),
    partIr(NULL),
    partXyt(NULL),
    partCompass(NULL),
    isSyncing_(false),
    isSynced_(false),
    isReading_(false),
    isRangeOn_(START_SONAR_)
{
    int num_irs = PIONEER2_NUM_IR;

    robotName = NULL;
    pthread_mutex_init(&serialMutex, NULL);
    numSonars_ = PIONEER2_NUM_SONAR;
    numIRs2 = num_irs;

    if ((report_level > HS_REP_LEV_NONE) && (report_level != HS_REP_LEV_DEBUG))
    {
        printTextWindow("Connecting to Pioneer2... ", false);
    }

    if (report_level >= HS_REP_LEV_DEBUG)
    {
        printTextWindow("Connecting to Pioneer2");
        printfTextWindow("robot serial port: %s", strPortString.c_str());
        printfTextWindow("%d sonar sensors + %d infared sensors", numSonars_, num_irs);
    }

    updateStatusBar_();
    setupSerial(&commFd_, strPortString.c_str(), B9600, report_level, "Robot", 1);
    sleep(1);
    sendSync_(CMD_SYNC2_);

    flushSerialLine_();
    sleep(1);

    if ((report_level > HS_REP_LEV_NONE) && (report_level != HS_REP_LEV_DEBUG))
    {
        printTextWindow("connected", true);
    }
    else if (report_level > HS_REP_LEV_NONE)
    {
        printTextWindow("Pioneer2 connected", true);
    }

    pthread_create(&readerThread_, NULL, &startReaderThread_, (void*)this);

    addSensors();

    runSyncer_();
}

//---------------------------------------------------------------------
// Destructor
//---------------------------------------------------------------------
Pioneer2::~Pioneer2(void)
{
    if (isSynced_)
    {
        sendSync_(CMD_CLOSE_);
        sendSync_(CMD_CLOSE_);
    }

    if (isSyncing_)
    {
        pthread_cancel(syncerThread_);
        pthread_join(syncerThread_, NULL);
    }

    isSyncing_ = false;

    if (isReading_)
    {
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    if (sensorSonar != NULL)
    {
        delete sensorSonar;
        sensorSonar = NULL;
    }

    if (sensorIr != NULL)
    {
        delete sensorIr;
        sensorIr = NULL;
    }

    if (sensorCompass != NULL)
    {
        delete sensorCompass;
        sensorCompass = NULL;
    }

    if (partSonar != NULL)
    {
        delete partSonar;
        partSonar = NULL;
    }

    if (partIr != NULL)
    {
        delete partIr;
        partIr = NULL;
    }

    if (partXyt != NULL)
    {
        delete partXyt;
        partXyt = NULL;
    }

    if (partCompass != NULL)
    {
        delete partCompass;
        partCompass = NULL;
    }

    printTextWindow("Pioneer2 disconnected");
    close(commFd_);

    if(robotName != NULL)
    {
        free(robotName);
        free(robotClass);
        free(subclass);
    }

    redrawWindows();
}

//---------------------------------------------------------------------
// This function flushes the buffer in the serial port.
//---------------------------------------------------------------------
void Pioneer2::flushSerialLine_(void)
{
    byte_t buf[100];
    int cnum;
  
    do
    {
        cnum=read(commFd_, buf, 100);
        if (cnum == -1)
        {
            cnum = 0;
        }
    } while(cnum);
}


//---------------------------------------------------------------------
// This function reads the next byte from the fd
//---------------------------------------------------------------------
byte_t Pioneer2::nextByte_(void)
{
    int cnum;
    byte_t byte;
  
    while(true)
    {
        cnum = read(commFd_, &byte, 1);
        if (cnum)
        {
            break;
        }
        // NEED TO SLEEP, DO NOT REMOVE
        usleep(READ_BACKOFF_USEC_);
    }
    return byte;
}

//---------------------------------------------------------------------
// This function reads the next remaining bytes from the fd
//---------------------------------------------------------------------
int Pioneer2::nextRemainingBytes_(byte_t *bytes, int max)
{
    int cnum;
    int numTries = 25;

    do
    {
        cnum = read(commFd_, bytes, max);
        if (cnum == 0)
        {
            numTries--;
            if (numTries == 0)
            {
                return -1;
            }
        }
        if (cnum == -1)
        {
            return -1;
        }
    } while ((cnum == 0) || (cnum == -1));
    return cnum;
}

void Pioneer2::getWord(char buf[])
{
    int i = 0;
    byte_t nchar;

    do
    {
        nchar = nextByte_();
        if ((nchar != 191) && (nchar != ' ') && (nchar != 10))
        {
            buf[i++] = nchar;
        }
    } while ((nchar != 191) && (nchar != ' ') && (nchar != 0));
    buf[i] = 0;
}

int Pioneer2::round(double n)
{
    if(n < 0)
    { 
        return((int)(n - 0.5));
    }
    else 
    {
        return((int)(n + 0.5));
    }
}

void Pioneer2::set_nibbles(int pos, byte_t buf[])
{
    buf[0] = (pos & 0xf000) >> 12;
    buf[1] = (pos & 0x0f00) >> 8;
    buf[2] = (pos & 0x00f0) >> 4;
    buf[3] = (pos & 0x000f) >> 0;
}

int Pioneer2::calc_chksum(byte_t* ptr)
{
    int n;
    int c = 0;
    n = *(ptr++);
    n -= 2;
    while(n > 1)
    {
        c += (*(ptr) << 8) | *(ptr + 1);
        c = c & 0xffff;
        n -= 2;
        ptr += 2;
    }
    if (n > 0)
    {
        c = c ^ ((int) *(ptr++));
    }
    return c;
}

int Pioneer2::shortAngle(int ang1, int ang2)
{
    int a, b;

    if (ang1 < ang2) {
        a = ang2 - ang1;
        b = 360 - ang2 + ang1;
    }
    else {
        a = ang1 - ang2;
        b = 360 - ang1 + ang2;
    }
    return (a < b) ? a : b;
}

//---------------------------------------------------------------------
// This function updates the status bar
//---------------------------------------------------------------------
void Pioneer2::updateStatusBar_()
{
    HSPose_t pose;
    char buf[100];

    statusStr[0] = 0;
  
    strcat(statusStr, "Pioneer2: ");

    sprintf(buf, "%c ", statusbarSpinner_->getStatus());
    strcat(statusStr, buf);

    if (isSyncing_)
    {
        strcat(statusStr, "Syncing");
    }
    else if(!isSynced_)
    {
        strcat(statusStr, "Not synced");
    }
    else
    {      
        getPose(pose);

        sprintf(
            buf,
            " %.2fV  x: %.2f  y: %.2f  t: %.2f ",
            battery,
            pose.loc.x,
            pose.loc.y,
            pose.rot.yaw);

        strcat(statusStr, buf);
    }
    statusbar->update(statusLine);
}

//---------------------------------------------------------------------
// This function sends a packet to the communication port.
//---------------------------------------------------------------------
int Pioneer2::sendPack_(byte_t pack[])
{
    int checksum;
    byte_t header[2];
    byte_t chksum[2];

    header[0] = 0xFA;
    header[1] = 0xFB;
    checksum = calc_chksum(pack);
    chksum[0] = (checksum & 0x0000ff00) >> 8;
    chksum[1] = checksum & 0x000000ff;

    pthread_cleanup_push((void(*)(void*)) pthread_mutex_unlock, (void*) &serialMutex);
    pthread_mutex_lock(&serialMutex);
    write(commFd_, header, 2);
    write(commFd_, pack, pack[0] - 1);
    write(commFd_, chksum, 2);
    pthread_cleanup_pop(1);
    return 0;
}

//---------------------------------------------------------------------
// This function sends a sync command to the robot.
//---------------------------------------------------------------------
int Pioneer2::sendSync_(int s)
{
    byte_t pack[6];
  
    pack[0] = 3;
    pack[1] = s;
    sendPack_(pack);
    if (report_level >= HS_REP_LEV_DEBUG)
    {
        printfTextWindow("Send Sync: %d", s);
    }
    return 0;
}


int Pioneer2::sendString(char str[], byte_t command)
{
    byte_t pack[255];
    int i;

    pack[0] = 5 + strlen(str);
    pack[1] = command;
    pack[2] = PIONEER2_ARGSTR;
    pack[3] = strlen(str);
    for(i = 0; i < (int) strlen(str); i++)
    {
        pack[4 + i] = str[i];
    }
    sendPack_(pack);
    return 0;
}

int Pioneer2::sendStringN(char str[], int len, byte_t command)
{
    byte_t pack[255];
    int i;

    pack[0] = 5 + len;
    pack[1] = command;
    pack[2] = PIONEER2_ARGSTR;
    pack[3] = len;
    for(i = 0; i < len; i++)
    {
        pack[4 + i] = str[i];
    }
    sendPack_(pack);
    return 0;
}  

int Pioneer2::sendInt_(int sint, byte_t command)
{
    byte_t pack[40];
  
    pack[0] = 6;
    pack[1] = command;
    if (sint < 0)
    {
        pack[2] = PIONEER2_ARGNINT;
        sint = abs(sint);
    }
    else pack[2] = PIONEER2_ARGINT;
    pack[4] = (sint&0x0000ff00) >> 8;
    pack[3] = sint&0x000000ff;
    sendPack_(pack);
    return 0;
}

//---------------------------------------------------------------------
// This function updates the robot info.
//---------------------------------------------------------------------
void Pioneer2::updateRobotInfo_(RobotInfo_t robotInfo)
{
    HSPose_t pose;
    HSPose_t deltaPose;
    static RobotInfo_t oldRobotInfo;
    static int battery_warn = 0;
    int i, xdif, ydif;
    double tdif;
    double tmpSonar;

    xdif = (int)((robotInfo.xPos - oldRobotInfo.xPos) * DistConvFactor);
    ydif = (int)((robotInfo.yPos - oldRobotInfo.yPos) * DistConvFactor);
    tdif = (double(robotInfo.thetaPos) - (double)(oldRobotInfo.thetaPos)) * DAngleConvFactor;

    if (xdif < -500)
    {
        xdif = (int)((MAX_ENCODER_VALUE_ - oldRobotInfo.xPos + robotInfo.xPos) * DistConvFactor);
    }
    if (xdif > 500)
    {
        xdif = -(int)((MAX_ENCODER_VALUE_ - robotInfo.xPos + oldRobotInfo.xPos) * DistConvFactor);
    }
    if (ydif < -500)
    {
        ydif = (int)((MAX_ENCODER_VALUE_ - oldRobotInfo.yPos + robotInfo.yPos) * DistConvFactor);
    }
    if (ydif > 500)
    {
        ydif = -(int)((MAX_ENCODER_VALUE_ - robotInfo.yPos + oldRobotInfo.yPos) * DistConvFactor);
    }

    ROBOT_CRAMPDEG(tdif, -200.0, 200.0);

    // Account for a reset robot
    if ((robotInfo.xPos == 0) && (robotInfo.yPos == 0) && (robotInfo.thetaPos == 0))
    {
        xdif = 0;
        ydif = 0;
        tdif = 0;
    }

    memset((HSPose_t *)&deltaPose, 0x0, sizeof(HSPose_t));
    deltaPose.loc.x = (double)xdif / (1000.0);
    deltaPose.loc.y = (double)ydif / (1000.0);
    deltaPose.rot.yaw = tdif;
    deltaPose.extra.time = getCurrentEpochTime();

    advancePose_(deltaPose);
    getPose(pose);

    if (ENABLE_COMPASS_)
    {
        compass[0] = (double)(robotInfo.compass * 2);
    }

    for(i = 0; i < numSonars_; i++)
    {
        if (robotInfo.sonarRange[i] > 0)
        {
            sonar[i] = (robotInfo.sonarRange[i] * RangeConvFactor) / 1000.0;

            // If the distance being read was bigger than sonar's detectable range, mark it as
            // no reading. Otherwise, adjust the readings by considering the width of the robot.
            if (sonar[i] > MAX_SONAR_READABLE_DISTANCE_)
            {
                sonar[i] = SONAR_READING_OUTSIDE_RANGE;
            }
            else
            {
                sonar[i] += SONAR_POS_[i];
            }

            sonar_loc[i][0] = pose.loc.x;
            sonar_loc[i][1] = pose.loc.y;
            sonar_loc[i][2] = pose.rot.yaw;
        }
    }

    battery = (double)(robotInfo.battery) / 10.0;

    if (battery < 10.8)
    {
        if ((battery_warn % 30) == 0)
        {
            printTextWindow("Warning: Battery low");
            printf("\a");
        }
        battery_warn++;
    }

    if (ENABLE_IR_ && (numIRs2 > 0))
    {
        // read the IR values as long as this isn't our
        // first robotInfo received
        int i, which = 0x2, base;
        base = numSonars_ - numIRs2;
        for (i = 0; i < numIRs2; i++)
        {
            if (robotInfo.digin & which)
            {
                tmpSonar = SONARMAX;
            }
            else
            {
                tmpSonar = IR_DISTANCE;
            }
            ir[i] = (double) (tmpSonar) / 1000.0;
            which = which << 1;
        } 

        sendInt_(0x0202, CMD_DIGOUT_); // hold the Vin high
        usleep(3000);                    // wait 6 milliseconds
        sendInt_(0x0200, CMD_DIGOUT_); // pull the Vin low
        // we need to wait atleast 56 milliseconds so
        // wait till the next robotInfo
    }

    oldRobotInfo = robotInfo;
}

//---------------------------------------------------------------------
// This function enables the sonar
//---------------------------------------------------------------------
void Pioneer2::sonarStart(void)
{
    char polling[20] = {8, 7, 6, 5, 4, 3, 2, 1, 16, 15, 14, 13 ,12 ,11, 10, 9, 0}; // ERIC

    sendString(polling, CMD_POLLING_);

    if (report_level >= HS_REP_LEV_DEBUG)
    { 
        printTextWindow("Pioneer2 sonar start");
    }

    isRangeOn_ = true;
}

//---------------------------------------------------------------------
// This function disables the sonar
//---------------------------------------------------------------------
void Pioneer2::sonarStop(void)
{
    char polling[12] = {0};

    sendString(polling, CMD_POLLING_);

    if (report_level >= HS_REP_LEV_DEBUG)
    { 
        printTextWindow("Pioneer2 sonar stop");
    }

    isRangeOn_ = false;
}

//---------------------------------------------------------------------
// This function decodes the Sync to a packet
//---------------------------------------------------------------------
void Pioneer2::decodeSync2Pack(byte_t pack[])
{
    isSyncing_ = false;
    isSynced_ = true;
    redrawWindows();
    robotName = strdup((char*) &(pack[2]));
    robotClass = strdup((char*) &(pack[3 + strlen(robotName)]));
    subclass = strdup((char*) &(pack[4 + strlen(robotName) + strlen(robotClass)]));
    if (report_level >= HS_REP_LEV_DEBUG)
    {
        printfTextWindow("Sync Completed  Name: %s  Class: %s  Subclass: %s", 
                          robotName, robotClass, subclass);
    }

    sendSync_(CMD_OPEN_);  
    //sonarStop();
    //sendInt_(1, (byte_t)CMD_ENABLE_);
    sendInt_(0x0101, CMD_DIGOUT_);  
}

//---------------------------------------------------------------------
// This function decodes the status packet.
//---------------------------------------------------------------------
void Pioneer2::decodeStatusPack(byte_t pack[])
{
    static RobotInfo_t robotInfo;
    int num, i, base;

    robotInfo.status = pack[1];
    robotInfo.xPos = ((int) pack[3] << 8) | (int) pack[2];
    robotInfo.yPos = ((int) pack[5] << 8) | (int) pack[4];
    robotInfo.thetaPos = ((int) pack[7] << 8) | (int) pack[6];
    robotInfo.leftVel  = ((int) pack[9] << 8) | (int) pack[8];
    robotInfo.rightVel  = ((int) pack[11] << 8) | (int) pack[10];
    robotInfo.battery = pack[12];
    robotInfo.leftBumper = pack[13];
    robotInfo.rightBumper = pack[14];
    robotInfo.control = ((int) pack[16] << 8) | (int) pack[15];
    robotInfo.PTU = ((int) pack[18] << 8) | (int) pack[17];
    robotInfo.compass = pack[19];
    robotInfo.sonarReadings = pack[20];

    memset(robotInfo.sonarRange, 0x0, sizeof(unsigned int)*numSonars_);

    if (robotInfo.sonarReadings > numSonars_)
    {
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printTextWindow("Bad packet recieved");
        }
        return;
    }

    for(i = 0; i < robotInfo.sonarReadings; i++)
    {
        num = pack[21+ (i*3)];
        robotInfo.sonarRange[num] = ((int) pack[23 + i * 3] << 8) | (int) pack[22 + i * 3];
    }

    base = 21 + (robotInfo.sonarReadings * 3);
    robotInfo.inputTimer = ((int) pack[base + 1] << 8) | (int) pack[base];
    robotInfo.analog = pack[base + 2];
    robotInfo.digin  = pack[base + 3];
    robotInfo.digout = pack[base + 4];

    updateRobotInfo_(robotInfo);
    updateStatusBar_();
}

//---------------------------------------------------------------------
// This function decodes the packet content.
//---------------------------------------------------------------------
int Pioneer2::decodePack(byte_t pack[])
{
    if (report_level >= HS_REP_LEV_DEBUG)
    {
        printTextWindow("pioneer packet: ", false);
    }
    switch(pack[1]) {

    case 0:
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printTextWindow("SYNC0", true);
        }
        sendSync_(1);  
        break;

    case 1:
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printTextWindow("SYNC1", true);
        }
        sendSync_(2);  
        break;
 
   case 2:
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printTextWindow("SYNC2", true);
        }
        decodeSync2Pack(pack);
        break;

    case VISIONpac:
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printTextWindow("VISIONpac", true);
        }
        //decodeVisionPack(pack);
        break;

    case VISIONIOpac:
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printTextWindow("VISIONIOpac", true);
        }
        //decodeVisionIoPack(pack);
        break;

    case FRAME24pac:
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printTextWindow("FRAME24pac", true);
        }
        //decodeFrame24Pack(pack);
        break;

    case DISTpac:
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printTextWindow("DISTpac", true);
        }
        break;

    case 0x32:
    case 0x33:
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printTextWindow("Status Packet", true);
        }
        decodeStatusPack(pack);
        break;

    default:
        if (report_level >= HS_REP_LEV_DEBUG)
        {
            printfTextWindow("Unknown: %2x", pack[1]);
        }
        break;
    }
    return pack[1];
}

//-----------------------------------------------------------------------
// This function controls the robot with the control speed.
//-----------------------------------------------------------------------
void Pioneer2::controlRobot_(void)
{
    HSSpeed_t controlSpeed;

    if (!isSynced_)
    {
        return;
    }

    getControlSpeed_(controlSpeed);

    // Note: Drive command is in mm/s.
	sendInt_((int)(ROBOT_M2MM(controlSpeed.driveSpeed.value)), CMD_VEL_);
    sendInt_((int)(controlSpeed.steerSpeed.value), CMD_RVEL_);
}

//---------------------------------------------------------------------
// This function starts the syncer thread
//---------------------------------------------------------------------
void *Pioneer2::startSyncerThread_(void *pioneer2Instance)
{
    ((Pioneer2*) pioneer2Instance)->syncerLoop_();
    return NULL;
}


//---------------------------------------------------------------------
// This function runs the syncer in a loop
//---------------------------------------------------------------------
void Pioneer2::syncerLoop_(void)
{
    while(true)
    {
        if (isSynced_)
        {
            sleep(1);

            // Turn on sonars.
            if (isRangeOn_)
            {
                sonarStart();
            }
            else
            {
                sonarStop();
            }

            sleep(1);

            // Turn on the motor.
            sendInt_(1, (byte_t)CMD_ENABLE_);

            pthread_exit(0);
        }

        usleep(SYNCER_LOOP_USLEEP_);
        sendSync_(CMD_PULSE_);
    }
}

//---------------------------------------------------------------------
// This function runs the syncer
//---------------------------------------------------------------------
void Pioneer2::runSyncer_(void)
{
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    if (report_level >= HS_REP_LEV_DEBUG)
    {
        printTextWindow("Pioneer2: Sync");
    }

    isSyncing_ = true;
    isSynced_ = false;

    updateStatusBar_();

    pthread_create(&syncerThread_, &attr, &startSyncerThread_, (void*)this);
}

//---------------------------------------------------------------------
// This function intialize the x, y, and theta
//---------------------------------------------------------------------
void Pioneer2::initRobot(void)
{
    //resetPose_();
    sendInt_(0, CMD_SETO_);
}

//---------------------------------------------------------------------
// This function reads a packet from the port.
//---------------------------------------------------------------------
int Pioneer2::readPack_()
{
    int count, checksum, rchecksum, n, len;
    byte_t pack[1024], hd[2], *data = NULL;

    n = nextRemainingBytes_(hd, 2);

    if (n == -1)
    {
        return -1;	// Read timeout in testing
    }

    if (n < 2)
    {
        n = nextRemainingBytes_(hd + 1, 1);
    }

    if (n == -1)
    {
        return -1;	// Read timeout in testing
    }

    if ((hd[0] != 0xFA) || (hd[1] != 0xFB))
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printTextWindow("Pioneer2: bad packet");
        }
        return -1;
    }

    len = count = pack[0] = nextByte_();
    data = pack + 1;

    while (len)
    {
        n = nextRemainingBytes_(data, len);

        if (n == -1) // Read timeout in testing
        {
            return -1;
        }

        len -= n;
        data += n;
    }

    checksum = (pack[count - 1] << 8) | (pack[count]);
    rchecksum = calc_chksum(pack);

    if (checksum != rchecksum)
    {
        printfTextWindow("Pioneer2: Checksum failed");
        return -1;
    }

    if (isSynced_ || isSyncing_)
    {
        decodePack(pack);
    }

    return pack[1];
}

//---------------------------------------------------------------------
// This function tests the robot connection.
//---------------------------------------------------------------------
int Pioneer2::testRobotConnection_(void)
{
    int rtn;

    sendSync_(CMD_SYNC2_);
    sendSync_(CMD_SYNC0_);
    sendSync_(CMD_SYNC0_);
    sendSync_(CMD_SYNC0_);

    rtn = readPack_();

    if (rtn != -1)
    {
        return 0;
    }

    return -1;
}

//---------------------------------------------------------------------
// This function starts the reader thread
//---------------------------------------------------------------------
void *Pioneer2::startReaderThread_(void* pioneer2Instance)
{
    ((Pioneer2*)pioneer2Instance)->readerLoop_();
    return NULL;
}

//---------------------------------------------------------------------
// This function runs the reader in a loop
//---------------------------------------------------------------------
void Pioneer2::readerLoop_(void)
{
    int count = 0;

    isReading_ = true;

    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Read the data from the robot.
        readPack_();

        // Compute the control speed.
        compControlSpeed_();

        // Control the robot.
        controlRobot_();

        // Check for the pthread termination again.
        pthread_testcancel();

        // Compute the frequency.
        compFrequency_();

        // Compute the velocity.
        updateVelocity_();

        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        usleep(READER_LOOP_USLEEP_);
    }
}

void Pioneer2::addSensors(void)
{
    partXyt = new RobotPartXyt();

    partSonar = new RobotPartSonar(PIONEER2_NUM_SONAR, (float *)SONAR_ANGLES_);

    sensorSonar = new Sensor(
        SENSOR_SONAR,
        sizeof(sonar_loc),
        (char *)sonar_loc,
        numSonars_,
        (float *)SONAR_ANGLES_,
        numSonars_,
        sonar,
        NULL);

    if (ENABLE_IR_)
    {
        partIr = new RobotPartIr(PIONEER2_NUM_IR, pioneer2_range_angle_IR);

        sensorIr = new Sensor(
            SENSOR_IR,
            0,
            NULL,
            PIONEER2_NUM_IR,
            pioneer2_range_angle_IR,
            PIONEER2_NUM_IR,
            ir,
            NULL);
    }

    if (ENABLE_COMPASS_)
    {
        partCompass = new RobotPartCompass();

        sensorCompass = new Sensor(
            SENSOR_COMPASS,
            0,
            NULL,
            1,
            compass,
            0,
            NULL,
            NULL);

        compass[0] = 0.0;
    }
}

void Pioneer2::control(void)
{
    int c;
    int done = 0;
    int msgData[3];

    msgData[0] = isRangeOn_;
    msgData[1] = 0; // "speed_factor" is obsolete.
    msgData[2] = 0; // "angular_speed_factor" is obsolete.
    messageDrawWindow(EnMessageType_PIONEER2, EnMessageErrType_NONE, msgData);
    redrawWindows();
    do
    {
        c = getch();
        switch(c) {

        case 'd':
            messageHide();
            delete this;
            return;
            break;
        case 's':
            if (!isSynced_ && !isSyncing_)
            {
                runSyncer_();
            }
            break;
        case 'p':
            isRangeOn_ = !isRangeOn_;
            msgData[0] = isRangeOn_;
            if (isRangeOn_)
            {
                sonarStart();
            }
            else
            {
                sonarStop();
            }
            messageDrawWindow(EnMessageType_PIONEER2, EnMessageErrType_NONE, msgData);
            break;
        case 'r':
            refreshScreen();
            break;
        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;
        }
    } while (!done);
    messageHide();
}

/*
void Pioneer2::setDigOut(int port, int mode)
{
   
    unsigned int myOut = (1<<(port+8))|(mode<<port);
    sendInt_(myOut, CMD_DIGOUT_);
}
*/

/**********************************************************************
# $Log: pioneer2.c,v $
# Revision 1.2  2008-07-15 22:33:52  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.6  2008/04/27 02:41:11  endo
# Regular backup.
#
# Revision 1.5  2008/04/18 21:46:36  endo
# Regular backup.
#
# Revision 1.4  2008/04/17 02:39:09  endo
# Regular backup.
#
# Revision 1.3  2008/04/17 01:29:06  endo
# Pioneer2 cleaned up a bit.
#
# Revision 1.2  2006/12/07 20:30:23  endo
# Regular backup.
#
# Revision 1.1.1.1  2006/07/21 17:42:10  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.4  2005/03/26 00:41:30  endo
# *** empty log message ***
#
# Revision 1.3  2005/03/26 00:30:55  endo
# check for isSynced_ added in controlRobot_().
#
# Revision 1.2  2004/09/10 19:41:07  endo
# New PoseCalc integrated.
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.1  2003/04/25 21:34:04  ebeowulf
# Initial revision
#
#**********************************************************************/
