/**********************************************************************
 **                                                                  **
 **                         rcom_handler.c                           **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: rcom_handler.c,v 1.1 2008-07-15 22:33:52 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <math.h>
#include <sys/time.h>

#include "RComIDs.h"
#include "RComAPIFactory.h"
#include "rcom_handler.h"
#include "hserver.h"
#include "sensors.h"
#include "robot.h"
#include "RemoteControl.h"

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------
RComHandler *gRComHandler = NULL;
bool gRComDebug = false;

//-----------------------------------------------------------------------
// Constatns
//-----------------------------------------------------------------------
const struct timespec RComHandler::READER_SLEEPTIME_NSEC_ = {0, 10000000}; // 10 msec.
const string RComHandler::EMPTY_STRING_ = "";
const int RComHandler::SKIP_STATUSBAR_UPDATE_ = 10;

//-----------------------------------------------------------------------
// Mutex initialization
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Constructor
//-----------------------------------------------------------------------
RComHandler::RComHandler(RComHandler** a) :
    Module((Module**)a, HS_MODULE_NAME_RCOM),
    rcom_(NULL),
    myID_(RCOM_USER_ID_HSERVER),
    readerThreadIsUp_(false)
{   
    // Initialize the RCom connection.
    if (!initialize(RCOMAPI_TYPE_RCOM_CLIENT))
    {
        printfTextWindow("Error(RComHandler): RCom could not be initialized.");
        delete this;
        return;
    }

    // Update the status bar.
    updateStatusBar_();

    // Run threads.
    run();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
RComHandler::~RComHandler(void)
{
    if (readerThreadIsUp_)
    {
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL );
    }

    sleep(1);

    if (rcom_ != NULL)
    {
        delete rcom_;
        rcom_ = NULL;
    }

    sleep(1);
    printfTextWindow("RComHandler exited.");
}

//-----------------------------------------------------------------------
// This function starts up a thread that reads data from the Robot.
//-----------------------------------------------------------------------
void *RComHandler::startReaderThread_(void *data)
{
    RComHandler *rcInstance;

    rcInstance = (RComHandler *)data;
    
    // Start the reader loop
    rcInstance->readerLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to reads data from Robot.
//-----------------------------------------------------------------------
void RComHandler::readerLoop_(void)
{
    int count = 0;

    readerThreadIsUp_ = true;

    while(true)
    {
        // Recieves the latest data to the board.
        updateSensors_();

        // Make sure the thread is not exited.
        pthread_testcancel();

        // Compute the frequency.
        compFrequency_();

        // Update the status bar.
        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        // Make sure the thread is not exited.
        pthread_testcancel();

        // Sleep a bit.
        nanosleep(&READER_SLEEPTIME_NSEC_, NULL);
    }
}

//-----------------------------------------------------------------------
// This function initializes RComHandler.
//-----------------------------------------------------------------------
bool RComHandler::initialize(int rcomAPIType)
{
    // Start RCom.
    rcom_ = RComAPIFactory::createRComAPI(rcomAPIType);

    if (rcom_ == NULL)
    {
        printfTextWindow("Error: RComHandler::initialize(). RComAPI could not be created.");
        return false;
    }

    // Initialize RCom.
    if (!(rcom_->initialize(myID_, cbProcessRecievedData_, (void *)this)))
    {
        printfTextWindow("Error: RComHandler::initialize(). RComAPI could not be initialized.");
        return false;
    }

    printfTextWindow("RComHandler activated.");

    return true;
}

//-----------------------------------------------------------------------
// This function update sensor readings from HServer.
//-----------------------------------------------------------------------
void RComHandler::updateSensors_(void)
{
    char *data = NULL, *extraData = NULL;
    int i, numSensors, dataSize, extraLength, sensorType, numAngles, numReadings, pos;
    float *angles = NULL, *readings = NULL;
    
    numSensors = sensors->numSensors();

    for (i = 0; i < numSensors; i++)
    {
        sensorType = sensors->latest_sensor[i]->getType();
        extraLength = sensors->latest_sensor[i]->getLenExtra();
        extraData = sensors->latest_sensor[i]->getExtra();
        sensors->latest_sensor[i]->getAngleInfo(angles, numAngles);
        sensors->latest_sensor[i]->getReadingInfo(readings, numReadings);

        if (angles == NULL)
        {
            numAngles = 0;
        }

        if (readings == NULL)
        {
            numReadings = 0;
        }

        // Sends:
        // - Sensor Type
        // - Extra Data Length
        // - Extra Data
        // - # of Angles
        // - Angles
        // - # of Readings
        // - Readings
        dataSize = (sizeof(int)*4) + extraLength + (sizeof(float)*numAngles) + (sizeof(float)*numReadings);
        data = new char[dataSize];
        memset(data, 0x0, dataSize);
        
        pos = 0;
        memcpy(data + pos, &sensorType, sizeof(int));
        pos += sizeof(int);

        memcpy(data + pos, &extraLength, sizeof(int));
        pos += sizeof(int);

        if (extraLength > 0)
        {
            memcpy(data + pos, &extraData, extraLength);
            pos += extraLength;
        }

        memcpy(data + pos, &numAngles, sizeof(int));
        pos += sizeof(int);

        if (numAngles > 0)
        {
            memcpy(data + pos, angles, sizeof(float)*numAngles);
            pos += sizeof(float)*numAngles;
        }

        memcpy(data + pos, &numReadings, sizeof(int));
        pos += sizeof(int);

        if (numReadings > 0)
        {
            memcpy(data + pos, readings, sizeof(float)*numReadings);
        }

        rcom_->sendData(
            RCOM_USER_ID_AIR_PERCEPTION,
            (const void *)data,
            dataSize,
            false);

        if (angles != NULL)
        {
            delete [] angles;
            angles = NULL;
        }

        if (readings != NULL)
        {
            delete [] readings;
            readings = NULL;
        }

        if (data != NULL)
        {
            delete [] data;
            data = NULL;
        }
    }
}

//-----------------------------------------------------------------------
// This function runs RComHandler.
//-----------------------------------------------------------------------
void RComHandler::run(void)
{
    // Start receiving data from the robot.
    pthread_create(&readerThread_, NULL, &startReaderThread_, (void *)this);
}

//-----------------------------------------------------------------------
// This callback function will get called when any message is delivered
// to RComHandler.
//-----------------------------------------------------------------------
void RComHandler::cbProcessRecievedData_(
    int recipientID,
    const void *data,
    size_t dataSize,
    bool reliableEnforced,
    void *extraArg)
{
    RComHandler *rcomHandlerInstrance = NULL;
    float* actionValues = NULL;
    int senderID;
    int i, numActionTypes;
    char *dataPtr = NULL;
    char msg[1024];
    const bool DEBUG_CB_PROCESS_RECIEVED_DATA = false;

    dataPtr = (char *)data;
    rcomHandlerInstrance = (RComHandler *)extraArg;

    memcpy(&senderID, dataPtr, sizeof(int));
    dataPtr += sizeof(int);

    memcpy(&numActionTypes, dataPtr, sizeof(int));
    dataPtr += sizeof(int);

    actionValues = new float[numActionTypes];

    for (i = 0; i < numActionTypes; i++)
    {
        memcpy(&(actionValues[i]), dataPtr, sizeof(float));
        dataPtr += sizeof(float);
    }

    if (numActionTypes > NUM_ACTION_TYPES)
    {
        printfTextWindow("Error: RComHandler::cbProcessRecievedData_(). Invalid numActionTypes.\n");
    }
    else
    {
        switch (senderID) {

        case RCOM_USER_ID_AIR_BEHAVIORAL:
            if ((gRemoteControl != NULL) && (gRemoteControl->controllerIsUp()))
            {
                // Manual controller is up. Do not execute the command.
            }
            else
            {
                if (gRobot != NULL)
                {
                    gRobot->setDesiredDriveSpeed(actionValues[ACTION_TYPE_TRANS_SPPED_X]);
                    gRobot->setDesiredSteerSpeed(actionValues[ACTION_TYPE_ANG_SPPED_YAW]);
                }
            }
            break;
        }
    }

    if ((rcomHandlerInstrance->reportLevel() >= HS_REP_LEV_DEBUG) ||
        DEBUG_CB_PROCESS_RECIEVED_DATA)
    {
        msg[0] = '\0';

        for (i = 0; i < numActionTypes; i++)
        {
            sprintf(msg, "%s%.2f ", msg, actionValues[i]);
        }

        printfTextWindow("Data received from RCom:\n");
        printfTextWindow(" Recipient ID: %d\n", recipientID);
        printfTextWindow(" Data Size: %d\n", dataSize);
        printfTextWindow(" Enforced Reliability: %s\n", reliableEnforced? "Yes" : "No");
        printfTextWindow(" Extra Arg from initialize(): %s\n", (extraArg)? "Specified" : "None");
        printfTextWindow(" Data: %s\n", msg);
    }

    delete [] actionValues;
    actionValues = NULL;
}

//-----------------------------------------------------------------------
// This function returns the current time (since Epoch) in second.
//-----------------------------------------------------------------------
double RComHandler::getCurrentEpochTime_(void)
{
    timeval tvp;
    double currentEpochTime;
    
    gettimeofday(&tvp, NULL);
    currentEpochTime = (double)tvp.tv_sec + (((double)tvp.tv_usec)/1000000.0);

    return currentEpochTime;
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void RComHandler::updateStatusBar_(void)
{
    char buf[100];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[0] = 0;

    if (readerThreadIsUp_)
    {
        sprintf(buf, "RCom: %c Connected", statusbarSpinner_->getStatus());

        if (DISPLAY_FREQUENCY)
        {
            getFrequency_(frequency);
            sprintf(buf, "%s : %.2f Hz", buf, frequency);
        }
    }
    else
    {
        sprintf(buf, "RCom: %c Connecting...", statusbarSpinner_->getStatus());
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

/**********************************************************************
# $Log: rcom_handler.c,v $
# Revision 1.1  2008-07-15 22:33:52  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.10  2008/06/02 18:01:32  endo
# Regular backup.
#
# Revision 1.9  2008/06/02 02:43:24  endo
# Regular backup.
#
# Revision 1.8  2006/11/29 23:13:28  endo
# OpenCV improved.
#
# Revision 1.7  2006/11/08 21:49:36  endo
# Regular backup.
#
# Revision 1.6  2006/11/07 23:59:33  endo
# Regular backup.
#
# Revision 1.5  2006/11/07 17:02:26  endo
# RComHandler::cbProcessRecievedData_() implemented.
#
# Revision 1.4  2006/09/06 22:18:22  endo
# Regular update.
#
# Revision 1.3  2006/09/05 22:24:29  endo
# Regular backups.
#
# Revision 1.2  2006/08/31 20:40:47  endo
# Regular update.
#
# Revision 1.1  2006/08/15 15:28:39  endo
# RComHandler implemented.
#
#**********************************************************************/
