/**********************************************************************
 **                                                                  **
 **                              robot.h                             **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Generic robot class for HServer                                 **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: robot.h,v 1.10 2010-05-03 22:10:10 zkira Exp $ */

#ifndef ROBOT_H
#define ROBOT_H

#include <pthread.h>
#include <string>

#include "HServerTypes.h"
#include "module.h"
#include "hserver.h"

#define ROBOT_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}
#define ROBOT_RAD2DEG(val) ((val)*180.0/M_PI)
#define ROBOT_DEG2RAD(val) ((val)*M_PI/180.0)
#define ROBOT_M2MM(val) ((val)*1000.0)
#define ROBOT_M2IN(val) ((val)*39.3700787)
#define ROBOT_IN2M(val) ((val)*0.0254)

using std::string;

class Robot : public Module {

    typedef struct VelocityBuffers_t {
        vector<double> xBuffer;
        vector<double> yBuffer;
        vector<double> zBuffer;
        vector<double> yawBuffer;
        vector<double> pitchBuffer;
        vector<double> rollBuffer;
        bool initialized;
    };

protected:
    HSPose_t shPose_; // Shared
    HSPose_t shLastPose_; // Shared
    HSSpeed_t shControlSpeed_; // Shared        // Actual control speed (factors in desired values, mins/maxes, time the control command was sent, etc.) 
    HSSpeed_t shDesiredSpeed_; // Shared        // Desired speed (includes speed, rotation, and altitude)
    HSRotation_t shDesiredRotation_; // Shared
    HSVelocity_t shVelocity_; // Shared
    VelocityBuffers_t velBuffers_;
    double shBatteryVoltage_; // Shared
    double shMinPointTurnAngle_; // Shared
    double shTravelDist_; // Shared
    double battery_level; // To be obsolete
    double maxSpeed_;
    double maxSteerSpeed_;
    double maxSteerSpeedAngle_;
    double maxAltitudeSpeed_;
    double maxNoCommandTimeSec_;
    double minSteerSpeed_;
    double stoppingSpeedThreshold_;
    double numVelocitySamples_;
    int robotType_;
    int poseCalcID_;
    int velIndex_;
    bool shReversableDrive_; // Shared
    bool velBufferFilled_;

    static pthread_mutex_t poseMutex_;
    static pthread_mutex_t velocityMutex_;
    static pthread_mutex_t controlSpeedMutex_;
    static pthread_mutex_t desiredSpeedMutex_;
    static pthread_mutex_t desiredRotationMutex_;
    static pthread_mutex_t travelDistMutex_;
    static pthread_mutex_t minPointTurnAngleMutex_;
    static pthread_mutex_t reversableDriveMutex_;
    static pthread_mutex_t batteryVoltageMutex_;

    void setPose_(HSPose_t pose);
    void resetPose_(void);
    void advancePose_(HSPose_t deltaPose);
    void advancePose2D_(HSSpeed_t speed, double deltaTimeSec);
    void advancePoseAltitude_(HSPose_t pose, HSSpeed_t speed, double deltaTimeSec);
    void compControlSpeed_(void);
    void getControlSpeed_(HSSpeed_t &speed);
    void getDesiredSpeed_(HSSpeed_t &desiredSpeed);
    void getDesiredRotation_(HSRotation_t &rot);
    void getReversableDrive_(bool &reversable);
    void getMinimumPointTurnAngle_(double &angle);
    void updateVelocity_(void);
    void logData_(double logTime);
    void setBatteryVoltage_(double batteryVoltage);
    void getBatteryVoltage_(double &batteryVoltage);
    void addTravelDist_(double travelDist);

    virtual void updateStatusBar_(void) = 0;
    virtual void controlRobot_(void) = 0;

public:
    static const double DEFAULT_MIN_POINT_TURN_ANGLE_;
    static const double DEFAULT_MAX_SPEED_;
    static const double DEFAULT_MAX_STEER_SPEED_;
    static const double DEFAULT_MAX_STEER_SPEED_ANGLE_;
    static const double DEFAULT_MAX_ALTITUDE_SPEED_;
    static double DEFAULT_MAX_NO_COMMAND_TIME_SEC_;
    static const double DEFAULT_MIN_STEER_SPEED_;
    static const double DEFAULT_STOPPING_SPEED_THRESHOLD_;
    static const int    DEFAULT_NUM_VELOCITY_SAMPLES_;

    Robot(void);
    Robot(Robot **a, int robotType, string name);
    double getBattery(void) {return battery_level;}
    void savePoseCalcID(int id);
    void getPose(HSPose_t &pose);
    void setDesiredRotation(HSRotation_t rot);
    void setDesiredSpeed(HSSpeed_t speed);
    void setDesiredDriveSpeed(double driveSpeed);
    void setDesiredAltitudeSpeed(double steerSpeed);
    void setDesiredSteerSpeed(double steerSpeed);
    void setDesiredArmSpeed(double armSpeed);
    void setReversableDrive(bool reversable);
    void setMinimumPointTurnAngle(double angle);

    void setMaxSpeed(double val);
    void setMaxSteerSpeed(double val);
    void setMaxSteerSpeedAngle(double val);
    void setMaxAltitudeSpeed(double val);
    void setMaxNoCommandTimeSec(double val);
    void setMinSteerSpeed(double val);
    void setStoppingSpeedThreshold(double val);
    void setNumVelocitySamples(double val);

    void stop(void);
    void setVelocity(HSVelocity_t vel);
    void getVelocity(HSVelocity_t &vel);
    void getTravelDist(double &travelDist);

    virtual ~Robot(void);
    virtual void sonarStart(void) {};
    virtual void sonarStop(void) {};
    virtual void initRobot(void);
    virtual void control(void) = 0;


    virtual FrameData GetFrame(void) { FrameData fd; fd.Data = NULL; fd.DataSize = 0; 	printfTextWindow("getting frame from robot"); return fd; };
};

inline void Robot::savePoseCalcID(int id)
{
    poseCalcID_ = id;
}

extern Robot *gRobot;

// some variables to deal with rovio lag
extern float desiredang;
extern bool startrot;
extern bool donerotate;

#endif

/**********************************************************************
# $Log: robot.h,v $
# Revision 1.10  2010-05-03 22:10:10  zkira
# Cricket sensor, blob tracking, rovio (inc. IR sensor), skybotix interface (unfinished)
#
# Revision 1.7  2009/04/17 17:51:48  zkira
# Added maxSpeed in robot (clamps speed to this max value), and ability to speed up usarsim data acquisition so that speeding up the USARSim simulator becomes possible
#
# Revision 1.6  2009/04/13 02:58:43  zkira
# RFID code
#
# Revision 1.5  2009/03/02 18:52:48  zkira
# Made lower level robot movement parameters variables so that they can be modified externally
#
# Revision 1.4  2008/10/27 21:44:47  zkira
# Changed movement parameters to allow movement in a building
# Added advancePoseAltitude to allow for control-signal based crude odometry (not used currently)
#
# Revision 1.3  2008/09/09 20:20:07  zkira
# Crude altitude control: Added altitudeSpeed & z component
#
# Revision 1.2  2008/07/15 22:33:53  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.3  2007/04/09 17:36:10  endo
# TravelDist and its logging added.
#
# Revision 1.2  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.2  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2004/09/10 19:43:28  endo
# New PoseCalc integrated.
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.8  2003/04/26 06:48:31  ebeowulf
# Added the Pioneer2-dxe controller.
#
# Revision 1.7  2003/04/02 22:26:24  zkira
# Added battery level variable to class
#
# Revision 1.6  2002/02/18 13:48:21  endo
# AmigoBot added.
#
# Revision 1.5  2002/01/16 22:33:33  ernest
# Added
# m_bposition_set flag.
#
# Revision 1.4  2001/05/29 22:35:28  blee
# Made changes to incorporate gps data w/ other position info.
#
# Revision 1.3  2000/11/13 20:09:23  endo
# ARTV-Jr class added to HServer.
#
# Revision 1.2  2000/09/19 03:49:46  endo
# RCS log added.
#
#
#**********************************************************************/
