/**********************************************************************
 **                                                                  **
 **                             sensors.c                            **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Sensor routines for HServer                                     **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: sensors.c,v 1.11 2009-04-13 03:57:39 zkira Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <string.h>
#include <stdlib.h>
#include <panel.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#include "hserver.h"
#include "sensors.h"


//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------
Sensors *sensors = NULL;

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t Sensor::localDataMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Contructor for Sensor
//-----------------------------------------------------------------------
Sensor::Sensor(
    int type,
    int len_extra,
    char *extra,
    int num_angles,
    float *angle,
    int num_readings,
    float *reading, 
    pthread_mutex_t *externalDataMutexPtr) :
    externalDataMutexPtr_(externalDataMutexPtr),
    m_pfAngles(angle),
    m_pfPitchAngle(0.0),
    m_pfReadings(reading),
    m_iType(type),
    m_iLenExtra(len_extra),
    m_iNumAngles(num_angles),
    m_iNumReadings(num_readings),
    m_pcExtra(extra)
{

    pthread_mutex_init(&localDataMutex_, NULL);
    m_iIndex = sensors->add(this);
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
Sensor::~Sensor()
{
    sensors->remove(m_iIndex);
}

//-----------------------------------------------------------------------
// This function returns the number of angles.
//-----------------------------------------------------------------------
int Sensor::getAngleCount()
{
    int iCount = 0;

    // aquire the angle lock
    if (!TryToLock(externalDataMutexPtr_))
    {
        pthread_mutex_lock(&localDataMutex_);
    }

    iCount = m_iNumAngles;

    // release the angle lock
    if (!TryToUnlock(externalDataMutexPtr_))
    {
        pthread_mutex_unlock(&localDataMutex_);
    }

    return iCount;
}

//-----------------------------------------------------------------------
// This function returns the number of readings.
//-----------------------------------------------------------------------
int Sensor::getReadingCount()
{
    int iCount = 0;

    // aquire the angle lock
    if (!TryToLock(externalDataMutexPtr_))
    {
        pthread_mutex_lock(&localDataMutex_);
    }

    iCount = m_iNumReadings;

    // release the angle lock
    if (!TryToUnlock(externalDataMutexPtr_))
    {
        pthread_mutex_unlock(&localDataMutex_);
    }

    return iCount;
}

//-----------------------------------------------------------------------
// This function returns the angles.
//-----------------------------------------------------------------------
void Sensor::getAngleInfo(float*& pfAngles, int& iCount)
{
    // aquire the angle lock
    if (!TryToLock(externalDataMutexPtr_))
    {
        pthread_mutex_lock(&localDataMutex_);
    }

    // make a copy of the angle info
    if (m_iNumAngles > 0)
    {
        iCount = m_iNumAngles;
        if ((pfAngles = new float[m_iNumAngles]) != NULL)
        {
            memcpy(pfAngles, m_pfAngles, m_iNumAngles * sizeof(float));
        }
    }
    else
    {
        pfAngles = NULL;
        iCount = 0;
    }

    // release the angle lock
    if (!TryToUnlock(externalDataMutexPtr_))
    {
        pthread_mutex_unlock(&localDataMutex_);
    }
}

//-----------------------------------------------------------------------
// This function returns the readings.
//-----------------------------------------------------------------------
void Sensor::getReadingInfo(float*& pfReadings, int& iCount)
{
    // aquire the readings lock
    if (!TryToLock(externalDataMutexPtr_))
    {
        pthread_mutex_lock(&localDataMutex_);
    }

    // make a copy of the readings info
    if (m_iNumReadings > 0)
    {
        iCount = m_iNumReadings;
        if ((pfReadings = new float[m_iNumReadings]) != NULL)
        {
            memcpy(pfReadings, m_pfReadings, m_iNumReadings * sizeof(float));
        }
    }
    else
    {
        pfReadings = NULL;
        iCount = 0;
    }

    // release the readings lock
    if (!TryToUnlock(externalDataMutexPtr_))
    {
        pthread_mutex_unlock(&localDataMutex_);
    }
}

//-----------------------------------------------------------------------
// General note on setAngleInfo() and setReadingInfo()
// If the actual sensor that this object represents did not
// pass in a mutex pointer to lock, then this object will 
// use its local mutex to protect the buffer. If the sensor
// did pass in a mutex pointer, then it is up to that sensor
// to lock it. It knows when it's own data needs to be 
// protected, and it's reasonable that it will want to lock 
// the data while this change takes place, so if we also try
// to lock it here, we'll have deadlock.
//-----------------------------------------------------------------------
void Sensor::setAngleInfo(float* pfAngles, int iCount)
{
    if (externalDataMutexPtr_ == NULL)
    {
        pthread_mutex_lock(&localDataMutex_);
    }

    m_pfAngles = pfAngles;
    m_iNumAngles = iCount;

    if (externalDataMutexPtr_ == NULL)
    {
        pthread_mutex_unlock(&localDataMutex_);
    }
}

void Sensor::setPitchAngle(float pfPitchAngle)
{
    if (externalDataMutexPtr_ == NULL)
    {
        pthread_mutex_lock(&localDataMutex_);
    }

    m_pfPitchAngle = pfPitchAngle;

    if (externalDataMutexPtr_ == NULL)
    {
        pthread_mutex_unlock(&localDataMutex_);
    }
}

//-----------------------------------------------------------------------
// See above general comment
//-----------------------------------------------------------------------
void Sensor::setReadingInfo(float* pfReadings, int iCount)
{
    if (externalDataMutexPtr_ == NULL)
    {
        pthread_mutex_lock(&localDataMutex_);
    }

    m_pfReadings = pfReadings;
    m_iNumReadings = iCount;

    if (externalDataMutexPtr_ == NULL)
    {
        pthread_mutex_unlock(&localDataMutex_);
    }
}

//-----------------------------------------------------------------------
// Contructor for Sensors
//-----------------------------------------------------------------------
Sensors::Sensors(int max)
{
    num_sensors = 0;
    max_sensors = max;
    latest_sensor = new Sensor*[max_sensors];
    memset(latest_sensor, -1, sizeof(Sensor*) * max_sensors);
    index = new char[max_sensors];
    memset(index, -1, sizeof(char) * max_sensors);
    windowInit();
    hide_panel(panel);
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
Sensors::~Sensors()
{
    delete [] latest_sensor;
    delete [] index;
}

//-----------------------------------------------------------------------
// This function add a sensor to the sensors list.
//-----------------------------------------------------------------------
int Sensors::add(Sensor* sensor)
{
    int i;
    int idx = -1;
    if (num_sensors == max_sensors)
    {
        return -1;
    }
    latest_sensor[num_sensors] = sensor;
    for(i = 0; i < max_sensors; i++)
    {
        if (index[i] == -1)
        {
            idx = i;
            break;
        }
    }
    index[idx] = num_sensors;
    num_sensors++;
    return idx;
}

//-----------------------------------------------------------------------
// This function removes a sensor from the sensors list.
//-----------------------------------------------------------------------
void Sensors::remove(int idx)
{
    int i, low;
    low = index[idx];
    index[idx] = -1;
    for(i = 0; i < max_sensors; i++)
    {
        if (index[i] > low)
        {
            index[i]--;
        }
        if ((i >= low) && (i < max_sensors - 1))
        {
            latest_sensor[i] = latest_sensor[i + 1];
        }
    }
    num_sensors--;
}

//-----------------------------------------------------------------------
// This function displays the updated sensor information of the robot in
// the sensor window.
//-----------------------------------------------------------------------
void Sensors::drawWindow(void)
{
    float rdg;
    char buf[100];
    char buf2[100];
    int line = 1;
    int i,j;
    int num, num_rdgs;
    int sensorType;
    
    pthread_cleanup_push((void(*)(void*))pthread_mutex_unlock, (void *) &screenMutex);
    pthread_mutex_lock(&screenMutex);
    werase(window);
    wmove(window,0,0);
    wborder(window,0,0,0,0,0,0,0,0);
    mvwaddstr(window, line++, 1, "Robot Sensors");
    mvwaddstr(window, line++, 1, "");

    for(i = 0; i < num_sensors; i++)
    {
        sensorType = latest_sensor[i]->getType();

        switch(sensorType) {

        case SENSOR_PANTILT_CAMERA:
            sprintf(
                buf,
                "%d.  pan/tilt  pan: %.0f tilt: %.0f zoom: %.0f focus %.0f",i,
                latest_sensor[i]->m_pfAngles[0],
                latest_sensor[i]->m_pfAngles[1],
                latest_sensor[i]->m_pfAngles[2],
                latest_sensor[i]->m_pfAngles[3]);
            break;

        case SENSOR_XYT:
            sprintf(buf, "%d.  Dead Reckoning: ", i);
            mvwaddstr(window,line++,1,buf);
            sprintf(
                buf,
                "    x: %.2f y: %.2f z:%.2f yaw: %.2f roll: %.2f pitch: %.2f",
                latest_sensor[i]->m_pfReadings[0],
                latest_sensor[i]->m_pfReadings[1],
                latest_sensor[i]->m_pfReadings[2],
                latest_sensor[i]->m_pfReadings[3],
                latest_sensor[i]->m_pfReadings[4],
                latest_sensor[i]->m_pfReadings[5]);
            break;

        case SENSOR_ARM:
            sprintf(buf,"%d.  arm  angle: %.3f",i,latest_sensor[i]->m_pfAngles[0]);
            break;

        case SENSOR_SICKLMS200_1:
        case SENSOR_SICKLMS200_2:
        case SENSOR_SICKLMS200_3:
        case SENSOR_SICKLMS200_4:
            sprintf(buf, "%d.  Laser:", i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    # of Readings: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_SONAR:
            num_rdgs = latest_sensor[i]->getReadingCount();
            sprintf(buf,"%d.  Sonar Readings: ",i);
            mvwaddstr(window,line++,1,buf);
            strcpy(buf,"    ");
            num = 1;

            for(j = 0; j < num_rdgs; j++)
            {
                rdg = latest_sensor[i]->m_pfReadings[j];
                if (rdg == (float)SONAR_READING_OUTSIDE_RANGE)
                {
                    sprintf(buf2,"---- ");
                }
                else
                {
                    sprintf(buf2,"%.2f ",rdg);
                }
                strcat(buf,buf2);

                if (num == 8)
                {
                    mvwaddstr(window,line++,1,buf);
                    strcpy(buf,"    ");
                    num = 0;
                }
                ++num;
            }
            break;

        case SENSOR_BUMPER:
            num_rdgs = latest_sensor[i]->getReadingCount();
            sprintf(buf,"%d.  Bumper Readings: ",i);
            mvwaddstr(window,line++,1,buf);
            strcpy(buf,"    ");
            num = 1;

            for(j = 0; j < num_rdgs; j++)
            {
                rdg = latest_sensor[i]->m_pfReadings[j];
                sprintf(buf2,"%.2f ",rdg);
                strcat(buf,buf2);

                if (num == 8)
                {
                    mvwaddstr(window,line++,1,buf);
                    strcpy(buf,"    ");
                    num = 0;
                }
                ++num;
            }
            break;

        case SENSOR_IR:
            sprintf(
                buf,
                "%d.  IR  # of Readings: %d",
                i,
                latest_sensor[i]->getReadingCount());
            mvwaddstr(window,line++,1,buf);
            strcpy(buf,"    ");

            if (latest_sensor[i]->getReadingCount() > 8)
            {
                num = 8;
            }
            else
            {
                num = latest_sensor[i]->getReadingCount();
            }

            for(j=0;j<num;j++)
            {
                sprintf(buf2,"%.2f ",latest_sensor[i]->m_pfReadings[j]);
                strcat(buf,buf2);
            }
            if (latest_sensor[i]->getReadingCount() > 8)
            {
                mvwaddstr(window,line++,1,buf);
                strcpy(buf,"    ");
                num = latest_sensor[i]->getReadingCount() - 8;
                for(j=0;j<num;j++)
                {
                    sprintf(buf2,"%.2f ",latest_sensor[i]->m_pfReadings[j+8]);
                    strcat(buf,buf2);
                }
            }
            break;

        case SENSOR_COGNACHROME:
            sprintf(buf, "%d.  Cognachrome:", i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    # of Blobs: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_OPENCV_BLOB_SIZE:
            sprintf(buf, "%d.  OpenCV [TrackBlob - Size]:", i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    # of Blobs: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_OPENCV_BLOB_COLOR:
            sprintf(buf, "%d.  OpenCV [TrackBlob - Color]:", i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    # of Blobs x 3 colors: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_OPENCV_FACE_SIZE:
            sprintf(buf, "%d.  OpenCV [FaceDetect - Size]:", i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    # of Faces: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_OPENCV_OBJECT_001:
        case SENSOR_OPENCV_OBJECT_002:
        case SENSOR_OPENCV_OBJECT_003:
        case SENSOR_OPENCV_OBJECT_004:
            sprintf(
                buf,
                "%d.  OpenCV [ObjectDetect(%d) - Size]:",
                i,
                (sensorType - SENSOR_OPENCV_OBJECT_001) + 1);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    # of Objects: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_RFID:
           sprintf(
                buf,
                "%d.  RFID:",
                i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    Max # of Objects: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_CHEMICAL:
           sprintf(
                buf,
                "%d.  Chemical:",
                i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    Max # of Objects: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_ENEMY:
           sprintf(
                buf,
                "%d.  Enemy:",
                i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    Max # of Objects: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_COMM:
           sprintf(
                buf,
                "%d.  Comm:",
                i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    Max # of Objects: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_VIDEO:
            sprintf(buf, "%d.  Video:", i);
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, "    # of Readings: %d", latest_sensor[i]->getReadingCount());
            mvwaddstr(window, line++, 1, buf);
            sprintf(buf, " ");
            break;

        case SENSOR_COMPASS:
            sprintf(
                buf,
                "%d.  Compass  Angle: %f\n",
                i,
                latest_sensor[i]->m_pfAngles[0]);
            break;

        case SENSOR_GPS:
            sprintf(
                buf,
                "%d.  gps lat: %f  lon: %f  height: %f  tog: %f sats: %.0f\n", 
                i,
                latest_sensor[i]->m_pfReadings[0],
                latest_sensor[i]->m_pfReadings[1],
                latest_sensor[i]->m_pfReadings[2],
                latest_sensor[i]->m_pfReadings[3],
                latest_sensor[i]->m_pfReadings[4]);
            break;

        case SENSOR_GROUND_TRUTH:
            sprintf(buf, "%d.  Ground Truth: ", i);
            mvwaddstr(window,line++,1,buf);
            sprintf(
                buf,
                "    x: %.2f y: %.2f z: %.2f",
                latest_sensor[i]->m_pfReadings[0],
                latest_sensor[i]->m_pfReadings[1],
                latest_sensor[i]->m_pfReadings[2]);
            break;

        default:
            sprintf(buf,"%d.  Unknown sensor type",i);
            break;
        }

        mvwaddstr(window,line++,1,buf);
    }

    pthread_cleanup_pop(1);
}

static void start_thread(void *sensors_instance)
{
    ((Sensors*)sensors_instance)->updator_thread();
}

void Sensors::updator_thread()
{
    while(true)
    {
        pthread_testcancel();
        drawWindow();
        redrawWindows();
        usleep(500000);
    }
}

void Sensors::capInt(int* i, int l, int u)
{
    if (*i < l) *i = l;
    if (*i > u) *i = u;
}

void Sensors::movePanel(int xdif, int ydif)
{
    panelX += xdif;
    panelY += ydif;
    capInt(&panelX, 0, mainPanelX - SENSORS_WIN_COL);
    capInt(&panelY, 0, mainPanelY - SENSORS_WIN_ROW);
    move_panel(panel, panelY, panelX);
    redrawWindows();
}

void Sensors::movePanelKey(int c)
{
    switch(c)
    {
    case '1':
        movePanel(-1,1);
        break;
    case '2':
        movePanel(0,1);
        break;
    case '3':
        movePanel(1,1);
        break;
    case '4':
        movePanel(-1,0);
        break;
    case '5':
        break;
    case '6':
        movePanel(1,0);
        break;
    case '7':
        movePanel(-1,-1);
        break;
    case '8':
        movePanel(0,-1);
        break;
    case '9':
        movePanel(1,-1);
        break;
    }
}


int Sensors::runKey(int c)
{
    switch(c){
    case 'r':
        drawWindow();
        refreshScreen();
        break;
    case 'x':
    case 'Q':
        pthread_cancel(updator);
        pthread_join(updator,NULL);
        return(1);
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
        movePanelKey(c);
        break;
    }
    return 0;
}


void Sensors::show()
{
    int c,done;
    
    done = false;
    show_panel(panel);
    drawWindow();
    redrawWindows();
    pthread_create(&updator, NULL, (void*(*)(void*))&start_thread, (void*) this);
    do
    {
        c = getch();
        done = runKey(c);
    } while (!done);
    hide_panel(panel);
    redrawWindows();
    return;
}

void Sensors::windowInit()
{
    panelX = SENSORS_WIN_X;
    panelY = SENSORS_WIN_Y;
    panelH = SENSORS_WIN_ROW;
    panelW = SENSORS_WIN_COL;
    if (panelH > screenY)
    {
        panelH = screenY;
    }
    if (panelW > screenX)
    {
        panelW = screenX;
    }
    if (panelX + panelW > screenX)
    {
        panelX = screenX - panelW;
    }
    if (panelY + panelH > screenY)
    {
        panelY = screenY - panelH;
    }
    window = newwin(panelH, panelW, panelY, panelX);
    panel = new_panel(window);
    drawWindow();
}

/**********************************************************************
# $Log: sensors.c,v $
# Revision 1.11  2009-04-13 03:57:39  zkira
# Enemy, chemical, and comm sensors (USARSim-DOE-MAST)
#
# Revision 1.9  2009/03/02 18:54:49  zkira
# Full 3D information
#
# Revision 1.8  2009/01/24 02:36:39  zkira
# Cleaned up code
#
# Revision 1.7  2008/11/20 03:51:46  ts2883
# Finished adding the pitch information for the laser.
#
# Revision 1.6  2008/11/05 22:14:20  zkira
# In the process of adding pitch information for laser
#
# Revision 1.5  2008/11/04 04:53:58  zkira
# Added pitch for 3D laser readings
#
# Revision 1.4  2008/10/30 22:50:50  zkira
# Added support for up to 4 laser sensors
#
# Revision 1.3  2008/09/09 20:16:44  zkira
# Crude altitude control: Added z component
#
# Revision 1.2  2008/07/15 22:33:53  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.8  2008/06/24 22:50:52  endo
# Ground-Truth sensor added.
#
# Revision 1.7  2008/04/17 01:29:06  endo
# Pioneer2 cleaned up a bit.
#
# Revision 1.6  2006/12/17 04:42:54  endo
# SENSOR_OPENCV_BLOB_COLOR added.
#
# Revision 1.5  2006/12/07 20:30:23  endo
# Regular backup.
#
# Revision 1.4  2006/12/01 21:14:54  endo
# Regular backup.
#
# Revision 1.3  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.2  2006/11/16 18:20:26  endo
# The virtual bumper added.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2006/05/15 03:06:53  endo
# gcc-3.4 upgrade
#
# Revision 1.1.1.1  2005/02/06 22:59:42  endo
# AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
#
# Revision 1.8  2002/01/16 22:40:54  ernest
# Initialize latest_sensor as precaution
#
# Revision 1.7  2001/03/08 23:32:08  blee
# Added initialization of member mutexes.
#
# Revision 1.6  2000/12/12 23:05:24  blee
# Changed member variable names and the constructor.
# Added getAngleCount, getReadingCount, getAngleInfo, getReadingInfo,
# setAngleInfo, and setReadingInfo. Made more multi-thread safe.
# Changed the case for SENSOR_SICKLMS200 in Sensors::drawWindow().
# Changed references in Sensors to Sensor member variables.
#
# Revision 1.5  2000/11/13 21:05:43  endo
# *** empty log message ***
#
# Revision 1.4  2000/11/13 20:09:23  endo
# ARTV-Jr class added to HServer.
#
# Revision 1.3  2000/09/19 03:49:46  endo
# RCS log added.
#
#
#**********************************************************************/
