
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/un.h>
#include <netdb.h>
#include <pthread.h>
#include <math.h>
#include <string.h>

#include "HServerTypes.h"
#include "hserver.h"
#include "sensors.h"
#include "robot_config.h"
#include "message.h"
#include "statusbar.h"
#include "skybotix-robot.h"
#include "Watchdog.h"
#include "LogManager.h"

//-----------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t SkybotixRobot::latestRawPoseMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t SkybotixRobot::sensorDataTableMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t SkybotixRobot::socketMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t SkybotixRobot::skybotixPoseOffsetMutex_ = PTHREAD_MUTEX_INITIALIZER;

const string SkybotixRobot::DEFAULT_SKYBOTIX_ROBOT_NAME_ = "skybotix_bot";
const string SkybotixRobot::DEFAULT_SKYBOTIX_ROBOT_CLASS_ = "Skybotix";
const string SkybotixRobot::DEFAULT_SKYBOTIX_SERVER_HOST_MACHINE_ = "127.0.0.1";
const string SkybotixRobot::EMPTY_STRING_ = "";
const int SkybotixRobot::DEFAULT_SKYBOTIX_SERVER_PORT_ = 3000;
const int SkybotixRobot::SKIP_STATUSBAR_UPDATE_ = 10;
const int SkybotixRobot::WATCHDOG_CHECK_INTERVAL_SEC_ = 2;
const int SkybotixRobot::MAIN_THREAD_USLEEP_ = 2000;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const SkybotixRobot::RobotSpec_t SkybotixRobot::ROBOT_SPECS_[NUM_SKYBOTIX_ROBOT_CLASSES] =
{
    // Note: The robot's Z-axis in Rovio points opposite of Z-axis in MissionLab.
    {
        SKYBOTIX_ROBOT_CLASS_DEFAULT,
        "Skybotix", // className
        0.5, // robotRadius
        0.1 // wheelRadius
    }
};

//-----------------------------------------------------------------------
// Contructor
//-----------------------------------------------------------------------
SkybotixRobot::SkybotixRobot(
    Robot** a,
    char *serverHostMachine,
    char *skybotixRobotName,
    char *skybotixRobotClass,
    int serverPort,
    float initX,
    float initY,
    float initZ,
    float initYaw,
    float initPitch,
    float initRoll) :
    Robot(a, HS_ROBOT_TYPE_SKYBOTIX, HS_MODULE_NAME_ROBOT_SKYBOTIX),
    partXyt_(NULL),
    initX_(initX),
    initY_(initY),
    initZ_(initZ),
    initYaw_(initYaw),
    initPitch_(initPitch),
    initRoll_(initRoll),
    robotSpecIndex_(-1),
    serverPort_(DEFAULT_SKYBOTIX_SERVER_PORT_),
    mainThreadIsUp_(false),
    readerThreadIsUp_(false)
{
    serverHostMachine_ = (serverHostMachine != NULL)?
        serverHostMachine : DEFAULT_SKYBOTIX_SERVER_HOST_MACHINE_;

    skybotixRobotName_ = (skybotixRobotName != NULL)?
        skybotixRobotName : DEFAULT_SKYBOTIX_ROBOT_NAME_;

    skybotixRobotClass_ = (skybotixRobotClass != NULL)?
        skybotixRobotClass : DEFAULT_SKYBOTIX_ROBOT_CLASS_;

    serverPort_ = (serverPort_ >= 0)?
        serverPort : DEFAULT_SKYBOTIX_SERVER_PORT_;

    pthread_mutex_init(&sensorDataTableMutex_, NULL);
    pthread_mutex_init(&socketMutex_, NULL);
    pthread_mutex_init(&skybotixPoseOffsetMutex_, NULL);
    pthread_mutex_init(&latestRawPoseMutex_, NULL);

    resetPose_();
    updateStatusBar_();

    if (!initializeSkybotix_())
    {
        printfTextWindow("Error: SkybotixRobot could not be initialized. Exiting Skybotix.\n");
        delete this;
        return;
    }

    addSensors_();

    pthread_create(&readerThread_, NULL, &startReaderThread_, (void*)this);
    pthread_create(&mainThread_, NULL, &startMainThread_, (void*)this);

    if (gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(NAME_, WATCHDOG_CHECK_INTERVAL_SEC_, getpid(), mainThread_);
    }

    refreshScreen();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
SkybotixRobot::~SkybotixRobot(void)
{
    if (mainThreadIsUp_)
    {
        pthread_cancel(mainThread_);
        pthread_join(mainThread_, NULL);
    }

    if (readerThreadIsUp_)
    {
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    delete partXyt_;
}

//-----------------------------------------------------------------------
// This function initialize skybotix
//-----------------------------------------------------------------------
bool SkybotixRobot::initializeSkybotix_(void)
{
    int i;

    skybotixInterface = new SkybotixInterface( strdup(serverHostMachine_.c_str()) );

    for (i = 0; i < NUM_SKYBOTIX_ROBOT_CLASSES; i++)
    {
        if (ROBOT_SPECS_[i].className == skybotixRobotClass_)
        {
            robotSpecIndex_ = i;
            break;
        }
    }

    if (robotSpecIndex_ < 0)
    {
        robotSpecIndex_ = 0;

        printfTextWindow(
            "Warning: SkybotixRobot::%s(). Robot spect not found. Using %s.\n",
            __FUNCTION__,
            ROBOT_SPECS_[robotSpecIndex_].className.c_str());
    }


    return true;
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void SkybotixRobot::updateStatusBar_(void)
{
    HSPose_t pose;
    char buf[100];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[0] = 0;

    if (mainThreadIsUp_)
    {
        strcat(statusStr, "SkybotixRobot [");

        getPose(pose);

        sprintf(
            buf,
            "%s]: %c x: %.2f  y: %.2f z: %.2f t: %.2f pitch:%.2f roll:%.2f",
            skybotixRobotName_.c_str(),
            statusbarSpinner_->getStatus(),
            pose.loc.x,
            pose.loc.y,
            pose.loc.z,
            pose.rot.yaw,
            pose.rot.pitch,
	    pose.rot.roll);

        if (DISPLAY_FREQUENCY)
        {
            getFrequency_(frequency);
            sprintf(buf, "%s : %.2f Hz", buf, frequency);
        }
    }
    else
    {
        sprintf(
            buf,
            "SkybotixRobot: Connecting to the server on %s [port %d]...",
            serverHostMachine_.c_str(),
            serverPort_);
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

//-----------------------------------------------------------------------
// This function starts the main thread.
//-----------------------------------------------------------------------
void *SkybotixRobot::startMainThread_(void *skybotixInstance)
{
    ((SkybotixRobot*) skybotixInstance)->mainLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function starts the reader thread.
//-----------------------------------------------------------------------
void *SkybotixRobot::startReaderThread_(void *skybotixInstance)
{
    ((SkybotixRobot*) skybotixInstance)->readerLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to move SkybotixRobot.
//-----------------------------------------------------------------------
void SkybotixRobot::mainLoop_(void)
{
    int count = 0;
    
    mainThreadIsUp_ = true;

    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Read the pose data from the skybotix.
        updatePose_();

        // Check for the pthread termination again.
        pthread_testcancel();

        // Compute the control speed.
        compControlSpeed_();

        // Control the robot.
        controlRobot_();

	//skybotixInterface->GetImage();

        // Compute the frequency.
        compFrequency_();

        // Compute the velocity.
        updateVelocity_();

        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        if (gLogManager != NULL)
        {
            manageLog_(readerThread_);
        }

        if (watchdog_ != NULL)
        {
            watchdog_->notifyUpdate();
        }

        // Ease the computational load.
        usleep(MAIN_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function loops around to receive the sensor from the server.
//-----------------------------------------------------------------------
void SkybotixRobot::readerLoop_(void)
{
    readerThreadIsUp_ = true;

    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();
		printfTextWindow("Readerloop");

        // Read the sensor data.
        if (skybotixInterface != NULL) {
		skybotixInterface->PeriodicUpdate();

	}

        // Check for the pthread termination again.
        pthread_testcancel();

        // Ease the computational load.
        usleep(MAIN_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function waits for the sensor data to arrive.
//-----------------------------------------------------------------------
void SkybotixRobot::waitForSensorData_(void)
{
    bool arrived = false;
    const struct timespec WAITER_SLEEP_TIME_NSEC = {1, 0}; // 1 second sleep

    while (true)
    {
        pthread_mutex_lock(&sensorDataTableMutex_);

        if (shSensorDataTable_.size() > 0)
        {
            arrived = true;
        }

        pthread_mutex_unlock(&sensorDataTableMutex_);

        if (arrived)
        {
            break;
        }

        nanosleep(&WAITER_SLEEP_TIME_NSEC, NULL);
    }
}

//-----------------------------------------------------------------------
// This function updates the current pose.
//-----------------------------------------------------------------------
void SkybotixRobot::updatePose_(void)
{
    HSPose_t pose, lastPose;
    HSSpeed_t controlSpeed;
    DataLine_t odoDataLine;
    map<string, DataLine_t>::iterator itr;
    string poseString = EMPTY_STRING_;
    vector<string> poseList;
    double dist;
    double dt, curTime;
    static double lastTime = 0;
    double x,y,theta;
    memset((HSPose_t *)&pose, 0x0, sizeof(HSPose_t));
    getLatestRawPose_(lastPose);

    switch (robotSpecIndex_) {

    case SKYBOTIX_ROBOT_CLASS_DEFAULT:
      //  if (skybotixInterface != NULL) skybotixInterface->GetPose(x, y, theta);
      //  else { 
          x = 0.0; //lastPose.loc.x + 0.01;
          y = 0.0; //lastPose.loc.y + 0.02;
          theta = 0.0; //lastPose.rot.yaw + 0.04;
	  //printfTextWindow("No skybotix interface: Made-up data! (%f,%f,%f)\n", x, y, theta);
	  // }
        break;
      default:
	printfTextWindow("Error: Invalid robot type %d\n", robotSpecIndex_);
    }

    pose.loc.x = x;
    pose.loc.y = y;
    pose.rot.yaw = theta;

    dist = sqrt(pow((pose.loc.x - lastPose.loc.x), 2.0) + pow((pose.loc.y - lastPose.loc.y), 2.0) + pow((pose.loc.z - lastPose.loc.z), 2.0));
    addTravelDist_(dist);
    setLatestRawPose_(pose);

    updatePoseFromSkybotixPoseOffset_();
}

//-----------------------------------------------------------------------
// This function sets the offset between the raw skybotix data and the
// current robot pose.
//-----------------------------------------------------------------------
void SkybotixRobot::setSkybotixPoseOffset_(HSPose_t poseOffset)
{
    pthread_mutex_lock(&skybotixPoseOffsetMutex_);
    shSkybotixPoseOffset_ = poseOffset;
    pthread_mutex_unlock(&skybotixPoseOffsetMutex_);
}

//-----------------------------------------------------------------------
// This function returns the offset between the raw skybotix data and the
// current robot pose.
//-----------------------------------------------------------------------
void SkybotixRobot::getSkybotixPoseOffset_(HSPose_t &poseOffset)
{
    pthread_mutex_lock(&skybotixPoseOffsetMutex_);
    poseOffset = shSkybotixPoseOffset_;
    pthread_mutex_unlock(&skybotixPoseOffsetMutex_);
}

//-----------------------------------------------------------------------
// This function saves the latest raw pose data.
//-----------------------------------------------------------------------
void SkybotixRobot::setLatestRawPose_(HSPose_t pose)
{
    pthread_mutex_lock(&latestRawPoseMutex_);
    shLatestRawPose_ = pose;
    pthread_mutex_unlock(&latestRawPoseMutex_);
}

//-----------------------------------------------------------------------
// This function returns the saved latest raw pose data.
//-----------------------------------------------------------------------
void SkybotixRobot::getLatestRawPose_(HSPose_t &pose)
{
    pthread_mutex_lock(&latestRawPoseMutex_);
    pose = shLatestRawPose_;
    pthread_mutex_unlock(&latestRawPoseMutex_);
}

//-----------------------------------------------------------------------
// This function updates the robot pose from the offset.
//-----------------------------------------------------------------------
void SkybotixRobot::updatePoseFromSkybotixPoseOffset_(void)
{
    HSPose_t pose;
    HSPose_t poseOffset;
    double x, y, z, angle, altAngle;

    getLatestRawPose_(pose);
    getSkybotixPoseOffset_(poseOffset);

    x = pose.loc.x;
    y = pose.loc.y;
    z = pose.loc.z;

    angle = poseOffset.rot.yaw;
    ROBOT_CRAMPDEG(angle, -180.0, 180.0);
    angle = ROBOT_DEG2RAD(angle);

    pose.loc.x = (x*cos(angle)) - (y*sin(angle)) + poseOffset.loc.x;
    pose.loc.y = (x*sin(angle)) + (y*cos(angle)) + poseOffset.loc.y;
    pose.rot.yaw += poseOffset.rot.yaw;
    ROBOT_CRAMPDEG(pose.rot.yaw, -180.0, 180.0);
    pose.rot.pitch += poseOffset.rot.pitch;
    ROBOT_CRAMPDEG(pose.rot.pitch, -180.0, 180.0);

    setPose_(pose);
}

//-----------------------------------------------------------------------
// This function controls the robot with the control speed.
//-----------------------------------------------------------------------
void SkybotixRobot::controlRobot_(void)
{
    HSSpeed_t controlSpeed;
    float vt, va, rr, rt, leftSpeed, rightSpeed;
    float altVelocity, linearVelocity, latVelocity, rotVelocity;
    bool normalized = false;

    getControlSpeed_(controlSpeed);

    switch (robotSpecIndex_) {

    case SKYBOTIX_ROBOT_CLASS_DEFAULT:
        vt = controlSpeed.driveSpeed.value;
        va = ROBOT_DEG2RAD(controlSpeed.steerSpeed.value);
        rr = ROBOT_SPECS_[robotSpecIndex_].robotRadius;
        rt = ROBOT_SPECS_[robotSpecIndex_].wheelRadius;

        if ((rr <= 0) || (rt <= 0))
        {
            printfTextWindow("Error:SkybotixRobot::controlRobot_(). Invalid radius.\n");
            return;
        }

        leftSpeed = (vt - rr*va)/rt;
        rightSpeed = (vt + rr*va)/rt;

	//        if (skybotixInterface != NULL) skybotixInterface->SetControl(vt, 0, va);
        break;
      default:
	printfTextWindow("Error: Invalid robot type %d\n", robotSpecIndex_);
    }

}

//-----------------------------------------------------------------------
// This function adds the sensors.
//-----------------------------------------------------------------------
void SkybotixRobot::addSensors_(void)
{
    float x, y, dist;

    partXyt_ = new RobotPartXyt();

}

//-----------------------------------------------------------------------
// This function sets up the user interface for the robot option.
//-----------------------------------------------------------------------
void SkybotixRobot::control(void)
{
    bool done = false;
    int msgData[2];

    msgData[0] = 0; // "speed_factor" is obsolete.
    msgData[1] = 0; // "angular_speed_factor" is obsolete.

    /*
      messageDrawWindow(
      EnMessageType_SKYBOTIX_ROBOT_CONTROL,
      EnMessageErrType_NONE,
      msgData);
    */

    do
    {
        int c = getch();

        switch(c) {

        case 'd':
            messageHide();
            delete this;
            return;
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;

        }
    } while (!done);

    messageHide();
}

