
#ifndef SKYBOTIX_ROBOT_H
#define SKYBOTIX_ROBOT_H

#include <curl/curl.h>
#include <pthread.h>
#include <string>
#include <vector>
#include <map>

#include "HServerTypes.h"
#include "robot.h"
#include "sensors.h"
#include "robot_config.h"
#include "skybotix_interface.h"

using std::string;
using std::vector;
using std::map;

//#define ROVIO_ROBOT_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}
//#define ROVIO_ROBOT_RAD2DEG(val) ((val)*180.0/M_PI)
//#define ROVIO_ROBOT_DEG2RAD(val) ((val)*M_PI/180.0)
//#define ROVIO_ROBOT_MAX_SONARS 100
//#define ROVIO_ROBOT_SONAR_LOC_DIMENSIONS 3

class SkybotixRobot : public Robot {

    typedef enum InitiConnectionStatus_t
    {
        INIT_CONNECTION_STATUS_FAILURE,
        INIT_CONNECTION_STATUS_NOT_READY,
        INIT_CONNECTION_STATUS_SUCCESS
    };

    typedef enum SkybotixRobotClasses_t
    {
        SKYBOTIX_ROBOT_CLASS_DEFAULT,
        NUM_SKYBOTIX_ROBOT_CLASSES
    };

    typedef struct RobotSpec_t
    {
        int robotClass;
        string className;
        float robotRadius;
        float wheelRadius;
    };

    typedef enum ERSLowSpeedGatePhase_t
    {
        ERS_LOWSPEED_GATE_PHASE_0,
        ERS_LOWSPEED_GATE_PHASE_1,
        ERS_LOWSPEED_GATE_PHASE_2,
        ERS_LOWSPEED_GATE_PHASE_3,
        ERS_LOWSPEED_GATE_PHASE_4,
        ERS_LOWSPEED_GATE_PHASE_5,
        ERS_LOWSPEED_GATE_PHASE_6,
        ERS_LOWSPEED_GATE_PHASE_7,
        ERS_LOWSPEED_GATE_PHASE_8,
        ERS_LOWSPEED_GATE_PHASE_9,
        NUM_ERS_LOWSPEED_GATE_PHASES
    };

    typedef vector<string> DataLine_t;


protected:
    HSPose_t shSkybotixPoseOffset_; // Shared
    HSPose_t shLatestRawPose_; // Shared
    RobotPartXyt *partXyt_;
    pthread_t mainThread_;
    pthread_t readerThread_;
    string serverHostMachine_;
    string skybotixRobotName_;
    string skybotixRobotClass_;
    map<string, DataLine_t> shSensorDataTable_;
    float initX_, initY_, initZ_;
    float initYaw_, initPitch_, initRoll_;
    int robotSpecIndex_;
    int shSocketFd_;
    int serverPort_;
    bool mainThreadIsUp_;
    bool readerThreadIsUp_;
    SkybotixInterface *skybotixInterface;

#ifndef NO_OPENCV
    IplImage *CurFrame;
#endif

    static pthread_mutex_t latestRawPoseMutex_;
    static pthread_mutex_t sensorDataTableMutex_;
    static pthread_mutex_t socketMutex_;
    static pthread_mutex_t skybotixPoseOffsetMutex_;

    static const RobotSpec_t ROBOT_SPECS_[NUM_SKYBOTIX_ROBOT_CLASSES];
    static const string DEFAULT_SKYBOTIX_ROBOT_CLASS_;
    static const string DEFAULT_SKYBOTIX_ROBOT_NAME_;
    static const string DEFAULT_SKYBOTIX_SERVER_HOST_MACHINE_;
    static const int DEFAULT_SKYBOTIX_SERVER_PORT_;
    static const int MAIN_THREAD_USLEEP_;
    static const int WATCHDOG_CHECK_INTERVAL_SEC_;
    static const int SKIP_STATUSBAR_UPDATE_;
    static const string EMPTY_STRING_;

    DataLine_t readDataLine_(void);
    vector<string> extractCommaSepratedValues_(string values);
    string readString_(void);
    string generateERSGateCommand_(HSSpeed_t controlSpeed);
    void addSensors_(void);
    void controlRobot_(void);
    void getLatestRawPose_(HSPose_t &pose);
    void getSkybotixPoseOffset_(HSPose_t &poseOffset);
    void mainLoop_(void);
    void parsePositionData_(HSPose_t &pose);
    void readerLoop_(void);
    void setLatestRawPose_(HSPose_t pose);
    void setSkybotixPoseOffset_(HSPose_t poseOffset);
    void updatePoseFromSkybotixPoseOffset_(void);
    void updatePose_(void);
    void updateStatusBar_(void);
    void waitForSensorData_(void);
    void writeString_(string value);
    bool initializeSkybotix_(void);
    bool readNBytes_(const int fd, unsigned char *buf, const int n);

    static void* startMainThread_(void *skybotixInstance);
    static void* startReaderThread_(void *skybotixInstance);

public:
    SkybotixRobot(
        Robot** a,
        char *serverHostMahine,
        char *skybotixRobotName,
        char *skybotixRobotClass,
        int serverPort,
        float initX,
        float initY,
        float initZ,
        float initYaw,
        float initPitch,
        float initRoll);
    ~SkybotixRobot(void);
    int getLaserData(float **laserReadings, float **laserAngles, const char *laserName);
    void control(void);


    FrameData GetFrame(void) { return skybotixInterface->GetCurFrame(); };
};    

#endif


