/**********************************************************************
 **                                                                  **
 **                          usarsim-robot.h                         **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for USARSim robot.                                    **
 **                                                                  **
 **  Copyright 2008, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: usarsim-robot.h,v 1.8 2009-04-17 17:51:49 zkira Exp $ */

#ifndef USARSIM_ROBOT_H
#define USARSIM_ROBOT_H

#include <pthread.h>
#include <string>
#include <vector>
#include <map>

#include "HServerTypes.h"
#include "robot.h"
#include "sensors.h"
#include "robot_config.h"

using std::string;
using std::vector;
using std::map;

#define USARSIM_ROBOT_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}
#define USARSIM_ROBOT_RAD2DEG(val) ((val)*180.0/M_PI)
#define USARSIM_ROBOT_DEG2RAD(val) ((val)*M_PI/180.0)
#define USARSIM_ROBOT_MAX_SONARS 100
#define USARSIM_ROBOT_SONAR_LOC_DIMENSIONS 3

// these are used to define the maximum elements needed in an array for the associated
// sensor. for example, we need to a sensor data array large enough to store the data
// for 10 RFIDs...
#define USARSIM_ROBOT_MAX_NUM_RFID 10
#define USARSIM_ROBOT_MAX_NUM_ENEMY 51
#define USARSIM_ROBOT_MAX_NUM_CHEMICAL 11
#define USARSIM_ROBOT_MAX_NUM_BUG 15

class USARSimRobot : public Robot {

    typedef enum InitiConnectionStatus_t
    {
        INIT_CONNECTION_STATUS_FAILURE,
        INIT_CONNECTION_STATUS_NOT_READY,
        INIT_CONNECTION_STATUS_SUCCESS
    };

    typedef enum USARSimRobotClasses_t
    {
        USARSIM_ROBOT_CLASS_AIRROBOT,
        USARSIM_ROBOT_CLASS_DRAGONFLYROBOT,
        USARSIM_ROBOT_CLASS_ATRVJR,
        USARSIM_ROBOT_CLASS_ERS,
        USARSIM_ROBOT_CLASS_HMMWV,
        USARSIM_ROBOT_CLASS_P2AT,
        USARSIM_ROBOT_CLASS_P2DX,
        USARSIM_ROBOT_CLASS_TALON,
        USARSIM_ROBOT_CLASS_TARANTULA,
        NUM_USARSIM_ROBOT_CLASSES
    };

    typedef struct RobotSpec_t
    {
        int robotClass;
        string className;
        int numSonars;
        float sonarAngles[USARSIM_ROBOT_MAX_SONARS];
        float sonarCoordinates[USARSIM_ROBOT_MAX_SONARS*2];
        float robotRadius;
        float wheelRadius;
    };

    typedef enum ERSLowSpeedGatePhase_t
    {
        ERS_LOWSPEED_GATE_PHASE_0,
        ERS_LOWSPEED_GATE_PHASE_1,
        ERS_LOWSPEED_GATE_PHASE_2,
        ERS_LOWSPEED_GATE_PHASE_3,
        ERS_LOWSPEED_GATE_PHASE_4,
        ERS_LOWSPEED_GATE_PHASE_5,
        ERS_LOWSPEED_GATE_PHASE_6,
        ERS_LOWSPEED_GATE_PHASE_7,
        ERS_LOWSPEED_GATE_PHASE_8,
        ERS_LOWSPEED_GATE_PHASE_9,
        NUM_ERS_LOWSPEED_GATE_PHASES
    };

    typedef vector<string> DataLine_t;


protected:
    HSPose_t shUSARSimPoseOffset_; // Shared
    HSPose_t shLatestRawPose_; // Shared
    // Create robot parts
    RobotPartXyt *partXyt_;
    RobotPartSonar *partSonar_;
    RobotPartRFID *partRFID_;
    RobotPartEnemy *partEnemy_;
    RobotPartChemical *partChemical_;
    RobotPartComm *partComm_;
    // Create robot sensors
    Sensor *sensorSonar_;
    Sensor *sensorRFID_;
    Sensor *sensorEnemy_;
    Sensor *sensorChemical_;
    Sensor *sensorComm_;
    pthread_t mainThread_;
    pthread_t readerThread_;
    string serverHostMachine_;
    string usarsimRobotName_;
    string usarsimRobotClass_;
    map<string, DataLine_t> shSensorDataTable_;
    float initX_, initY_, initZ_;
    float initYaw_, initPitch_, initRoll_;
    float *sonarMountDist_;
    float *sonarReadings_;
    float (*sonarLocs_)[USARSIM_ROBOT_SONAR_LOC_DIMENSIONS];
    int numSonars_;
    // arrays to store the sensor data:
    float *RFIDData_;
    int numRFID_;
    float *enemyData_;
    int numEnemy_;
    float *chemicalData_;
    int numChemical_;
    float *commData_;
    int numComm_;
    int robotSpecIndex_;
    int shSocketFd_;
    int serverPort_;
    bool mainThreadIsUp_;
    bool readerThreadIsUp_;

    static pthread_mutex_t latestRawPoseMutex_;
    static pthread_mutex_t sensorDataTableMutex_;
    static pthread_mutex_t socketMutex_;
    static pthread_mutex_t usarsimPoseOffsetMutex_;

    static const RobotSpec_t ROBOT_SPECS_[NUM_USARSIM_ROBOT_CLASSES];
    static const string EMPTY_STRING_;
    static const string CMD_KEY_DRIVE_;
    static const string CMD_KEY_INIT_;
    static const string CMD_KEY_MULTIDRIVE_;
    static const string CMD_STRING_ALTITUDE_VELOCITY_;
    static const string CMD_STRING_CLASS_NAME_;
    static const string CMD_STRING_FALSE_;
    static const string CMD_STRING_LATERAL_VELOCITY_;
    static const string CMD_STRING_LEFT_;
    static const string CMD_STRING_LINEAR_VELOCITY_;
    static const string CMD_STRING_LOCATION_;
    static const string CMD_STRING_NAME_;
    static const string CMD_STRING_NORMALIZED_;
    static const string CMD_STRING_RIGHT_;
    static const string CMD_STRING_ROTATION_;
    static const string CMD_STRING_ROTATIONAL_VELOCITY_;
    static const string CMD_STRING_TRUE_;
    static const string DATA_HEADER_SENSOR_;
    static const string DATA_KEY_FOV_;
    static const string DATA_KEY_LOCATION_;
    static const string DATA_KEY_ORIENTATION_;
    static const string DATA_KEY_POSE_;
    static const string DATA_KEY_RANGE_;
    static const string DATA_KEY_RFID_ID_;
    static const string DATA_KEY_RFID_LOCATION_;

    static const string DATA_KEY_TIME_;

    static const string DATA_KEY_ENEMY_ID_;
    static const string DATA_KEY_ENEMY_DISTANCE_;
    static const string DATA_KEY_ENEMY_LOCATION_;
    static const string DATA_KEY_ENEMY_PODSELF_;
    static const string DATA_KEY_ENEMY_PODENEMY_;

    static const string DATA_KEY_CHEMICAL_ID_;
    static const string DATA_KEY_CHEMICAL_DISTANCE_;
    static const string DATA_KEY_CHEMICAL_LOCATION_;
    static const string DATA_KEY_CHEMICAL_PODCHEM_;

    static const string DATA_KEY_COMM_BUGNUM_;
    static const string DATA_KEY_COMM_PD_;

    static const string DATA_KEY_TYPE_;
    static const string DATA_VALUE_INS_;
    static const string DATA_VALUE_LASER_;
    static const string DATA_VALUE_ODOMETRY_;
    static const string DATA_VALUE_GROUND_TRUTH_;
    static const string DATA_VALUE_SONAR_;
    static const string DATA_VALUE_RFID_;
    static const string DATA_VALUE_ENEMY_;
    static const string DATA_VALUE_CHEMICAL_;
    static const string DATA_VALUE_COMM_;
    static const string DATA_VALUE_TOUCH_;
    static const string DEFAULT_USARSIM_ROBOT_CLASS_;
    static const string DEFAULT_USARSIM_ROBOT_NAME_;
    static const string DEFAULT_USARSIM_SERVER_HOST_MACHINE_;
    static const int DEFAULT_USARSIM_SERVER_PORT_;
    static const int INVALID_SOCKET_FD_;
    static const int MAIN_THREAD_USLEEP_;
    static const int MAX_CMD_LENGTH_;
    static const int SKIP_STATUSBAR_UPDATE_;
    static const int WATCHDOG_CHECK_INTERVAL_SEC_;
    static const char CHAR_BRACE_START_;
    static const char CHAR_BRACE_END_;
    static const char CHAR_CARRIAGE_RETURN_;
    static const char CHAR_COMMA_;
    static const char CHAR_NEW_LINE_;
    static const char CHAR_NULL_;
    static const char CHAR_WHITE_SPACE_;
    static const bool REPORT_SONARS_;
    static const bool REPORT_RFID_;
    static const bool REPORT_ENEMY_;
    static const bool REPORT_CHEMICAL_;
    static const bool REPORT_COMM_;

    DataLine_t readDataLine_(void);
    vector<string> extractCommaSepratedValues_(string values);
    string readString_(void);
    string generateERSGateCommand_(HSSpeed_t controlSpeed);
    int connectToServer_(string hostMacine, int port);
    void addSensors_(void);
    void controlRobot_(void);
    void getLatestRawPose_(HSPose_t &pose);
    void getUSARSimPoseOffset_(HSPose_t &poseOffset);
    void mainLoop_(void);
    void parsePositionData_(HSPose_t &pose);
    void parseSonarData_(void);
    void parseRFIDData_(void);
    void parseEnemyData_(void);
    void parseChemicalData_(void);
    void parseCommData_(void);
    void readerLoop_(void);
    void readSensorData_(void);
    void setLatestRawPose_(HSPose_t pose);
    void setUSARSimPoseOffset_(HSPose_t poseOffset);
    void updatePoseFromUSARSimPoseOffset_(void);
    void updatePose_(void);
    void updateStatusBar_(void);
    void waitForSensorData_(void);
    void writeString_(string value);
    bool initializeUSARSim_(void);
    bool readNBytes_(const int fd, unsigned char *buf, const int n);
    bool startRobot_(void);


    static void* startMainThread_(void *usarsimInstance);
    static void* startReaderThread_(void *usarsimInstance);

public:
    static const int MAIN_UT_SLOMO_;
    USARSimRobot(
        Robot** a,
        char *serverHostMahine,
        char *usarsimRobotName,
        char *usarsimRobotClass,
        int serverPort,
        float initX,
        float initY,
        float initZ,
        float initYaw,
        float initPitch,
        float initRoll);
    ~USARSimRobot(void);
    int getLaserData(float **laserReadings, float **laserAngles, const char *laserName);
    void control(void);
};    

#endif

/**********************************************************************
# $Log: usarsim-robot.h,v $
# Revision 1.8  2009-04-17 17:51:49  zkira
# Added maxSpeed in robot (clamps speed to this max value), and ability to speed up usarsim data acquisition so that speeding up the USARSim simulator becomes possible
#
# Revision 1.7  2009/04/13 03:57:40  zkira
# Enemy, chemical, and comm sensors (USARSim-DOE-MAST)
#
# Revision 1.5  2008/10/30 22:50:51  zkira
# Added support for up to 4 laser sensors
#
# Revision 1.4  2008/10/27 21:39:26  zkira
# Added separate dragonfly model
#
# Revision 1.3  2008/09/18 07:36:02  zkira
# Added ground truth
# BUG fix: Changed MAIN_LOOP_SLEEP because it was too slow (only every 10th or so message
# is an odometry reading!).  It may have to be changed back if it causes network troubles though.
#
# Revision 1.2  2008/08/26 19:40:50  endo
# Improving the USARSim module.
#
# Revision 1.5  2008/02/01 23:18:26  endo
# USARSimLaser added.
#
# Revision 1.4  2008/01/13 00:35:40  endo
# Regular backup.
#
# Revision 1.3  2008/01/11 15:51:18  endo
# Regular backup.
#
# Revision 1.2  2008/01/10 23:02:54  endo
# Regular backup.
#
# Revision 1.1  2008/01/09 22:54:03  endo
# USARSimRobot added.
#
#**********************************************************************/
