/**********************************************************************
 **                                                                  **
 **                         usarsim-video.c                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for USARSim video (via ImageSrv).                     **
 **                                                                  **
 **  Copyright 2008, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: usarsim-video.c,v 1.2 2009-08-28 19:46:40 zkira Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/un.h>
#include <netdb.h>
#include <pthread.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "HServerTypes.h"
#include "hserver.h"
#include "sensors.h"
#include "robot_config.h"
#include "message.h"
#include "statusbar.h"
#include "usarsim-video.h"
#include "Watchdog.h"
#include "LogManager.h"

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t USARSimVideo::socketMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Statics
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const string USARSimVideo::DEFAULT_USARSIM_SERVER_HOST_MACHINE_ = "127.0.0.1";
const string USARSimVideo::DEFAULT_USARSIM_VIDEO_NAME_ = "us_video";
const string USARSimVideo::USARSIM_ACK_STRING_ = "OK";
const int USARSimVideo::DEFAULT_USARSIM_SERVER_PORT_ = 5003;
const int USARSimVideo::IMAGE_BUFFER_SIZE_ = 2;
const int USARSimVideo::INVALID_SOCKET_FD_ = -1;
const unsigned int USARSimVideo::MAX_READ_BLOCK_SIZE_ = 2048;

//-----------------------------------------------------------------------
// Contructor
//-----------------------------------------------------------------------
USARSimVideo::USARSimVideo(
    Video **a,
    char *serverHostMachine,
    char *usarsimVideoName,
    int serverPort,
    bool showImage) :
    Video(a, HS_ROBOT_TYPE_USARSIM, HS_MODULE_NAME_VIDEO_USARSIM, showImage),
    lastTimeStamp_(0),
    shSocketFd_(INVALID_SOCKET_FD_),
    serverPort_(DEFAULT_USARSIM_SERVER_PORT_),
    imageBufferIndex_(-1)
{
    ImageData_t emptyImageData;
    string displayName;
    int i, width, height;

    serverHostMachine_ = (serverHostMachine != NULL)?
        serverHostMachine : DEFAULT_USARSIM_SERVER_HOST_MACHINE_;

    usarsimVideoName_ = (usarsimVideoName != NULL)?
        usarsimVideoName : DEFAULT_USARSIM_VIDEO_NAME_;

    serverPort_ = (serverPort_ >= 0)?
        serverPort : DEFAULT_USARSIM_SERVER_PORT_;

    pthread_mutex_init(&socketMutex_, NULL);

    // Reset the image data.
    emptyImageData.imageSize = 0;
    emptyImageData.image = NULL;

    if (imageBufferIndex_ < 0)
    {
        for (i = 0; i < IMAGE_BUFFER_SIZE_; i++)
        {
            imageBuffer_.push_back(emptyImageData);
        }
    }

    // Update the status bar.
    updateStatusBar_();

    // Initialize the usarsim.
    if (!initializeUSARSim_())
    {
        printfTextWindow("Error: USARSimVideo could not be initialized. Exiting Usarsim.\n");
        delete this;
        a = NULL;
        return;
    }

    // Start the reader thread.
    pthread_create(&readerThread_, NULL, &startReaderThread_, (void*)this);

    // Start the watchdog.
    if (gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(
            NAME_,
            WATCHDOG_CHECK_INTERVAL_SEC_,
            getpid(),
            readerThread_);
    }

    if (showImage_)
    {
        pthread_mutex_lock(&socketMutex_);
        width = shCurrentImageWidth_;
        height = shCurrentImageHeight_;
        pthread_mutex_unlock(&socketMutex_);

        displayName = NAME_ + ": " + usarsimVideoName_;
        new Xwindow(&xwindow_, (char *)(displayName.c_str()), 0, 0, width, height);

        // Start the display thread.
        pthread_create(&displayThread_, NULL, &startDisplayThread_, (void*)this);
    }

    refreshScreen();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
USARSimVideo::~USARSimVideo(void)
{
    if (displayThreadIsUp_)
    {
        displayThreadIsUp_ = false;
        pthread_cancel(displayThread_);
        pthread_join(displayThread_, NULL);
    }

    if (readerThreadIsUp_)
    {
        readerThreadIsUp_ = false;
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    printTextWindow("USARSimVideo disconnected");
}

//-----------------------------------------------------------------------
// This function initialize usarsim
//-----------------------------------------------------------------------
bool USARSimVideo::initializeUSARSim_(void)
{
    const struct timespec RETRY_SLEEP_TIME_NSEC = {1, 0}; // 1 second sleep
    int connStatus;
    bool done = false;

    while (!done)
    {
        // Connect to server.
        connStatus = connectToServer_(serverHostMachine_, serverPort_);

        switch (connStatus) {

        case INIT_CONNECTION_STATUS_NOT_READY:
            nanosleep(&RETRY_SLEEP_TIME_NSEC, NULL);
            break;

        case INIT_CONNECTION_STATUS_SUCCESS:
            printfTextWindow("USARSimVideo: Server connected.\n");
            done = true;
            break;

        default:
            printfTextWindow(
                "Error: USARSimVideo::initializeUSARSim(). Failed to connect to the server.\n");
            return false;
        }
    }

    if (!readVideoData_())
    {
        printfTextWindow(
            "Error: USARSimVideo::initializeUSARSim(). Failed to read the video data.\n");
        return false;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function connects to the server through a socket.
//-----------------------------------------------------------------------
int USARSimVideo::connectToServer_(string hostMacine, int port)
{
    struct sockaddr_in sin;
    struct hostent *hent = NULL;
    int sockfd;
    int status = INIT_CONNECTION_STATUS_FAILURE;

    sin.sin_family = PF_INET;
    sin.sin_port = htons(port);

    if (hostMacine == EMPTY_STRING_)
    {
        printfTextWindow(
            "Error: USARSimVideo::connectToServer_(). Host name is empty.\n");
        return INIT_CONNECTION_STATUS_FAILURE;
    }

    if (isdigit(hostMacine[0]))
    {
        sin.sin_addr.s_addr = inet_addr(hostMacine.c_str());
    }
    else
    {
        // Get the IP address.
        if((hent = gethostbyname(hostMacine.c_str())) == NULL)
        {
            printfTextWindow(
                "Error: USARSimVideo::connectToServer_(). Unknown host: %s\n",
                hostMacine.c_str());
            return INIT_CONNECTION_STATUS_FAILURE;
        }

        memcpy((char*) &sin.sin_addr, hent->h_addr, sizeof(sin.sin_addr));
    }

    bzero(&(sin.sin_zero), 8);

    // Open the socket.
    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        printfTextWindow(
            "Error: USARSimVideo::connectToServer_(). Socket could not be opened.\n");
        return INIT_CONNECTION_STATUS_FAILURE;
    }

    // Connect
    if (connect(sockfd, (sockaddr*)&sin, sizeof(sin)))
    {
        close(sockfd);

        switch(errno) {

        case ECONNREFUSED:
            status = INIT_CONNECTION_STATUS_NOT_READY;
            break;

        case ENETUNREACH:
            printfTextWindow(
                "Error: USARSimVideo::connectToServer_(). Network unreachable from this host.");
            status = INIT_CONNECTION_STATUS_FAILURE;
            break;

        default:
            printfTextWindow(
                "Error: USARSimVideo::connectToServer_(). Could not connect to socket.");
            status = INIT_CONNECTION_STATUS_FAILURE;
            break;
        }

        return status;
    }

    shSocketFd_ = sockfd;

    return INIT_CONNECTION_STATUS_SUCCESS;
}

//-----------------------------------------------------------------------
// This function reads n bytes from the file descriptor.
//-----------------------------------------------------------------------
bool USARSimVideo::readNBytes_(const int fd, unsigned char *buf, const int n)
{
    int msgLength = -1, index = 0, remaining = 0;

    remaining = n;

    while (remaining > 0)
    {
        msgLength = read(fd, &buf[index], remaining);

        if (msgLength == 0)
        {
            printfTextWindow(
                "USARSimVideo: WARNING! Server may have died. Reconnection invoked.\n");
            initializeUSARSim_();
            return false;
        }
        else if (msgLength <= 0)
        {
            printfTextWindow(
                "USARSimVideo: WARNING! Read error %d occured. Reconnection invoked.\n",
                errno);
            initializeUSARSim_();
            return false;
        }

        index += msgLength;
        remaining -= msgLength;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void USARSimVideo::updateStatusBar_(void)
{
    char buf[100];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[0] = 0;
    strcat(statusStr, "USARSimVideo [");

    if (frameRate_ > 0)
    {
        sprintf(
            buf,
            "%s]: %c %.2f frames/sec",
            usarsimVideoName_.c_str(),
            statusbarSpinner_->getStatus(),
            frameRate_);
    }
    else
    {
        sprintf(
            buf,
            "%s]: %c Connecting...",
            usarsimVideoName_.c_str(),
            statusbarSpinner_->getStatus());
    }

    if (DISPLAY_FREQUENCY)
    {
        getFrequency_(frequency);
        sprintf(buf, "%s : %.2f Hz", buf, frequency);
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

//-----------------------------------------------------------------------
// This function reads the video data from usarsim.
//-----------------------------------------------------------------------
bool USARSimVideo::readVideoData_(void)
{
    fd_set readMask;
    unsigned char oneByteBuf[1];
    unsigned char *origImagePtr = NULL;
    double currentTime;
    int imageType;
    int imageWidth, imageHeight, imageSize;
    int nfds, nfd;
    unsigned int dataSize, readSize, readCount;
    unsigned short shortValue;
    const bool BLOCK_ON = false;

    currentTime = getCurrentEpochTime();
    readCount = 0;

    pthread_mutex_lock(&socketMutex_);

    // Get the image type.
    readNBytes_(shSocketFd_, (unsigned char *)&oneByteBuf, 1);
    imageType = (int)oneByteBuf[0];

    switch (imageType) {

    case USARSIM_IMAGETYPE_RAW:
        break;

    case USARSIM_IMAGETYPE_JPEG_QUALITYSUPERB:
    case USARSIM_IMAGETYPE_JPEG_QUALITYGOOD:
    case USARSIM_IMAGETYPE_JPEG_QUALITYNORMAL:
    case USARSIM_IMAGETYPE_JPEG_QUALITYAVERAGE:
    case USARSIM_IMAGETYPE_JPEG_QUALITYBAD:
        printfTextWindow(
            "Error: USARSimVideo::readVideoData(). JPEG not supported.\n");
        pthread_mutex_unlock(&socketMutex_);
        return false;

    default:
        printfTextWindow(
            "Error: USARSimVideo::readVideoData(). Corrupted Data.\n");
        pthread_mutex_unlock(&socketMutex_);
        return false;
    }

    // Get the data length.
    readNBytes_(shSocketFd_, (unsigned char *)&dataSize, sizeof(unsigned int));
    dataSize = ntohl(dataSize);
    imageSize = (int)dataSize;

    // If it is a raw, get the dimensions.
    if (imageType == USARSIM_IMAGETYPE_RAW)
    {
        readNBytes_(shSocketFd_, (unsigned char *)&shortValue, sizeof(unsigned short));
        imageWidth = (int)ntohs(shortValue);
        readNBytes_(shSocketFd_, (unsigned char *)&shortValue, sizeof(unsigned short));
        imageHeight = (int)ntohs(shortValue);

        imageSize -= (2*sizeof(unsigned short));
        readCount += (2*sizeof(unsigned short));
    }

    // Prepare the image data.
    imageBufferIndex_++;
    imageBufferIndex_ = (imageBufferIndex_ >= IMAGE_BUFFER_SIZE_)? 0 : imageBufferIndex_;

    if (imageBuffer_[imageBufferIndex_].image != NULL)
    {
        if (imageBuffer_[imageBufferIndex_].imageSize != imageSize)
        {
            delete [] imageBuffer_[imageBufferIndex_].image;
            imageBuffer_[imageBufferIndex_].image = new unsigned char[imageSize];
            imageBuffer_[imageBufferIndex_].imageSize = imageSize;
        }
    }
    else
    {
        imageBuffer_[imageBufferIndex_].image = new unsigned char[imageSize];
        imageBuffer_[imageBufferIndex_].imageSize = imageSize;
    }

    origImagePtr = imageBuffer_[imageBufferIndex_].image;

    while (readCount < dataSize)
    {
        if (BLOCK_ON)
        {
            FD_ZERO(&readMask);
            FD_SET(shSocketFd_, &readMask);

            nfds = shSocketFd_ + 1;
            nfd = select(nfds, &readMask, NULL, NULL, NULL);

            if ((nfd == -1) || (!FD_ISSET(shSocketFd_, &readMask)))
            {
                printfTextWindow(
                    "Error: USARSimVideo::readVideoData(). select() failed.\n");
                pthread_mutex_unlock(&socketMutex_);
                return false;
            }
        }

        readSize = dataSize - readCount;
        readSize = (readSize <= MAX_READ_BLOCK_SIZE_)? readSize : MAX_READ_BLOCK_SIZE_;
        readNBytes_(shSocketFd_, imageBuffer_[imageBufferIndex_].image, (int)readSize);
        imageBuffer_[imageBufferIndex_].image += readSize;
        readCount += readSize;
    }

    // Send the acknowledgment.
    write(shSocketFd_, (const void *)USARSIM_ACK_STRING_.c_str(), USARSIM_ACK_STRING_.size());

    pthread_mutex_unlock(&socketMutex_);

    // Correct the image.
    imageBuffer_[imageBufferIndex_].image = origImagePtr;

    flipY_(
        imageWidth,
        imageHeight,
        &(imageBuffer_[imageBufferIndex_].image));

    pthread_mutex_lock(&currentImageMutex_);
    shCurrentImage_ = imageBuffer_[imageBufferIndex_].image;
    shCurrentImageSize_ = imageSize;
    shCurrentImageWidth_ = imageWidth;
    shCurrentImageHeight_ = imageHeight;
    shCurrentImageTimeStamp_ = currentTime;
    pthread_mutex_unlock(&currentImageMutex_);

    if ((lastCapturedTime_ != 0) && (currentTime != lastCapturedTime_))
    {
        frameRate_ = 1.0 / (currentTime - lastCapturedTime_);
    }

    lastCapturedTime_ = currentTime;

    return true;
}

//-----------------------------------------------------------------------
// This function sets the channel.
//-----------------------------------------------------------------------
void USARSimVideo::setChannel(int channel)
{
    // Channel not supported in USARSimVideo.
}

//-----------------------------------------------------------------------
// This function sets up the user interface for the USARSimVideo.
//-----------------------------------------------------------------------
void USARSimVideo::control(void)
{
    bool done = false;
    int msgData[2];
    int c;

    msgData[0] = 0;
    msgData[1] = 0;

    /*
    messageDrawWindow(
        EnMessageType_USARSIM_VIDEO_CONTROL,
        EnMessageErrType_NONE,
        msgData);
    */

    do
    {
        c = getch();

        switch(c) {

        case 'd':
            messageHide();
            delete this;
            return;
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;

        }
    } while (!done);

    messageHide();
}

/**********************************************************************
# $Log: usarsim-video.c,v $
# Revision 1.2  2009-08-28 19:46:40  zkira
# Rovio fixes and added header includes to fix compilation on other machines
#
# Revision 1.1  2008/07/15 22:33:55  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.2  2008/01/21 22:23:14  endo
# USARSimVideo implemented.
#
# Revision 1.1  2008/01/21 19:14:16  endo
# Regular backup.
#
#**********************************************************************/
