/**********************************************************************
 **                                                                  **
 **                              video.c                             **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Video framegrabber control for HServer                          **
 **                                                                  **
 **  Copyright 2000 - 2008, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: video.c,v 1.2 2008-07-15 22:33:55 endo Exp $ */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "HServerTypes.h"
#include "hserver.h"
#include "message.h"
#include "statusbar.h"
#include "video.h"
#include "xwindow.h"
#include "robot_config.h"

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------
Video *gVideo = NULL;

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t Video::currentImageMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
//const int Video::READER_THREAD_USLEEP_ = 20000;
const int Video::READER_THREAD_USLEEP_ = 1000;
const int Video::DISPLAY_THREAD_USLEEP_ = 20000;
const int Video::SKIP_STATUSBAR_UPDATE_ = 10;
const int Video::WATCHDOG_CHECK_INTERVAL_SEC_ = 2;

//-----------------------------------------------------------------------
// Constructor
//-----------------------------------------------------------------------
Video::Video(
    Video** a,
    int videoType,
    string name,
    bool showImage) :
    Module((Module**)a, name),
    xwindow_(NULL),
    frameRate_(0),
    lastCapturedTime_(0),
    shCurrentImageTimeStamp_(0),
    videoType_(videoType),
    shCurrentImageWidth_(0),
    shCurrentImageHeight_(0),
    shCurrentImageSize_(0),
    shCurrentImage_(NULL),
    showImage_(showImage),
    readerThreadIsUp_(false),
    displayThreadIsUp_(false)
{
    // Initialize mutexes.
    pthread_mutex_init(&currentImageMutex_, NULL);
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
Video::~Video()
{
    if (displayThreadIsUp_)
    {
        displayThreadIsUp_ = false;
        pthread_cancel(displayThread_);
        pthread_join(displayThread_, NULL);
    }

    if (readerThreadIsUp_)
    {
        readerThreadIsUp_ = false;
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    if (xwindow_ != NULL)
    {
        delete xwindow_;
        xwindow_ = NULL;
    }
}

//-----------------------------------------------------------------------
// This function starts the reader thread.
//-----------------------------------------------------------------------
void *Video::startReaderThread_(void *videoInstance)
{
    ((Video*)videoInstance)->readerLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function starts the display thread.
//-----------------------------------------------------------------------
void *Video::startDisplayThread_(void *videoInstance)
{
    ((Video*)videoInstance)->displayLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to grab the video image.
//-----------------------------------------------------------------------
void Video::readerLoop_(void)
{
    int count = 0;

    readerThreadIsUp_ = true;
    
    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Read the data from the gazebo.
        readVideoData_();

        // Check for the pthread termination again.
        pthread_testcancel();

        // Compute the frequency.
        compFrequency_();

        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        if (watchdog_ != NULL)
        {
            watchdog_->notifyUpdate();
        }

        usleep(READER_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function loops around to display the video image.
//-----------------------------------------------------------------------
void Video::displayLoop_(void)
{
    displayThreadIsUp_ = true;
    
    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Show the current image.
        showCurrentImage_();

        // Check for the pthread termination again.
        pthread_testcancel();

        usleep(DISPLAY_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function displays the current image in the xwindow.
//-----------------------------------------------------------------------
void Video::showCurrentImage_(void)
{
    double timeStamp;
    int width, height, imageSize;
    unsigned char *image = NULL;

    image = duplicateCurrentImage(&width, &height, &imageSize, &timeStamp);

    if ((image == NULL) || (xwindow_ == NULL))
    {
        return;
    }

    xwindow_->displayRGBImage(
        0,
        0,
        width,
        height,
        image);

    delete [] image;
    image = NULL;

}

//-----------------------------------------------------------------------
// This function convert a BGR base image into a RGB image. 
//-----------------------------------------------------------------------
bool Video::bgr2rgb_(
    int w,
    int h,
    const unsigned char *bgrImage,
    unsigned char **rgbImage)
{
    int x, y, idx;
    unsigned char *rgbImagePtr = NULL, *rgbImagePtrOrig = NULL;
    unsigned char *bgrImagePtr = NULL;

    rgbImagePtrOrig = *rgbImage;

    for (y = 0; y < h; y++)
    {
        for (x = 0; x < w; x++)
        {
            idx = ((y*w)+x)*3;

            rgbImagePtr = (unsigned char *)(rgbImagePtrOrig + idx + 2);
            bgrImagePtr = (unsigned char *)(bgrImage + idx);

            if (rgbImagePtr != NULL)
            {
                // Copy blue.
                *rgbImagePtr = *bgrImagePtr;

                // Copy green.
                rgbImagePtr--;
                bgrImagePtr++;
                *rgbImagePtr = *bgrImagePtr;

                // Copy red.
                rgbImagePtr--;
                bgrImagePtr++;
                *rgbImagePtr = *bgrImagePtr;
            }
            else
            {
                printfTextWindow("Xwindow::bgr2rgb_(): Insufficient memory allocation.");
                return false;
            }
        }
    }

    *rgbImage = rgbImagePtrOrig;
    
    return true;
}

//-----------------------------------------------------------------------
// This function convert a BGR base image into a RGB image. 
//-----------------------------------------------------------------------
bool Video::flipY_(
    int w,
    int h,
    unsigned char **image)
{
    const int NUM_COLORS = 3;
    int i, x, y, idxFront, idxBack;
    unsigned char frontColor, backColor;
    unsigned char *frontPtr = NULL, *backPtr = NULL, *imagePtrOrig = NULL;

    imagePtrOrig = *image;

    for (y = 0; y < (h/2); y++)
    {
        for (x = 0; x < w; x++)
        {
            idxFront = ((y*w) + x)*NUM_COLORS;
            idxBack = (((h-1-y)*w)+x)*NUM_COLORS;

            frontPtr = (unsigned char *)(imagePtrOrig + idxFront);
            backPtr = (unsigned char *)(imagePtrOrig + idxBack);

            if ((frontPtr != NULL) && (backPtr != NULL))
            {
                for (i = 0; i < NUM_COLORS; i++)
                {
                    // Copy the color
                    frontColor = *frontPtr;
                    backColor = *backPtr;
                    
                    // Switch the colors
                    *frontPtr = backColor;
                    *backPtr = frontColor;

                    // Increment the pointers.
                    frontPtr++;
                    backPtr++;
                } 
            }
            else
            {
                printfTextWindow("Xwindow::flipY_(): Insufficient memory allocation.");
                return false;
            }
        }
    }

    *image = imagePtrOrig;
    
    return true;
}

//-----------------------------------------------------------------------
// This function dumps the most recent image into a PPM file.
//-----------------------------------------------------------------------
void Video::dumpCurrentImage(FILE *ppmFile)
{
    double timeStamp;
    int i, j;
    int width, height, imageSize;
    unsigned char *image = NULL;

    if (ppmFile == NULL)
    {
        printfTextWindow("Error: Video::dumpCurrentImage_(). Output file is NULL.");
        return;
    }
	
    image = duplicateCurrentImage(&width, &height, &imageSize, &timeStamp);

    if (image == NULL)
    {
        printfTextWindow("Error: Video::dumpCurrentImage_(). Current image is NULL.");
        return;
    }
    
    fprintf(ppmFile, "P6\n %d %d\n 255\n", width, height);

    for(i = 0; i < height; i++)
    {
        for(j = 0; j < width; j++)
        {
            fwrite(image + ((i*width) + j) * 3 + 2, 1, 1, ppmFile);
            fwrite(image + ((i*width) + j) * 3 + 1, 1, 1, ppmFile);
            fwrite(image + ((i*width) + j) * 3,     1, 1, ppmFile);
        }
    }

    fflush(ppmFile);

    delete [] image;
    image = NULL;
}

//-----------------------------------------------------------------------
// This function returns the copy of the current image. Make sure to
// free the memory of the returned image after this function is called.
//-----------------------------------------------------------------------
unsigned char *Video::duplicateCurrentImage(
    int *width,
    int *height,
    int *imageSize,
    double *timeStamp)
{
    unsigned char *imageCopy = NULL;

    pthread_mutex_lock(&currentImageMutex_);

    *width = shCurrentImageWidth_;
    *height = shCurrentImageHeight_;
    *imageSize = (*width) * (*height) * 3;
    *timeStamp = shCurrentImageTimeStamp_;

    if (shCurrentImage_ != NULL)
    {
        imageCopy = new unsigned char[(*imageSize)];
        memcpy(imageCopy, shCurrentImage_, *imageSize);
    }

    pthread_mutex_unlock(&currentImageMutex_);
    

    return imageCopy;
}

/**********************************************************************
# $Log: video.c,v $
# Revision 1.2  2008-07-15 22:33:55  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.10  2008/04/11 20:22:44  endo
# FrameGrabber fixed.
#
# Revision 1.9  2008/01/21 22:23:14  endo
# USARSimVideo implemented.
#
# Revision 1.8  2008/01/21 19:14:17  endo
# Regular backup.
#
# Revision 1.7  2006/12/07 20:30:23  endo
# Regular backup.
#
# Revision 1.6  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.5  2006/11/24 23:17:14  endo
# OpenCV BlobTrack improved.
#
# Revision 1.4  2006/11/22 22:24:28  endo
# OpenCV BlobTrack implemented.
#
# Revision 1.3  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.2  2006/11/20 16:43:37  endo
# Xwindow improved.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2006/05/15 06:01:03  endo
# gcc-3.4 upgrade
#
# Revision 1.1.1.1  2005/02/06 22:59:43  endo
# AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
#
# Revision 1.6  2003/04/06 13:32:37  endo
# gcc 3.1.1
#
# Revision 1.5  2001/03/23 21:32:46  blee
# altered to use a config file
#
# Revision 1.4  2000/12/12 23:23:24  blee
# Changed #defines for the user interface to enumerated types.
#
# Revision 1.3  2000/09/19 03:51:29  endo
# RCS log added.
#
#
#**********************************************************************/
