/* MissionLab log analyzer
   by Zsolt Kira
*/

#include <stdio.h>
#include <math.h>
#include <vector.h>

// Skip header lines in log files (hack, should really be done analytically)
#define NUM_JUNK_LINES 12

int main(int argc, char *argv[]) {

  char line[10000], out_file[10000];
  char ch;
  FILE *in, *out;
  int ret = -1;

  // Fields in log file
  double time=0, x=0, y=0, z=0, heading=0, roll=0, pitch=0, dist=0, vx=0, vy=0, vz=0, previous_x=0, previous_y=0, previous_z=0;
  int id, fsa_num, state, previous_state;

  // Decides whether ALL missions are included in averages, regardless of whether mission was completed
  int count_all = 1;

  // Decides whether number of steps counting should stop once state is reached, or should continues until end of file
  int keep_state = 0;

  // Decides whether to output averages, etc. into a file or stdout
  int use_output_file = 0;

  // Use z component for various calculations
  int use_3D = 1;

  // Start and end states
  int begin_state = -1, end_state = -1;

  // Booleans signifying whether start or end states have been reached
  int begin = 0, end = 0;

  double start_time = 0, end_time = -1;
 
  /*************  Statistical information that is kept ******************/

  // Distance travelled; 
  double stat_distance_travelled = 0, stat_total_distance_travelled = 0;

  // Time Taken
  double stat_time_taken, stat_total_time_taken = 0;

  // Steps taken
  int stat_step = 0, stat_total_steps = 0;

  // Keep all history of data for standard deviation purposes
  vector <double> stat_distances;
  vector <double> stat_times;
  vector <int> stat_steps;

  /**********************************************************************/

  // Number of instances encountered (num_instances are number that are counted, total_num_instances are all instances encountered)
  int num_instances = 0, total_num_instances = 0;

  // Track numbers of error conditions
  int num_could_not_open = 0, num_did_not_start = 0, num_did_not_end = 0, num_incomplete = 0;
  
  int line_counter;
  int i;
  int result;
    
  if (argc < 2) {
    printf("Usage: %s [options] files\n", argv[0]);
    printf("Options: -ss int    Start state - Begin counting log once this state is first reached (-1 to start from first state, default is -1)\n");
    printf("         -es int    End state - Stop counting log once this state is reached (-1 for never, default is -1)\n");
//    printf("         -o string  NOT IMPLEMENTED Output string specifying output for each file\n");
//    printf("                    Format is:\n"); 
    printf("         -k         Keeps the last state (counts all times steps, even after in finish state) (default is off)\n");
    printf("         -s         Output the stats averaged over all files in a file of the given name (default is stdout).\n");
    printf("         -nc        Do not count failed missions in averaging of various statistics (eg. distance or times steps).\n");
    printf("         -ts        Start time - Time of mission after which statistics should be counte (in seconds)\n");
    printf("         -te        End time - Time after which statistics should not be counted (in seconds)\n");
    exit(0);
  }

  // Take in arguments
  for (i = 1; i < argc; i++) {

    if (strcmp(argv[i], "-k") == 0) {
      printf("Option: Keep state\n");
      keep_state = 1;
      continue;
    } else if (strcmp(argv[i], "-nc") == 0) {
      printf("Option: Not counting failed missions in averaging\n");
      count_all = 0;
      continue;
    } else if (strcmp(argv[i], "-s") == 0) {
      if (i+1 == argc) {
        fprintf(stderr, "%s: Error: Option -s given without file name after it\n", argv[0]);
        exit(0);
      }
      strcpy(out_file, argv[++i]);
      use_output_file = 1;
      printf("Option: Output to file %s\n", out_file);
      continue;
    } else if (strcmp(argv[i], "-ss") == 0) {
      if (i+1 == argc) {
        fprintf(stderr, "%s: Error: Option -ss given without number after it\n", argv[0]);
        exit(0);
      }
      begin_state = atoi(argv[++i]);
      printf("Option: Start state is now %d\n", begin_state);
      continue;
    } else if (strcmp(argv[i], "-es") == 0) {
      if (i+1 == argc) {
        fprintf(stderr, "%s: Error: Option -ss given without number after it\n", argv[0]);
        exit(0);
      }
      end_state = atoi(argv[++i]);
      printf("Option: End state is now %d\n", end_state);
      continue;
    } else if (strcmp(argv[i], "-es") == 0) {
      if (i+1 == argc) {
        fprintf(stderr, "%s: Error: Option -ss given without number after it\n", argv[0]);
        exit(0);
      }
      end_state = atoi(argv[++i]);
      printf("Option: End state is now %d\n", end_state);
      continue;
    } else if (strcmp(argv[i], "-ts") == 0) {
      if (i+1 == argc) {
        fprintf(stderr, "%s: Error: Option -ts given without number after it\n", argv[0]);
        exit(0);
      }
      start_time = atof(argv[++i]);
      printf("Option: Start time is now %f\n", start_time);
      continue;
    } else if (strcmp(argv[i], "-te") == 0) {
      if (i+1 == argc) {
        fprintf(stderr, "%s: Error: Option -te given without number after it\n", argv[0]);
        exit(0);
      }
      end_time = atof(argv[++i]);
      printf("Option: End time is now %f\n", end_time);
      continue;
    }

    total_num_instances++;
  
    in = fopen(argv[i], "r");
    if (in == NULL) {
      fprintf(stderr, "%s: Error: Could not open file %s\n", argv[0], argv[i]);
      num_could_not_open++;
      continue;
    }

    begin = 0;
    end = 0;
    line_counter = 0;

    stat_step = 0;
    stat_distance_travelled = 0;
    stat_time_taken = -1;
    
    while (1) {

      line_counter++;
    
      result = fscanf(in, "%[^\n]s", line);

      // Get rid of newline
      fscanf(in, "%c", &ch);
      
      // Skip first lines filled with junk
      if (line_counter <= NUM_JUNK_LINES)
        continue;
 
      // If reached file end, and we have reached the desired end state, note time taken to do so     
      if (result == EOF) {
        if (end_state < 0 | end == 1) {
          stat_time_taken = time;
        }
        break;
      }

      // Read in line of data      
      if ( (ret = sscanf(line, "%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %d (%d,%d) %lf", &time, &x, &y, &z, &heading, &roll, &pitch, &vx, &vy, &vz, &id, &fsa_num, &state, &dist)) != 14) {
        fprintf(stderr, "%s: %s: Error: Could not get all data (got %d) on line %d containing\n", argv[0], argv[i], ret, line_counter);
        fprintf(stderr, "%s\n", line);
        num_incomplete++;
        break;
      } 

      //printf("State: %d (start state=%d) end=%d\n", state, begin_state, end);
      
      /********* DATA ANALYSIS HERE *********/
      
      if (line_counter > NUM_JUNK_LINES) {

        if (end == 1) {
          if (state != end_state) {
            stat_time_taken = time;
            break;
          }
        }

        if (!begin) {
          if (begin_state < 0 || state == begin_state)
            begin = 1;
          else
            continue;
        }

        /*
        if (time < start_time) {
          continue;
        }

        if (time > end_time) {
          continue;
        }
        */
      }

      // Make sure this is not first line of real data (no valid previous_x or _y)
      if (line_counter > NUM_JUNK_LINES+1 ) {

        if (use_3D) {
	        stat_distance_travelled  += sqrt( (previous_x - x) * (previous_x - x) + 
        	                                  (previous_y - y) * (previous_y - y) +
                	                          (previous_z - z) * (previous_z - z) );
	} else {
	        stat_distance_travelled  += sqrt( (previous_x - x) * (previous_x - x) + 
        	                                  (previous_y - y) * (previous_y - y));
	}
        stat_time_taken = time;
        stat_step++;

        // If reached end state, exit or continue depending on keep_state bool        
        if (state == end_state) {
          end = 1;
          if (keep_state) {
          } else {
            stat_time_taken = time;
            break;
          }
        }
        
      }
      
      previous_x = x;
      previous_y = y;
      previous_z = z;
      previous_state = state;
      
      /**************************************/
      
    }

    if (begin == 0) {
      fprintf(stderr, "%s: %s: Error: Never reached start state (%d)\n", argv[0], argv[i], begin_state);
      num_did_not_start++;
      //continue;
    }

    if (num_instances == 0)
      printf(" Distance    Time    Steps         Filename\n");

    num_instances++;

    if (!end) {
      //fprintf(stderr, "%s: %s: Error: Time taken to complete not available (never finished?)\n", argv[0], argv[i]);
      num_did_not_end++;

      if (!count_all) {
        printf("%f %f %d %s\n", stat_distance_travelled, stat_time_taken, -stat_step, argv[i]);
        continue;
      }

      //printf("%f %f %d %s\n", stat_distance_travelled, stat_time_taken, -stat_step, argv[i]);
      //continue;
    }
    else {
      //printf("%s completed\n", argv[i]);
    }

    printf("%f %f %d %s\n", stat_distance_travelled, stat_time_taken, stat_step, argv[i]);

    stat_total_distance_travelled += stat_distance_travelled;
    stat_total_time_taken += stat_time_taken;
    stat_total_steps += stat_step;
            
    stat_distances.push_back(stat_distance_travelled);
    stat_times.push_back(stat_time_taken);
    stat_steps.push_back(stat_step);
    
    fclose(in);
  }

  double stat_mean_distance;
  double stat_mean_time;
  double stat_mean_steps;

  if (use_output_file) {
    out = fopen(out_file, "w");
    if (out == NULL) {
      fprintf(stderr, "%s: Error: Could not open output file %s\n", argv[0], out_file);
      exit(1);
    }
  }  else {
    out = stdout;
  }
  
  if (num_instances > 0) {
    stat_mean_distance = stat_total_distance_travelled / num_instances;
    stat_mean_time = stat_total_time_taken / num_instances;
    stat_mean_steps = stat_total_steps / num_instances;
  } else {
    fprintf(out, "No valid data\n");
  }
    
  double stat_stdev_distance = 0;
  double stat_stdev_time = 0;
  double stat_stdev_step = 0;

  if (num_instances > 1) {
 
    i = num_instances - 1;
    for ( ; i >= 0; i--) {

      //printf("%d: %f %f\n", i, stat_distances[i], stat_times[i]);
      
      stat_stdev_distance += (stat_distances[i] - stat_mean_distance) *
                             (stat_distances[i] - stat_mean_distance);
 
      stat_stdev_time += (stat_times[i] - stat_mean_time) *
                         (stat_times[i] - stat_mean_time);
 
      stat_stdev_step += (stat_steps[i] - stat_mean_steps) *
                         (stat_steps[i] - stat_mean_steps);
    }

    stat_stdev_distance = stat_stdev_distance / (num_instances - 1);
    stat_stdev_time = stat_stdev_time / (num_instances - 1);
    stat_stdev_step = stat_stdev_step / (num_instances - 1);
    
    stat_stdev_distance = sqrt(stat_stdev_distance);
    stat_stdev_time = sqrt(stat_stdev_time);
    stat_stdev_step = sqrt(stat_stdev_step);

  }

  if (num_instances > 0) {
    fprintf(out, "\n%d instances:\n (%d total files,\n  %d could not be opened,\n  %d did not start,\n  %d did not end,\n  %d with incomplete lines)\n\n",
            num_instances, total_num_instances, num_could_not_open, num_did_not_start, num_did_not_end, num_incomplete);
    fprintf(out, "Avg distance: %10.2f   (stdev: %.2f)\n", stat_mean_distance, stat_stdev_distance);
    fprintf(out, "Avg time:     %10.2f   (stdev: %.2f)\n", stat_mean_time, stat_stdev_time);
    fprintf(out, "Avg steps:    %10.2f   (stdev: %.2f)\n", stat_mean_steps, stat_stdev_step);
  }
  
  if (use_output_file)
    fclose(out);
}
